""" ESP Wi-Fi Allaince Sigma agent software

   Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
   Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.


   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 """


def get_params_attribute(params, key):
    """ Parse capi params attributes

    Parameters:
        CAPI params (dict): CAPI command parameters in dict

    Returns:
        Param Value (str|dict|list): CAPI parameters values
    """
    for k in params.keys():
        if k.lower() == key.lower():
            return params[k]
    return None
    '''
    if key in params:
        return params[key]
    else:
        return None
    '''


class ca_capi_cmd:
    """ Capi public classes initialised with UCC commands

    """

    class ca_get_version:
        def __init__(self, params):
            pass

    class device_get_info:
        def __init__(self, params):
            pass

    class device_list_interfaces:
        def __init__(self, params):
            self.interfaceType = get_params_attribute(params, 'interfaceType')
    class sta_reset_default:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.prog = get_params_attribute(params, 'program')
            self.type = get_params_attribute(params, 'type')

    class sta_get_info:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')

    class sta_get_mac_address:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')

    class sta_get_parameter:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.parameter = get_params_attribute(params, 'parameter')

    class sta_set_encryption:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ssid = get_params_attribute(params, 'ssid')
            self.encpType = get_params_attribute(params, 'encpType')
            self.key1 = get_params_attribute(params, 'key1')

    class sta_scan:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.parameter   = get_params_attribute(params, 'GetParameter')

    class dev_send_frame:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.name = get_params_attribute(params, 'NAME')
            self.program = get_params_attribute(params, 'program')
            self.dest_mac = get_params_attribute(params, 'Dest_Mac')
            self.framename = get_params_attribute(params, 'FrameName')
            self.wnm_notify_element = get_params_attribute(
                params, 'WNM_Notify_Element')

    class dev_exec_action:
        def __init__(self, params):
            self.ServerCertTrust = get_params_attribute(
                params, 'ServerCertTrust')

    class sta_set_pwrsave:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.mode      = get_params_attribute(params, 'mode')

                
    class sta_set_eaptls:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ssid = get_params_attribute(params, 'ssid')
            self.username = get_params_attribute(params, 'username')
            self.trustedRootCA = get_params_attribute(params, 'trustedRootCA')
            self.clientCertificate = get_params_attribute(params, 'clientCertificate')
            self.encpType = get_params_attribute(params, 'encpType')
            self.KeyMgmtType = get_params_attribute(params, 'keymgmttype')
            self.passphrase = get_params_attribute(params, 'passphrase')

    class sta_set_rfeature:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ltf = get_params_attribute(params, 'LTF')
            self.gi = get_params_attribute(params, 'GI')
            self.prog = get_params_attribute(params, 'prog')
            self.transmitOMI = get_params_attribute(params, 'transmitOMI')
            self.OMCtrl_TxNSTS = get_params_attribute(params, 'OMCtrl_TxNSTS')
            self.OMCtrl_ChnlWidth = get_params_attribute(params, 'OMCtrl_ChnlWidth')
            self.OMCtrl_ULMUDisable = get_params_attribute(params, 'OMCtrl_ULMUDisable')
            self.OMCtrl_ULMUDataDisable = get_params_attribute(params, 'OMCtrl_ULMUDataDisable')
            #mbo

            self.ch_pref_num = get_params_attribute(params, 'Ch_Pref_Num')
            self.ch_pref = get_params_attribute(params, 'Ch_Pref')
            self.ch_op_class = get_params_attribute(params, 'Ch_Op_Class')
            self.ch_reason_code = get_params_attribute(params, 'Ch_Reason_Code')

            self.NDPPagingInd = get_params_attribute(params, 'NDPPagingInd')
            self.RespPMMode = get_params_attribute(params, 'RespPMMode')
            self.NegotiationType = get_params_attribute(params, 'NegotiationType')
            self.TWT_Setup = get_params_attribute(params, 'TWT_Setup')
            self.SetupCommand = get_params_attribute(params, 'SetupCommand')
            self.TWT_Trigger = get_params_attribute(params, 'TWT_Trigger')
            self.implicit = get_params_attribute(params, 'implicit')
            self.FlowType = get_params_attribute(params, 'FlowType')
            self.WakeIntervalExp = get_params_attribute(params, 'WakeIntervalExp')
            self.Protection = get_params_attribute(params, 'Protection')
            self.NominalMinWakeDur = get_params_attribute(params, 'NominalMinWakeDur')
            self.WakeIntervalMantissa = get_params_attribute(params, 'WakeIntervalMantissa')
            self.TWT_Channel = get_params_attribute(params, 'TWT_Channel')

            self.FlowID = get_params_attribute(params, 'FlowID')
            #TWT Information
            self.TWT_Operation = get_params_attribute(params, 'TWT_Operation')
            #BF test
            self.BF_Type = get_params_attribute(params, 'BF_Type')
            self.RuStart = get_params_attribute(params, 'RuStart')
            self.RuEnd = get_params_attribute(params, 'RuEnd')
            self.device_num = get_params_attribute(params, 'device')
            # ERSU
            self.PPDUTxType = get_params_attribute(params, 'PPDUTxType')
            self.RUAllocTone = get_params_attribute(params, 'RUAllocTone')


    class sta_set_11n:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.prog = get_params_attribute(params, 'prog')
            self.transmitOMI = get_params_attribute(params, 'transmitOMI')
            self.OMCtrl_TxNSTS = get_params_attribute(params, 'OMCtrl_TxNSTS')
            self.OMCtrl_ChnlWidth = get_params_attribute(
                params, 'OMCtrl_ChnlWidth')
            self.OMCtrl_ULMUDisable = get_params_attribute(
                params, 'OMCtrl_ULMUDisable')
            self.OMCtrl_ULMUDataDisable = get_params_attribute(
                params, 'OMCtrl_ULMUDataDisable')
            self.NDPPagingInd = get_params_attribute(params, 'NDPPagingInd')
            self.RespPMMode = get_params_attribute(params, 'RespPMMode')
            self.NegotiationType = get_params_attribute(
                params, 'NegotiationType')
            self.TWT_Setup = get_params_attribute(params, 'TWT_Setup')
            self.SetupCommand = get_params_attribute(params, 'SetupCommand')
            self.TWT_Trigger = get_params_attribute(params, 'TWT_Trigger')
            self.implicit = get_params_attribute(params, 'implicit')
            self.FlowType = get_params_attribute(params, 'FlowType')
            self.WakeIntervalExp = get_params_attribute(
                params, 'WakeIntervalExp')
            self.Protection = get_params_attribute(params, 'Protection')
            self.NominalMinWakeDur = get_params_attribute(
                params, 'NominalMinWakeDur')
            self.WakeIntervalMantissa = get_params_attribute(
                params, 'WakeIntervalMantissa')
            self.TWT_Channel = get_params_attribute(params, 'TWT_Channel')
            # iTWT Teardown
            # sta_set_rfeature,Interface,espsta,prog,HE,TWT_Setup,teardown,FlowID,1,NegotiationType,0
            self.FlowID = get_params_attribute(params, 'FlowID')
            # TWT Information
            self.TWT_Operation = get_params_attribute(params, 'TWT_Operation')
            # BF test
            self.BF_Type = get_params_attribute(params, 'BF_Type')
            self.RuStart = get_params_attribute(params, 'RuStart')
            self.RuEnd = get_params_attribute(params, 'RuEnd')
            self.device_num = get_params_attribute(params, 'device')
            # ERSU
            self.PPDUTxType = get_params_attribute(params, 'PPDUTxType')
            self.RUAllocTone = get_params_attribute(params, 'RUAllocTone')


    class sta_set_power_save:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.program = get_params_attribute(params, 'program')
            self.powersave = get_params_attribute(params, 'powersave')

    class sta_set_eapttls:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ssid = get_params_attribute(params, 'ssid')
            self.username = get_params_attribute(params, 'username')
            self.trustedRootCA = get_params_attribute(params, 'trustedRootCA')
            self.clientCertificate = get_params_attribute(params, 'clientCertificate')
            self.encpType = get_params_attribute(params, 'encpType')
            self.KeyMgmtType = get_params_attribute(params, 'keymgmttype')
            self.passphrase = get_params_attribute(params, 'password')

    class sta_set_peap:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ssid = get_params_attribute(params, 'ssid')
            self.username = get_params_attribute(params, 'username')
            self.trustedRootCA = get_params_attribute(params, 'trustedRootCA')
            self.encpType = get_params_attribute(params, 'encpType')
            self.KeyMgmtType = get_params_attribute(params, 'keymgmttype')
            self.passphrase = get_params_attribute(params, 'password')

    class sta_send_addba:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.tid = get_params_attribute(params, 'tid')
            self.Dest_mac = get_params_attribute(params, 'Dest_mac')

    class sta_set_security:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ssid = get_params_attribute(params, 'ssid')
            self.encpType = get_params_attribute(params, 'encpType')
            self.type = get_params_attribute(params, 'Type')
            self.KeyMgmtType = get_params_attribute(params, 'KeyMgmtType')
            self.passphrase = get_params_attribute(params, 'passphrase')
            self.key1 = get_params_attribute(params, 'key1')
            self.username = get_params_attribute(params, 'username')
            self.trustedRootCA = get_params_attribute(params, 'trustedRootCA')
            self.CertType = get_params_attribute(params, 'CertType')
            self.clientCertificate = get_params_attribute(
                params, 'clientCertificate')
            self.keymgmtType = get_params_attribute(params, 'keymgmtType')
            self.PairwiseCipher = get_params_attribute(
                params, 'PairwiseCipher')
            self.GroupCipher = get_params_attribute(params, 'GroupCipher')
            self.GroupMgntCipher = get_params_attribute(
                params, 'GroupMgntCipher')
            self.password = get_params_attribute(params, 'password')
            self.PMKSACaching = get_params_attribute(params, 'PMKSACaching')
            self.AKMSuiteType = get_params_attribute(params, 'AKMSuiteType')
            self.innerEAP = get_params_attribute(params, 'innerEAP')
            self.peapVersion = get_params_attribute(params, 'peapVersion')

    class sta_set_ip_config:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.dhcp = get_params_attribute(params, 'dhcp')
            self.ip = get_params_attribute(params, 'ip')
            self.mask = get_params_attribute(params, 'mask')
            self.defaultGateway = get_params_attribute(
                params, 'defaultGateway')

    class sta_associate:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ssid = get_params_attribute(params, 'ssid')
            self.bssid = get_params_attribute(params, 'bssid')

    class sta_is_connected:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')

    class sta_get_bssid:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')

    class sta_get_ip_config:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')

    class sta_disconnect:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.maintain_profile = get_params_attribute(
                params, 'maintain_profile')

    class sta_reassoc:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.channel = get_params_attribute(params, 'Channel')
            self.bssid = get_params_attribute(params, 'bssid')

    class sta_preset_testparameters:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.supplicant = get_params_attribute(params, 'supplicant')
            self.program = get_params_attribute(params, 'program')
            self.ch_pref_num = get_params_attribute(params, 'Ch_Pref_Num')
            self.ch_pref = get_params_attribute(params, 'Ch_Pref')
            self.ch_op_class = get_params_attribute(params, 'Ch_Op_Class')
            self.ch_reason_code = get_params_attribute(params, 'Ch_Reason_Code')
            self.cellular_data_cap = get_params_attribute(params, 'Cellular_Data_Cap')
            self.roaming = get_params_attribute(params, 'Roaming')

    class sta_set_psk:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.ssid = get_params_attribute(params, 'ssid')
            self.passphrase = get_params_attribute(params, 'passphrase')
            self.encpType = get_params_attribute(params, 'encpType')
            self.keymgmttype = get_params_attribute(params, 'keymgmttype')

    class sta_set_wireless:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
            self.program = get_params_attribute(params, 'program')
            self.txsp_stream = get_params_attribute(params, 'txsp_stream')
            self.rxsp_stream = get_params_attribute(params, 'rxsp_stream')
            self.amsdu = get_params_attribute(params, 'amsdu')
            self.ldpc = get_params_attribute(params, 'LDPC')
            self.mcs = get_params_attribute(params, 'mcs_fixedrate')
            self.bcc = get_params_attribute(params, 'BCC')
            self.width = get_params_attribute(params, 'width')
            self.omcontrol = get_params_attribute(params, 'OMControl')
            self.uph = get_params_attribute(params, 'UPH')

    class traffic_agent_reset:
        def __init__(self, params):
            pass

    class sta_set_systime:
        def __init__(self, params):
            self.year = int(get_params_attribute(params, 'year'))
            self.month = int(get_params_attribute(params, 'month'))
            self.hours = int(get_params_attribute(params, 'hours'))
            self.day = int(get_params_attribute(params,'date'))
            self.minutes = int(get_params_attribute(params, 'minutes'))
            self.seconds = int(get_params_attribute(params, 'seconds'))

    class traffic_send_ping:
        def __init__(self, params):
            self.destination = get_params_attribute(params, 'destination')
            self.framesize = get_params_attribute(params, 'framesize')
            self.frameRate = get_params_attribute(params, 'frameRate')
            self.duration = get_params_attribute(params, 'duration')

    class traffic_stop_ping:
        def __init__(self, params):
            self.streamID = get_params_attribute(params, 'streamID')

    class traffic_agent_config:
        def __init__(self, params):
            self.profile = get_params_attribute(params, 'profile')
            self.direction = get_params_attribute(params, 'direction')
            self.source = get_params_attribute(params, 'source')
            self.destination = get_params_attribute(params, 'destination')
            self.destinationPort = get_params_attribute(params, 'destinationPort')
            self.sourcePort = get_params_attribute(params, 'sourcePort')
            self.frameRate = get_params_attribute(params, 'frameRate')
            self.duration = get_params_attribute(params, 'duration')
            self.payloadSize = get_params_attribute(params, 'payloadSize')
            self.startDelay = get_params_attribute(params, 'startDelay')
            self.trafficClass = get_params_attribute(params, 'trafficClass')
            self.startDelay = get_params_attribute(params, 'startDelay')
            self.HTI = get_params_attribute(params, 'HTI')
            self.maxCnt = get_params_attribute(params, 'maxCnt')

    class traffic_agent_send:
        def __init__(self, params):
            self.streamID = get_params_attribute(params, 'streamID')

    class traffic_agent_receive_start:
        def __init__(self, params):
            self.streamID = get_params_attribute(params, 'streamID')

    class traffic_agent_receive_stop:
        def __init__(self, params):
            self.streamID = get_params_attribute(params, 'streamID')

    class sta_btm_query:
        def __init__(self, params):
            self.reason_code = get_params_attribute(params, "reason_code")
            self.list = get_params_attribute(params, "list")

    class sta_loopback_server_start:
        def __init__(self, params):
            self.port = get_params_attribute(params, "port")

    class device_reset:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')

    class sta_loopback_server_stop:
        def __init__(self, params):
            self.interface = get_params_attribute(params, 'interface')
