""" ESP Wi-Fi Allaince Sigma agent software

   Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
   Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.


   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 """
from common import gvars, logger
from device import dut, response

special_response_format = {'traffic_agent_send': 'traffic_agent_send_resp',
	                   'traffic_agent_receive_stop': 'traffic_agent_receive_stop_resp'}


class api_response:
    def __init__(self, caCmdResp, response):
        self.caCmdResp = caCmdResp
        self.response = response

    def traffic_agent_send_resp(self):
        response = "%s,streamID,%s,txFrames,%s,rxFrames,%s,txPayloadBytes,%s,rxPayloadBytes,%s,outOfSequenceFrames,%s" % (
            self.response, self.caCmdResp.streamID, self.caCmdResp.txFrames, self.caCmdResp.rxFrames,
            self.caCmdResp.txPayloadBytes, self.caCmdResp.rxPayloadBytes, self.caCmdResp.outOfSequenceFrames)
        return response

    def traffic_agent_receive_stop_resp(self):
        response = "%s,streamID,%s,txFrames,%s,rxFrames,%s,txPayloadBytes,%s,rxPayloadBytes,%s,outOfSequenceFrames,%s" % (
            self.response, self.caCmdResp.streamID, self.caCmdResp.txFrames, self.caCmdResp.rxFrames,
            self.caCmdResp.txPayloadBytes, self.caCmdResp.rxPayloadBytes, self.caCmdResp.outOfSequenceFrames)
        # response = "%s,streamID,%s,txFrames,%s,rxFrames,%s,txPayloadBytes,%s,rxPayloadBytes,%s,outOfSequenceFrames,%s" % (
        #     self.response, self.caCmdResp.streamID, "0", self.caCmdResp.rxFrames,
        #     "0", self.caCmdResp.rxPayloadBytes, self.caCmdResp.outOfSequenceFrames)
        return response

    def generic_resp(self):
        for attr, value in self.caCmdResp.__dict__.items():
            if "scan" not in attr:
                self.response += ",%s,%s" % (attr, value)
            else :
                self.response += ",%s" % (value)
        return self.response

    def sta_scan_resp(self):
        response = "%s" % self.response
        if self.caCmdResp.scan is not None:
            for item in self.caCmdResp.scan:
                response+=",%s" % item
        return response

def dump_object_attr(obj):
    for attr, value in obj.__dict__.items():
        if value is not None:
            label = "\t%s: %s"  % (str(attr).ljust(15, ' ') ,str(value).ljust(15, ' '))
            logger.PRINT(label)

    print("\n")


def process_ca_api_request(api, caCmd, caCmdResp):
    dut_target_api = getattr(dut, api)
    if gvars.VERBOSE:
        dump_object_attr(caCmd)
    ret = dut_target_api(caCmd, caCmdResp)

    if ret.status == gvars.SUCCESS:
        response = gvars.CAPI_RESP_SUCCESS
        respObj = api_response(caCmdResp, response)

        if api in special_response_format:
            response = getattr(respObj, special_response_format[api])()
        else:
            response = respObj.generic_resp()

        logger.INFO(response + "\n")
    else:
        logger.ERROR(ret.data)
        response = gvars.CAPI_RESP_INVALID
    # TODO : Ensure this works... Check previous commit changes
    if gvars.QT_ENABLED:
        return ret
    return response
