""" ESP Wi-Fi Allaince Sigma agent software

   Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
   Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.


   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 """

import threading
import serial
import time
import os
from common import gvars

class LoggerReceiver(threading.Thread):
    def __init__(self, dut):
        super(LoggerReceiver, self).__init__()
        self.setDaemon(True)
        self.dut = dut
        self._exit_event = threading.Event()
        self._lock = threading.Lock()
        if not os.path.exists('logs'):
            os.makedirs('logs')
        self._log_file = gvars.LOG_FILE_LOCATION
        self._data_cache = b""

    def run(self):
        while True:
            if self._exit_event.is_set():
                return
            try:
                _data = self.dut.read(1024)
                if not _data:
                    continue
                with self._lock:
                    self._data_cache += _data
                    with gvars.g_data_cache_lock:
                        gvars.g_data_cache += _data
                    with open(self._log_file, "ab+") as f:
                        f.write(_data)
            except Exception:
                pass

    def read_data(self):
        _data = b""
        with self._lock:
            self._data_cache = b""
            with gvars.g_data_cache_lock:
                _data = gvars.g_data_cache
                gvars.g_data_cache = b""
        return _data

    def close(self):
        self._exit_event.set()


class DutPipe(object):
    def __init__(self, *args, **kwargs):
        self.dut_inst = serial.Serial(*args, **kwargs)
        self.receiver = LoggerReceiver(self.dut_inst)
        self.receiver.start()

    def write(self, data):
        self.dut_inst.write(data.encode())
    def read_data(self):
        time.sleep(0.05)
        data = self.receiver.read_data().decode(encoding = "utf-8", errors = "ignore")
        return data
    def close(self):
        self.receiver.close()
        self.dut_inst.close()
