""" ESP Wi-Fi Allaince Sigma agent software

   Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
   Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.


   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 """

import time
import os
import sys
import threading
import serial
from datetime import datetime

from common import gvars, logger ,dutpipe
last_data = dict()


def sync_ca_dut_prompts():

    gvars.g_traffic_send_data_lock = threading.Lock()
    gvars.serialWriteLock = threading.Lock()
    gvars.serialReadLock = threading.Lock()
    gvars.SIGMA_TEST_DATA['data_streams'] = {}
    gvars.serialReadPipeLock = threading.Lock()
    gvars.readers_count = 0
    gvars.g_data_cache_lock = threading.Lock()

    retry = 30
    while retry:
        log = send_serial_command(gvars.DUT_SERIAL_PIPE, "version")
        if 'IDF Version' not in log.strip():
            logger.PRINT("\n>Error: CA/DUT prompt are not in sync, please press reset button on esp device!!!")
            waitResponse("Reconnect after", 30)
        else:
            break
        retry -=1

    if not retry:
        gvars.DUT_SERIAL_PIPE.close()
        sys.exit(0)

    for line in log.split("\n"):
        if 'IDF Version' in line:
            gvars.IDF_VERSION = ":".join(line.split(":")[1:]).strip()
            break

    branch = send_local_command("git rev-parse --abbrev-ref HEAD")
    commit = send_local_command("git rev-parse HEAD")
    gvars.SW_VERSION  = "%s (%s)" % (branch.strip(), commit.strip()[0:7])
    cur_time = time.time()
    log=send_serial_command(gvars.DUT_SERIAL_PIPE,"settime {}".format(int(cur_time)))
    logger.PRINT("\nSystem Time set to  : {}".format(datetime.utcfromtimestamp(cur_time)))
    log = send_serial_command(gvars.DUT_SERIAL_PIPE,"set_prog --program any")
    gvars.SIGMA_TEST_DATA["prog"] = "any"

def waitResponse(m, t):
    for i in range(t, 0, -1):
        print(">{} [{}] sec.".format(m,i), end="\r", flush=True)
        time.sleep(1)

def send_local_command(command):
    if gvars.VERBOSE:
        logger.PRINT(" [lcmd] " + command)

    stream = os.popen(command)
    resp = stream.read()

    if gvars.VERBOSE:
        logger.PRINT(" [resp] " + resp)
    return resp

def send_serial_command(pipe, command, wait=True, waitTime=1):
    with gvars.serialWriteLock:
        pipe = gvars.DUT_SERIAL_PIPE = dutpipe.DutPipe(gvars.DUT, gvars.BAUD_RATE, serial.EIGHTBITS, serial.PARITY_NONE,serial.STOPBITS_ONE, 0.01, False, False, None)
        time.sleep(0.05)
        if gvars.VERBOSE:
            logger.PRINT(" [scmd] " + command)
        pipe.write(command + "\r\n")
        if not wait:
            pipe.close()
            return None
        time.sleep(waitTime)
        log = read_serial_console(pipe)
        if gvars.VERBOSE:
            logger.PRINT("Got Response for {} as {}".format(command,log))
        pipe.close()
    return log

def read_serial_console(pipe, timeout=5):
    time.sleep(0.05)
    start = time.time()
    with gvars.serialReadLock:
        while True:
            data = pipe.read_data()
            if data:
                if gvars.VERBOSE:
                    logger.PRINT(" [console] " + data)
                break
            if time.time() - start > timeout:
                break
    return data
