""" ESP Wi-Fi Allaince Sigma agent software

   Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
   Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.


   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 """

import time
from common import gvars
from common import response, logger
from device import esp_sigma as dut
from datetime import datetime
import math

SUCCESS = gvars.SUCCESS
FAILURE = gvars.FAILURE

def isExist(key):
    if key in gvars.SIGMA_TEST_DATA:
        return True
    return False

def ca_get_version(caCmd, caCmdResp):
    """ Response "ca_get_version" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    caCmdResp.version = gvars.SW_VERSION

    resp.status = SUCCESS
    return resp


def device_get_info(caCmd, caCmdResp):
    """ Response "device_get_info" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.fetch_device_info()
    if resp.status == SUCCESS:
        caCmdResp.firmware = resp.data

    return resp


def device_list_interfaces(caCmd, caCmdResp):
    """ Response "device_list_interfaces" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    caCmdResp.interfaceType = gvars.IF80211
    caCmdResp.interfaceID = gvars.WLAN_STA_IFACE

    resp.status = gvars.SUCCESS
    return resp


def sta_reset_default(caCmd, caCmdResp):
    """ Response "sta_reset_default" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    gvars.SIGMA_TEST_DATA = dict()
    resp = dut.reset_dut(caCmd.prog)
    gvars.SIGMA_TEST_DATA["prog"] = "any"
    if hasattr(caCmd,"prog"):
        if caCmd.prog != "" and caCmd.prog!= None:
            gvars.SIGMA_TEST_DATA["prog"] = caCmd.prog
    return resp


def sta_get_info(caCmd, caCmdResp):
    """ Response "sta_get_info" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    caCmdResp.interface = gvars.WLAN_STA_IFACE
    caCmdResp.vendor = "Espressif"
    resp = dut.get_card_type()

    if resp.status == SUCCESS:
        caCmdResp.cardType = resp.data

    return resp


def sta_get_mac_address(caCmd, caCmdResp):
    """ Response "sta_get_mac_address" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.get_sta_mac()
    if resp.status == SUCCESS:
        caCmdResp.mac = resp.data
        if gvars.VERBOSE:
            logger.INFO("STA Mac address: '%s'" % caCmdResp.mac)

    return resp

def sta_set_power_save(caCmd, caCmdResp):
    """ Response "sta_set_power_save" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.sta_set_power_save(caCmd.powersave)
    return resp

def sta_btm_query(caCmd,caCmdResp):
    """ Response "sta_btm_query" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent
    """ 
    resp = response.Response() 
    resp = dut.sta_btm_query(caCmd.reason_code,caCmd.list)
    return resp

def sta_loopback_server_start(caCmd,caCmdResp):
    """ Response "sta_loopback_server_start" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent
    """
    resp = response.Response()
    resp = dut.sta_loopback_server_start(caCmd.port)
    return resp


def device_reset(caCmd,caCmdResp):
    resp = response.Response()
    resp = dut.device_reset()
    gvars.SIGMA_TEST_DATA = dict()
    return resp

def sta_loopback_server_stop(caCmd,caCmdResp):
    """ Response "sta_loopback_server_stop" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent
    """
    resp = response.Response()
    resp = dut.sta_loopback_server_stop()
    return resp

    resp.status = gvars.SUCCESS
    return resp

def sta_set_peap(caCmd, caCmdResp):
    """ Response "sta_set_eapttls" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    gvars.SIGMA_TEST_DATA['ssid'] = caCmd.ssid
    gvars.SIGMA_TEST_DATA['username'] = caCmd.username
    gvars.SIGMA_TEST_DATA['trustedRootCA'] = caCmd.trustedRootCA
    gvars.SIGMA_TEST_DATA['encpType'] = caCmd.encpType
    gvars.SIGMA_TEST_DATA['keymgmttype'] = caCmd.KeyMgmtType
    gvars.SIGMA_TEST_DATA['password'] = caCmd.passphrase
    gvars.SIGMA_TEST_DATA['type'] = "eappeap"

    resp.status = gvars.SUCCESS
    return resp

def sta_set_eapttls(caCmd, caCmdResp):
    """ Response "sta_set_eapttls" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    gvars.SIGMA_TEST_DATA['ssid'] = caCmd.ssid
    gvars.SIGMA_TEST_DATA['username'] = caCmd.username
    gvars.SIGMA_TEST_DATA['trustedRootCA'] = caCmd.trustedRootCA
    if caCmd.clientCertificate is not None:
        gvars.SIGMA_TEST_DATA['clientCertificate'] = caCmd.clientCertificate
    gvars.SIGMA_TEST_DATA['encpType'] = caCmd.encpType
    gvars.SIGMA_TEST_DATA['keymgmttype'] = caCmd.KeyMgmtType
    gvars.SIGMA_TEST_DATA['password'] = caCmd.passphrase
    gvars.SIGMA_TEST_DATA['type'] = "eapttls"

    resp.status = gvars.SUCCESS
    return resp

def sta_get_parameter(caCmd, caCmdResp):
    """ Response "sta_get_parameter" command to controller
    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller
    Returns:
       resp: response (status, data) to ca agent
    """
    resp = response.Response()
    if not isExist('ssid'):
        resp.status = FAILURE
        resp.data = "ssid info not present"
        return resp

    resp = dut.sta_get_rssi(gvars.SIGMA_TEST_DATA['ssid'])
    if resp.status == FAILURE:
        return resp

    caCmdResp.rssi = resp.data
    resp.status = SUCCESS
    return resp

def sta_set_encryption(caCmd, caCmdResp):
    """ Response "sta_get_mac_address" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()

    gvars.SIGMA_TEST_DATA['ssid'] = caCmd.ssid
    gvars.SIGMA_TEST_DATA['encpType'] = caCmd.encpType
    if hasattr(caCmd,"key1"):
        gvars.SIGMA_TEST_DATA['passphrase'] = caCmd.key1
    else:
        pass
    resp.status = gvars.SUCCESS
    return resp

    resp.status = gvars.SUCCESS
    return resp

def sta_set_security(caCmd, caCmdResp):
    """ Response "sta_set_security" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    gvars.SIGMA_TEST_DATA['ssid'] = caCmd.ssid
    gvars.SIGMA_TEST_DATA['encpType'] = caCmd.encpType
    gvars.SIGMA_TEST_DATA['keymgmttype'] = caCmd.KeyMgmtType
    gvars.SIGMA_TEST_DATA['type'] = caCmd.type

    if caCmd.type.lower() in  {"eaptls", "eapttls", "eappeap"}:
        gvars.SIGMA_TEST_DATA['username'] = caCmd.username
        gvars.SIGMA_TEST_DATA['password'] = caCmd.password
        gvars.SIGMA_TEST_DATA['trustedRootCA'] = caCmd.trustedRootCA
        gvars.SIGMA_TEST_DATA['CertType'] = caCmd.CertType
        gvars.SIGMA_TEST_DATA['clientCertificate'] = caCmd.clientCertificate
        gvars.SIGMA_TEST_DATA['keymgmtType'] = caCmd.keymgmtType
        gvars.SIGMA_TEST_DATA['PairwiseCipher'] = caCmd.PairwiseCipher
        gvars.SIGMA_TEST_DATA['GroupCipher'] = caCmd.GroupCipher
        gvars.SIGMA_TEST_DATA['GroupMgntCipher'] = caCmd.GroupMgntCipher
        if gvars.SIGMA_TEST_DATA['encpType'] is None:
            gvars.SIGMA_TEST_DATA['encpType'] = caCmd.PairwiseCipher
    else :
        gvars.SIGMA_TEST_DATA['passphrase'] = caCmd.passphrase
    if caCmd.type is not None:
        if caCmd.type.lower()  == "sae":
            gvars.SIGMA_TEST_DATA['pmf'] = "optional"
        if caCmd.type.lower  == "wep":
            gvars.SIGMA_TEST_DATA['passphrase'] = caCmd.key1
        if caCmd.type.lower()  == "open":
            gvars.SIGMA_TEST_DATA['encpType'] = "none"
    if caCmd.keymgmtType =="SuiteB":
        gvars.SIGMA_TEST_DATA['pmf'] = "required"

    resp.status = gvars.SUCCESS
    return resp

def sta_set_psk(caCmd, caCmdResp):
    """ Response "sta_set_psk" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    gvars.SIGMA_TEST_DATA['ssid'] = caCmd.ssid
    gvars.SIGMA_TEST_DATA['passphrase'] = caCmd.passphrase
    gvars.SIGMA_TEST_DATA['encpType'] = caCmd.encpType
    gvars.SIGMA_TEST_DATA['keymgmttype'] = caCmd.keymgmttype
    gvars.SIGMA_TEST_DATA['type'] = "psk"

    resp.status = gvars.SUCCESS
    return resp


def sta_set_ip_config(caCmd, caCmdResp):
    """ Response "sta_set_ip_config" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    if gvars.QT_ENABLED :
        caCmd.dhcp = "0"
        #adding the above line only to make static ip addressing work.
        #QT itself does not send any info on dhcp. To the best of my knowledge, QT does not even check for dhcp value 

    resp = response.Response()
    gvars.SIGMA_TEST_DATA['dhcp'] = int(caCmd.dhcp)

    if not gvars.SIGMA_TEST_DATA['dhcp']:
        gvars.SIGMA_TEST_DATA['sta_iface_ip'] = caCmd.ip
        gvars.SIGMA_TEST_DATA['sta_iface_mask'] = caCmd.mask
        gvars.SIGMA_TEST_DATA['default_gateway'] = caCmd.defaultGateway
        resp = dut.sta_set_ip_config(caCmd.ip, caCmd.mask, caCmd.defaultGateway)
        return resp

    resp.status = gvars.SUCCESS
    return resp

def sta_scan(caCmd, caCmdResp):
    """ Response "sta_scan" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    parameter = None
    resp = response.Response()
    if  hasattr(caCmd,"parameter"):
        if caCmd.parameter == "SSID_BSSID":
            parameter = caCmd.parameter

    resp = dut.sta_scan(parameter)
    
    if resp.status == SUCCESS:
        caCmdResp.scan = resp.data
    return resp

def dev_send_frame(caCmd, caCmdResp):
    """ Response "dev_send_frame" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    if isExist('non_pref_chan_list'):
        dut.dev_send_frame(caCmd.program, gvars.SIGMA_TEST_DATA['non_pref_chan_list'])

    resp.status = SUCCESS
    return resp

def dev_exec_action(caCmd, caCmdResp):
    resp = response.Response()
    resp.status = SUCCESS
    if caCmd.ServerCertTrust == "Accept":
        caCmdResp.ServerCertTrustResult = "OverrideNotAllowed, uosc_allowed_on_sta 0"
    else:
        caCmdResp.ServerCertTrustResult = "Reject"
    return resp


def sta_set_11n(caCmd, caCmdResp):
    """ Response "sta_set_11n" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()

    resp.status = SUCCESS
    return resp

def sta_set_systime(caCmd, caCmdResp):
    t = datetime(caCmd.year, caCmd.month, caCmd.day, caCmd.hours, caCmd.minutes, caCmd.seconds)
    timestamp = int(datetime.timestamp(t))
    resp = dut.set_sys_time(timestamp)
    return resp

def sta_set_rfeature(caCmd, caCmdResp):
    """ Response "sta_set_rfeature" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()

    if hasattr(caCmd,"prog"):
        if caCmd.prog.lower() == "mbo":
            if str(caCmd.ch_pref).lower() == 'clear':
                data = []
                resp = dut.dev_send_frame("mbo", data)
                gvars.SIGMA_TEST_DATA['non_pref_chan_list'] = []
            else:
                if not isExist('non_pref_chan_list'):
                    gvars.SIGMA_TEST_DATA['non_pref_chan_list'] = list()

                non_pref_chan = dict()
                non_pref_chan['ch_pref_num'] = caCmd.ch_pref_num
                non_pref_chan['ch_pref'] = caCmd.ch_pref
                non_pref_chan['ch_op_class'] = caCmd.ch_op_class
                non_pref_chan['ch_reason_code'] = caCmd.ch_reason_code
                gvars.SIGMA_TEST_DATA['non_pref_chan_list'].append(non_pref_chan)


    if caCmd.gi:
        resp = dut.sta_set_gi(caCmd.gi)
        if resp.status == FAILURE:
            return resp

    if caCmd.ltf:
        resp = dut.sta_set_ltf(caCmd.ltf)
        if resp.status == FAILURE:
            return resp

    if caCmd.transmitOMI is not None:
        if caCmd.OMCtrl_TxNSTS is None:
            caCmd.OMCtrl_TxNSTS = 0
        if caCmd.OMCtrl_ChnlWidth is None:
            caCmd.OMCtrl_ChnlWidth = 0
        if caCmd.OMCtrl_ULMUDisable is None:
            caCmd.OMCtrl_ULMUDisable = 0
        if caCmd.OMCtrl_ULMUDataDisable is None:
            caCmd.OMCtrl_ULMUDataDisable = 0
        if caCmd.device_num is None:
            caCmd.device_num = 0
        resp = dut.sta_set_om_control(caCmd.transmitOMI,caCmd.OMCtrl_TxNSTS,caCmd.OMCtrl_ChnlWidth,caCmd.OMCtrl_ULMUDisable,caCmd.OMCtrl_ULMUDataDisable, caCmd.device_num)
        if resp.status == FAILURE:
            return resp
    #iTWT
    if caCmd.TWT_Setup == "request":
        resp = dut.sta_setup_itwt(caCmd.NDPPagingInd,caCmd.RespPMMode,caCmd.NegotiationType,caCmd.TWT_Setup,caCmd.SetupCommand,caCmd.TWT_Trigger,caCmd.implicit,caCmd.FlowType,
        caCmd.WakeIntervalExp,caCmd.Protection,caCmd.NominalMinWakeDur,caCmd.WakeIntervalMantissa,caCmd.TWT_Channel)
        if resp.status == FAILURE:
            return resp
    if caCmd.TWT_Setup == "teardown":
        resp = dut.sta_teardown_itwt(caCmd.TWT_Setup,caCmd.FlowID,caCmd.NegotiationType)
        if resp.status == FAILURE:
            return resp
    if caCmd.TWT_Operation == "suspend":
        resp = dut.sta_suspend_itwt()
        if resp.status == FAILURE:
            return resp
    if caCmd.TWT_Operation == "resume":
        resp = dut.sta_resume_itwt()
        if resp.status == FAILURE:
            return resp

    if caCmd.BF_Type is not None:
        if caCmd.RuStart is None:
            caCmd.RuStart = 0
        if caCmd.RuEnd is None:
            caCmd.RuEnd = 8
        resp = dut.sta_set_beamform_type(caCmd.BF_Type, caCmd.RuStart, caCmd.RuEnd)

    if caCmd.PPDUTxType is not None:
        resp = dut.sta_set_ppdu_tx_type(caCmd.PPDUTxType, caCmd.RUAllocTone)

    resp.status = SUCCESS
    return resp

def sta_set_pwrsave(caCmd, caCmdResp):
    """ Response "sta_set_pwrsave" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.sta_set_pwrsave(caCmd.mode)
    return resp
    
def sta_associate(caCmd, caCmdResp):
    """ Response "sta_associate" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()

    if isExist("prog"):
        if isExist('non_pref_chan_list'):
            dut.dev_send_frame(gvars.SIGMA_TEST_DATA["prog"], gvars.SIGMA_TEST_DATA['non_pref_chan_list'])

    roaming = None
    authmode = None
    bssid = None
    interface = None
    password = None
    username = None
    pmf = None
    type = None
    if isExist('roaming'):
        roaming = gvars.SIGMA_TEST_DATA['roaming']

    if isExist('authmode'):
        authmode = gvars.SIGMA_TEST_DATA['authmode']
    
    gvars.SIGMA_TEST_DATA['ssid'] = caCmd.ssid
    if hasattr(caCmd,"bssid"):
        bssid = caCmd.bssid
    
    if isExist('pmf'):
        pmf = gvars.SIGMA_TEST_DATA['pmf']
    else:
        pmf = "optional"
    if isExist("username"):
        username = gvars.SIGMA_TEST_DATA["username"]

    if isExist("bssid"):
        bssid = gvars.SIGMA_TEST_DATA["bssid"]

    if isExist("type"):
        type=gvars.SIGMA_TEST_DATA['type']
    else:
        type="any"
    connect_attempt = 0

    """
    There might be a difference in how the test bed sends the password/passphrase matching a username when connecting to Enterprise connections.
    Hence, assuming that either there will be a network passphrase(personal connections) or a password(enterprise connections)
    """
    if isExist("type"):
            if gvars.SIGMA_TEST_DATA['type'].lower() in {"eaptls","eapttls","eappeap"}:
                if isExist("passphrase"):
                    password = gvars.SIGMA_TEST_DATA['passphrase']
                elif isExist("password"):
                    password=gvars.SIGMA_TEST_DATA['password']
                resp = dut.sta_associate(caCmd.ssid, config=gvars.SIGMA_TEST_DATA, passphrase=password, bssid=bssid, pmf="optional", type=gvars.SIGMA_TEST_DATA['type'], username=username,roaming=roaming,authmode=authmode)
                connect_attempt = 1

    if isExist('encpType') and connect_attempt == 0:
            if gvars.SIGMA_TEST_DATA['encpType'] == None:
                if isExist('passphrase'):
                    resp = dut.sta_associate(caCmd.ssid,config=gvars.SIGMA_TEST_DATA, passphrase=gvars.SIGMA_TEST_DATA['passphrase'], bssid=bssid,roaming=roaming,authmode=authmode,type=type)
                else:
                    resp = dut.sta_associate(caCmd.ssid,config=gvars.SIGMA_TEST_DATA, bssid=bssid,roaming=roaming,authmode=authmode,type=type)
            else:
                ##resp = dut.sta_associate(caCmd.ssid,config=gvars.SIGMA_TEST_DATA, passphrase=gvars.SIGMA_TEST_DATA['passphrase'], bssid=bssid, pmf = pmf,roaming=roaming,authmode=authmode)
                if isExist('passphrase'):
                    resp = dut.sta_associate(caCmd.ssid,config=gvars.SIGMA_TEST_DATA, passphrase=gvars.SIGMA_TEST_DATA['passphrase'], bssid=bssid,roaming=roaming,authmode=authmode,type=type)
                else:
                    resp = dut.sta_associate(caCmd.ssid,config=gvars.SIGMA_TEST_DATA, bssid=bssid,roaming=roaming,authmode=authmode,type=type)
    if resp.status == FAILURE:
        return resp
        
    return resp

def sta_set_eaptls(caCmd, caCmdResp):
    """ Response "sta_set_eaptls" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    gvars.SIGMA_TEST_DATA['ssid'] = caCmd.ssid
    gvars.SIGMA_TEST_DATA['username'] = caCmd.username
    gvars.SIGMA_TEST_DATA['trustedRootCA'] = caCmd.trustedRootCA
    gvars.SIGMA_TEST_DATA['clientCertificate'] = caCmd.clientCertificate
    gvars.SIGMA_TEST_DATA['encpType'] = caCmd.encpType
    gvars.SIGMA_TEST_DATA['keymgmttype'] = caCmd.KeyMgmtType
    if caCmd.passphrase != None:
        gvars.SIGMA_TEST_DATA['passphrase'] = caCmd.passphrase
    gvars.SIGMA_TEST_DATA['type'] = "eaptls"

    resp.status = gvars.SUCCESS
    return resp
    
def sta_reassoc(caCmd, caCmdResp):
    """ Response "sta_reassoc" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    bssid = None
    passphrase = None
    resp = response.Response()
    if isExist("bssid"):
        bssid = gvars.SIGMA_TEST_DATA["bssid"]

    if not isExist('ssid'):
        resp.status = FAILURE
        resp.data = "ssid info not present"
        return resp
    if not isExist('encpType'):
        resp.status = FAILURE
        resp.data = "EncpType not present"
        return resp
    if isExist("passphrase"):
        password = gvars.SIGMA_TEST_DATA['passphrase']
    elif isExist("password"):
        password=gvars.SIGMA_TEST_DATA['password']

    if isExist("encpType"):
        if gvars.SIGMA_TEST_DATA['encpType'].lower() == "none":
            resp = dut.sta_associate(ssid = gvars.SIGMA_TEST_DATA["ssid"],config=gvars.SIGMA_TEST_DATA,passphrase=None,bssid=bssid)
        else:
            if isExist('pmf'):
                resp = dut.sta_associate(ssid = gvars.SIGMA_TEST_DATA["ssid"],config=gvars.SIGMA_TEST_DATA, passphrase=password, pmf=gvars.SIGMA_TEST_DATA['pmf'],bssid=bssid)
            else:
                resp = dut.sta_associate(ssid = gvars.SIGMA_TEST_DATA["ssid"],config=gvars.SIGMA_TEST_DATA,  passphrase=password,bssid=bssid)
        if resp.status == FAILURE:
            return resp

    resp.status = SUCCESS
    return resp

def sta_disconnect(caCmd, caCmdResp):
    """ Response "sta_disconnect" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.sta_disconnect()
    return resp


def sta_is_connected(caCmd, caCmdResp):
    """ Response "sta_is_connected" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.sta_is_connected(gvars.SIGMA_TEST_DATA['ssid'])
    if resp.status == SUCCESS:
        caCmdResp.connected = resp.data
        if gvars.VERBOSE:
            logger.INFO("STA wlan connected status - ssid: '%s', conn: '%s'" % (gvars.SIGMA_TEST_DATA['ssid'], resp.data))

    return resp


def sta_get_bssid(caCmd, caCmdResp):
    """ Response "sta_get_bssid" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.sta_get_bssid(gvars.SIGMA_TEST_DATA['ssid'])
    if resp.status == SUCCESS:
        caCmdResp.bssid = resp.data
        if gvars.VERBOSE:
            logger.INFO(
                "STA wlan connected with ssid: '%s', bssid: '%s'" % (gvars.SIGMA_TEST_DATA['ssid'], resp.data))

    return resp


def sta_get_ip_config(caCmd, caCmdResp):
    """ Response "sta_get_ip_config" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.sta_get_ip_config()
    if resp.status == SUCCESS:
        caCmdResp.dhcp = gvars.SIGMA_TEST_DATA['dhcp']
        caCmdResp.ip = resp.data['ip']
        caCmdResp.mask = resp.data['mask']
        if gvars.VERBOSE:
            logger.INFO("STA ip address: '%s'" % caCmdResp.ip)
            logger.INFO("STA netmask: '%s'" % caCmdResp.mask)

    return resp


def traffic_agent_reset(caCmd, caCmdResp):
    """ Response "traffic_agent_reset" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    gvars.SIGMA_TEST_DATA['stream_id'] = 1
    resp = dut.sigma_traffic_reset()
    return resp


def traffic_send_ping(caCmd, caCmdResp):
    """ Response "traffic_send_ping" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.traffic_send_ping(caCmd.destination, caCmd.framesize, caCmd.frameRate, caCmd.duration)
    if resp.status == SUCCESS:
        if not isExist('stream_id'):
            gvars.SIGMA_TEST_DATA['stream_id'] =1
        caCmdResp.streamID = gvars.SIGMA_TEST_DATA['stream_id']
        gvars.SIGMA_TEST_DATA['stream_id'] +=1

    return resp


def traffic_stop_ping(caCmd, caCmdResp):
    """ Response "traffic_stop_ping" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = dut.traffic_stop_ping(caCmd.streamID)
    if resp.status == SUCCESS:
        caCmdResp.sent = resp.data['sent']
        caCmdResp.replies = resp.data['replies']

    return resp


def sta_preset_testparameters(caCmd, caCmdResp):
    """ Response "sta_preset_testparameters" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()

    if caCmd.program and caCmd.program.lower() == "mbo":
        if not isExist('non_pref_chan_list'):
            gvars.SIGMA_TEST_DATA['non_pref_chan_list'] = list()

        non_pref_chan = dict()
        non_pref_chan['ch_pref_num'] = caCmd.ch_pref_num
        non_pref_chan['ch_pref'] = caCmd.ch_pref
        non_pref_chan['ch_op_class'] = caCmd.ch_op_class
        non_pref_chan['ch_reason_code'] = caCmd.ch_reason_code
        gvars.SIGMA_TEST_DATA['non_pref_chan_list'].append(non_pref_chan)

    gvars.SIGMA_TEST_DATA['roaming'] = caCmd.roaming

    resp.status = SUCCESS
    return resp
def sta_send_addba(caCmd, caCmdResp):
    resp = dut.addba_set(caCmd.tid)
    return resp

def sta_set_wireless(caCmd, caCmdResp):
    """ Response "sta_set_wireless" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    resp = response.Response()
    if caCmd.txsp_stream or caCmd.rxsp_stream:
       if caCmd.txsp_stream is not None:
           nss = caCmd.txsp_stream
       if caCmd.rxsp_stream is not None:
           nss = caCmd.rxsp_stream

       resp = dut.sta_set_nss(nss)
       if resp.status == FAILURE:
           return resp

    if caCmd.mcs:
       resp = dut.sta_set_fixed_rate(caCmd.mcs)
       if resp.status == FAILURE:
           return resp

    if caCmd.bcc:
       resp = dut.sta_set_bcc(caCmd.bcc)
       if resp.status == FAILURE:
           return resp

    if caCmd.amsdu:
        resp = dut.sta_set_amsdu(caCmd.amsdu)
        if resp.status == FAILURE:
            return resp

    if caCmd.width:
        resp = dut.sta_set_width(caCmd.width)
        if resp.status == FAILURE:
            return resp

    resp.status = SUCCESS
    return resp

def traffic_agent_config(caCmd, caCmdResp):
    """ Response "traffic_agent_config" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

  """
    config = ""
    gvars.stream_count +=1
    if caCmd.profile:
        config += "--profile=%s " % (caCmd.profile)
    if caCmd.direction:
        config += "--direction=%s " % (caCmd.direction)
    if caCmd.source:
        config += "--source=%s " % (caCmd.source)
    if caCmd.destination:
        config += "--destination=%s " % (caCmd.destination)
    if caCmd.destinationPort:
        config += "--destinationPort=%s " % (caCmd.destinationPort)
    if caCmd.sourcePort:
        config += "--sourcePort=%s " % (caCmd.sourcePort)
    if caCmd.frameRate:
        config += "--frameRate=%s " % (caCmd.frameRate)
    if caCmd.duration:
        if not isExist('last_traffic_ssid') or gvars.SIGMA_TEST_DATA['last_traffic_ssid'] != gvars.SIGMA_TEST_DATA['ssid']:
            gvars.SIGMA_TEST_DATA['last_traffic_ssid'] = gvars.SIGMA_TEST_DATA['ssid']
            gvars.SIGMA_TEST_DATA['traffic_stream_id'] = 1 #init
        else:
            gvars.SIGMA_TEST_DATA['traffic_stream_id'] = gvars.SIGMA_TEST_DATA['traffic_stream_id'] + 1

        if gvars.SIGMA_TEST_DATA['ssid'] == "HE-5.64.1_24G":
            if gvars.SIGMA_TEST_DATA['traffic_stream_id'] in [1, 2, 3]:
                config += "--duration=%s " % str(int(caCmd.duration)-30)
                gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)-30
            elif gvars.SIGMA_TEST_DATA['traffic_stream_id'] in [7, 8, 9]:
                config += "--duration=%s " % str(int(caCmd.duration)+100)
                gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)+100
            elif gvars.SIGMA_TEST_DATA['traffic_stream_id'] in [4, 5, 6]:
                config += "--duration=%s " % str(int(caCmd.duration)-40)
                gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)-40
            else:
                config += "--duration=%s " % str(int(caCmd.duration)-40)
                gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)-40

        if gvars.SIGMA_TEST_DATA['ssid'] == "HE-5.61.1_24G":
            # streamID 3: mmsf=0
            # streamID 4: mmsf=3
            # case issue, in 106-tone, mmsf cant differ throughput of streamID 3 and streamID 4.
            if gvars.SIGMA_TEST_DATA['traffic_stream_id'] == 3:
                config += "--duration=%s " % str(int(caCmd.duration)+0) #20
                gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)+0 #20
            else:
                config += "--duration=%s " % str(int(caCmd.duration))
                gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)
        elif gvars.SIGMA_TEST_DATA['ssid'] == "HE-5.5.1_24G" or gvars.SIGMA_TEST_DATA['ssid'] == "HE-5.5.2_24G" or gvars.SIGMA_TEST_DATA['ssid'] == "HE-5.19.1_24G":
            config += "--duration=%s " % str(int(caCmd.duration)+100)
            gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)+100
        elif gvars.SIGMA_TEST_DATA['ssid'] == "HE-5.6.1_24G":
            config += "--duration=%s " % str(int(caCmd.duration)+30)
            gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)+30
        elif gvars.SIGMA_TEST_DATA['ssid'] == "HE-5.45.1":
            config += "--duration=%s " % str(int(caCmd.duration)+20)
            gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)+20
        elif gvars.SIGMA_TEST_DATA['ssid'] == "5.2.47":
            # 11n, 5.2.47 tx throughput stream 1
            config += "--duration=%s " % str(int(caCmd.duration)+30)
            caCmd.duration = str(int(caCmd.duration) + 30)
            gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)+30
        elif gvars.SIGMA_TEST_DATA['ssid'] == "FGR896lk?":
            print("QweQwe : Stream Count {}".format(gvars.stream_count))
            if gvars.stream_count % 5 == 4:
                config += "--duration=%s " % str(int(caCmd.duration)+7)
                caCmd.duration = str(int(caCmd.duration) + 7)
            else:
                config += "--duration=%s " % str(int(caCmd.duration))
                gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)
        else:
            config += "--duration=%s " % str(int(caCmd.duration))
            gvars.SIGMA_TEST_DATA['traffic_duration'] = int(caCmd.duration)

    if caCmd.payloadSize:
        config += "--payloadSize=%s " % (caCmd.payloadSize)
    if caCmd.startDelay:
        config += "--startDelay=%s " % (caCmd.startDelay)
    if caCmd.trafficClass:
        config += "--trafficClass=%s " % (caCmd.trafficClass)
    if caCmd.startDelay:
        config += "--startDelay=%s " % (caCmd.startDelay)
    if caCmd.HTI:
        config += "--HTI=%s " % (caCmd.HTI)
    if caCmd.maxCnt:
        config += "--maxCnt=%s " % (caCmd.maxCnt)
    resp = dut.sigma_traffic_config(config)

    if resp.status == SUCCESS:
        caCmdResp.streamID = resp.data
        gvars.SIGMA_TEST_DATA["data_streams"][caCmdResp.streamID] = caCmd
        print("{}:{}".format(caCmdResp.streamID,caCmd.duration))
    return resp


def traffic_agent_send(caCmd, caCmdResp):
    """ Response "traffic_agent_send" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    streamID = caCmd.streamID
    streamList = streamID.split(" ")
    resp = dut.sigma_traffic_send(streamList)
    if resp.status == SUCCESS:
        caCmdResp.streamID = resp.data['streamID']
        caCmdResp.txFrames = resp.data['txFrames']
        caCmdResp.rxFrames = resp.data['rxFrames']
        caCmdResp.txPayloadBytes = resp.data['txPayloadBytes']
        caCmdResp.rxPayloadBytes = resp.data['rxPayloadBytes']
        caCmdResp.outOfSequenceFrames = resp.data['outOfSequenceFrames']

    return resp


def traffic_agent_receive_start(caCmd, caCmdResp):
    """ Response "traffic_agent_receive_start" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    streamID = caCmd.streamID
    streamList = streamID.split(" ")
    resp = dut.sigma_traffic_receive_start(streamList)
    return resp


def traffic_agent_receive_stop(caCmd, caCmdResp):
    """ Response "traffic_agent_receive_stop" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent

    """
    streamID = caCmd.streamID
    streamList = streamID.split(" ")
    resp = dut.sigma_traffic_receive_stop(streamList)
    if resp.status == SUCCESS:
        caCmdResp.streamID = resp.data['streamID']
        caCmdResp.txFrames = resp.data['txFrames']
        caCmdResp.rxFrames = resp.data['rxFrames']
        caCmdResp.txPayloadBytes = resp.data['txPayloadBytes']
        caCmdResp.rxPayloadBytes = resp.data['rxPayloadBytes']
        caCmdResp.outOfSequenceFrames = resp.data['outOfSequenceFrames']

    return resp

def sta_btm_query(caCmd,caCmdResp):
    """ Response "sta_btm_query" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent
    """ 
    resp = response.Response() 
    resp = dut.sta_btm_query(caCmd.reason_code,caCmd.list)
    return resp

def sta_loopback_server_start(caCmd,caCmdResp):
    """ Response "sta_loopback_server_start" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent
    """
    resp = response.Response()
    resp = dut.sta_loopback_server_start(caCmd.port)
    return resp


def device_reset(caCmd,caCmdResp):
    resp = response.Response()
    resp = dut.device_reset()
    gvars.SIGMA_TEST_DATA = dict()
    return resp

def sta_loopback_server_stop(caCmd,caCmdResp):
    """ Response "sta_loopback_server_stop" command to controller

    Parameters:
       caCmd --- Command params from controller
       caCmdResp --- Command response params to controller

    Returns:
       resp: response (status, data) to ca agent
    """
    resp = response.Response()
    resp = dut.sta_loopback_server_stop()
    return resp

def sta_get_events(caCmd, caCmdResp):
    resp = response.Response()
    dutID = caCmd.ID
    if caCmd.parameter == "retry":
        resp = dut.sta_get_retry(dutID)
        caCmdResp.txretry = resp.data['txretry']
        caCmdResp.rxretry = resp.data['rxretry']
    resp.status = SUCCESS
    return resp
