from hmac import compare_digest
from xml.dom.pulldom import default_bufsize
from common import gvars, logger
import socket
import struct

def cidr_to_netmask(cidr):
    network, net_bits = cidr.split('/')
    host_bits = 32 - int(net_bits)
    netmask = socket.inet_ntoa(struct.pack('!I', (1 << 32) - (1 << host_bits)))
    return network, netmask

def generate_config(qt_cmd):
    default_wep_key = ""
    ssid = ""
    for cfg_item in qt_cmd[1:]:
        if "ssid=" in cfg_item:
            gvars.SIGMA_TEST_DATA['ssid'] = cfg_item.split("=")[1]
            ssid = cfg_item.split("=")[1]
        elif "wep_key0=" in cfg_item:
            gvars.SIGMA_TEST_DATA['passphrase'] = cfg_item.split("=")[1]
        elif "wep_default_key=" in cfg_item:
            default_wep_key = cfg_item.split("=")[1]
        elif "eap=" in cfg_item.lower():
            if "ttls" in cfg_item.lower():
                gvars.SIGMA_TEST_DATA['type'] = "eapttls"
            elif "tls" in cfg_item.lower():
                gvars.SIGMA_TEST_DATA['type'] = "eaptls"
            elif "peap" in cfg_item.lower():
                gvars.SIGMA_TEST_DATA['type'] = "eappeap"
        elif "pmf=" in cfg_item:
            if cfg_item.split("=")[1] == "1":
                gvars.SIGMA_TEST_DATA['pmf'] = "optional"
        elif "ieee80211w=" in cfg_item.lower():
            if cfg_item.split("=")[1] == "1":
                gvars.SIGMA_TEST_DATA['pmf'] = "optional"
            elif cfg_item.split("=")[1] == "2":
                gvars.SIGMA_TEST_DATA['pmf'] = "required"
        elif "psk=" in cfg_item:
            gvars.SIGMA_TEST_DATA["passphrase"] = cfg_item.split("=")[1]
        elif "key_mgmt=" in cfg_item.lower():
            gvars.SIGMA_TEST_DATA['encpType'] = cfg_item.split("=")[1]
            if cfg_item.split("=")[1] == "SAE":
                gvars.SIGMA_TEST_DATA['authmode'] = "wpa3"
            elif cfg_item.split("=")[1] == "WPA-PSK":
                gvars.SIGMA_TEST_DATA['authmode'] = "wpa2"
            #TODO : Fix this whenever QT adds support for 192 Bit ser
            elif cfg_item.split("=")[1] == "SuiteB":
                gvars.SIGMA_TEST_DATA['authmode'] = "wpa2-eap"
                gvars.SIGMA_TEST_DATA['pmf'] = "required"
                gvars.SIGMA_TEST_DATA['keymgmtType'] = 'SuiteB'
        elif "username=" in cfg_item.lower():
            gvars.SIGMA_TEST_DATA['username'] = cfg_item.split("=")[1]
        elif "password=" in cfg_item.lower():
            gvars.SIGMA_TEST_DATA['password'] = cfg_item.split("=")[1]
        elif "identity=" in cfg_item.lower():
            gvars.SIGMA_TEST_DATA['identity'] = cfg_item.split("=")[1]
        elif "ca_cert=" in cfg_item.lower():
            gvars.SIGMA_TEST_DATA["trustedRootCA"] = cfg_item.split("=")[1]
        elif "client_cert=" in cfg_item.lower():
            gvars.SIGMA_TEST_DATA["clientCertificate"] = cfg_item.split("=")[1]
        elif "pairwise=" in cfg_item.lower():
            if "TKIP" in cfg_item.split("=")[1]:
                gvars.SIGMA_TEST_DATA['authmode'] = "wpa"

    if "type" in gvars.SIGMA_TEST_DATA.keys():
        if gvars.SIGMA_TEST_DATA["type"] in ["eappeap","eapttls"]:
            if "username" not in gvars.SIGMA_TEST_DATA.keys():
                if "identity" in gvars.SIGMA_TEST_DATA.keys():
                    gvars.SIGMA_TEST_DATA['username'] = gvars.SIGMA_TEST_DATA["identity"]

    if "encpType" not in gvars.SIGMA_TEST_DATA.keys():
        gvars.SIGMA_TEST_DATA['encpType'] = "none"

    if "passphrase" not in gvars.SIGMA_TEST_DATA.keys() and default_wep_key != "":
        gvars.SIGMA_TEST_DATA['passphrase'] = default_wep_key
    return ssid

def sta_get_ip(qt_cmd):
    return "sta_get_ip_config"


def sta_get_mac(qt_cmd):
    return "sta_get_mac_address"


def sta_set_static_ip(qt_cmd):
    ip, mask = cidr_to_netmask(qt_cmd[1])
    return "sta_set_ip_config,ip,{},mask,{}".format(ip, mask)


def sta_associate(qt_cmd):
    ssid = generate_config(qt_cmd) 
    return "sta_associate,ssid,{}".format(ssid)

def sta_disassociate(qt_cmd):
    return "sta_disconnect"

def sta_reconnect(qt_cmd):
    return "sta_reassoc,ssid,{}".format(gvars.SIGMA_TEST_DATA["ssid"])


def sta_set_parameter(qt_cmd):
    pmf = qt_cmd[1].split("=")[1]
    if gvars.SIGMA_TEST_DATA['pmf'] == "optional" and pmf == "2":
        gvars.SIGMA_TEST_DATA['pmf'] = "required"
    elif gvars.SIGMA_TEST_DATA['pmf'] == "required" and pmf == "1":
        gvars.SIGMA_TEST_DATA['pmf'] = "optional"
    
    #TODO : Figure out if we should call reconnect here.. disconnect and connect here
    return "sta_reassoc,ssid,{}".format(gvars.SIGMA_TEST_DATA["ssid"])

def sta_btm_query(qt_cmd):
    reason_code = -1
    list = 0
    for param in qt_cmd:
        param_key_value = param.split("=")
        if param_key_value[0] == "reason_code":
            reason_code = param_key_value[1]
        elif param_key_value == "list":
            list = 1
    
    return "sta_btm_query,reason_code,{},candidate_list,{}".format(reason_code,list)

def sta_loopback_server_start(qt_cmd):
    port = int(qt_cmd[1].split("=")[1])
    return "sta_loopback_server_start,port,{}".format(port)

def device_reset(qt_cmd):
    return "device_reset"

def sta_loopback_server_stop(qt_cmd):
    return "sta_loopback_server_stop"

def sta_set_channel_width(qt_cmd):
    #TODO Figure out Sigma Equivalent
    logger.INFO("Not supported yet")

def sta_enable_power_save(qt_cmd):
    #TODO Figure out what data QT actually sends to us for this. For now just enabling it
    return "sta_set_power_save,mode,on"




