from common import gvars, logger
from capi import capi
from capi import capi_cmd, capi_cmd_resp, capi_requests
from quicktrack import qt_capi_translate,capi_qt_translate


QT_CMDS_LIST = [
    "sta_get_ip",
    "sta_get_mac",
    "sta_set_static_ip",
    "sta_associate",
    "sta_disassociate",
    "sta_reconnect",
    "sta_set_parameter",
    "sta_btm_query",
    "sta_set_channel_width",
    "sta_enable_power_save",
    "sta_loopback_server_start",
    "sta_loopback_server_stop",
    "device_reset"]



CAPI_CMDS_LIST = ["ca_get_version",
                  "traffic_send_ping",
                  "traffic_stop_ping",
                  "traffic_agent_config",
                  "traffic_send_ping",
                  "traffic_agent_send",
                  "traffic_agent_reset",
                  "traffic_agent_receive_start",
                  "traffic_agent_receive_stop",
                  "sta_get_ip_config",
                  "sta_set_ip_config",
                  "sta_get_mac_address",
                  "sta_set_mac_address",
                  "sta_is_connected",
                  "sta_verify_ip_connection",
                  "sta_get_bssid",
                  "sta_get_stats",
                  "sta_set_encryption",
                  "sta_set_psk",
                  "sta_set_eaptls",
                  "sta_set_eapttls",
                  "sta_set_eapsim",
                  "sta_set_peap",
                  "sta_set_ibss",
                  "sta_associate",
                  "device_list_interfaces",
                  "device_get_info",
                  "sta_get_info",
                  "sta_set_mode",
                  "sta_up_load",
                  "sta_debug_set",
                  "sta_set_uapsd",
                  "sta_set_wmm",
                  "sta_disconnect",
                  "sta_reassociate",
                  "sta_reassoc",
                  "dev_send_frame",
                  "dev_exec_action",
                  "sta_set_security",
                  "sta_preset_testparameters",
                  "sta_set_eapfast",
                  "sta_set_eapaka",
                  "sta_set_eapakaprime",
                  "sta_set_systime",
                  "sta_set_power_save",
                  "sta_reset_default",
                  "sta_set_11n",
                  "sta_set_wireless",
                  "sta_send_addba",
                  "sta_set_rifs_test",
                  "sta_send_coexist_mgmt",
                  "sta_get_p2p_dev_address",
                  "sta_set_p2p",
                  "sta_p2p_connect",
                  "sta_start_autonomous_go",
                  "sta_p2p_start_group_formation",
                  "sta_p2p_dissolve",
                  "sta_send_p2p_invitation_req",
                  "sta_accept_p2p_invitation_req",
                  "sta_send_p2p_provision_dis_req",
                  "sta_set_wps_pbc",
                  "sta_wps_read_pin",
                  "sta_wps_enter_pin",
                  "sta_get_psk",
                  "sta_p2p_reset",
                  "sta_wps_read_label",
                  "sta_get_p2p_ip_config",
                  "sta_send_service_discovery_req",
                  "sta_send_p2p_presence_req",
                  "sta_set_sleep",
                  "sta_scan",
                  "sta_set_opportunistic_ps",
                  "sta_add_arp_table_entry",
                  "sta_block_icmp_response",
                  "sta_set_radio",
                  "sta_set_rfeature",
                  "start_wfd_connection",
                  "wfa_cli_cmd",
                  "connect_go_start_wfd",
                  "sta_generate_event",
                  "reinvoke_wfd_session",
                  "sta_get_parameter",
                  "sta_exec_action",
                  "sta_get_events",
                  "sta_btm_query",
                  "sta_loopback_server_start",
                  "sta_loopback_server_stop",
                  "device_reset"]

def translate_qt_call(qt_call):
    qt_cmd = qt_call.split(",")
    capi_cmd = getattr(qt_capi_translate,qt_cmd[0],gvars.INVALID_QT_COMMAND)(qt_cmd)
    if capi_cmd == gvars.INVALID_QT_COMMAND:
        logger.ERROR(gvars.invalid_qt_command(capi_cmd))
    return capi_cmd

def translate_capi_resp(capi_resp,qt_call):
    qt_resp = getattr(capi_qt_translate,qt_call,gvars.INVALID_CAPI_COMMAND)(capi_resp)
    if qt_resp == gvars.INVALID_CAPI_COMMAND:
        logger.ERROR(gvars.invalid_capi_command(qt_resp))
    return qt_resp    
   
def process_capi_message(message):
    """ Process UCC CAPI message

    Parameters:
        CAPI commands (str): CAPI comma seperated message

    Returns:
        Config dict (dict): Return Configuration dict

    """
    api_config_data = dict()
    data = message.split(",")[1:]
    if len(data) > 0:
        index = 0
        while index < len(data):
            api_config_data[data[index]] = data[index + 1]
            index += 2

    return api_config_data




def process_qt_request(qt_call):
    """ Process UCC CAPI command requests

    Parameters:
        CAPI commands (str): CAPI command string which config Info

    Returns:
        CAPI status (str): Return CAPI status string

    """
    #capi_call = "Ignore"
    capi_call = translate_qt_call(qt_call)
    capi_call = translate_qt_call(qt_call)
    #logger.CAPI(qt_call)
    capi_api_info = capi_call.split(",")
    api_name = capi_api_info[0].strip()
    
    logger.INFO("Basic Info : QT_Input : {}\n  CAPI_Equivalent : {}".format(qt_call,capi_call))
    if api_name not in CAPI_CMDS_LIST:
        logger.ERROR("Not a valid CAPI command !")
        return gvars.CAPI_RESP_INVALID
  
    configs = process_capi_message(capi_call)

    caCmd = getattr(capi_cmd.ca_capi_cmd, api_name)(configs)
    caCmdResp = getattr(capi_cmd_resp.ca_capi_cmd_resp, api_name)()
    capi_resp = capi_requests.process_ca_api_request(api_name, caCmd, caCmdResp)

    qt_resp = translate_capi_resp(capi_resp,qt_call.split(",")[0])
    return qt_resp