/* CAPI commands

   This example code is in the Public Domain (or CC0 licensed, at your option.)

   Unless required by applicable law or agreed to in writing, this
   software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, either express or implied.
*/

#include <string.h>
#include "esp_log.h"
#include "esp_console.h"
#include "argtable3/argtable3.h"

/* P118 CAPI Spec */
typedef struct {
    struct arg_str *abftl_range;            //string: "Gt1"- >=1, specify the min value of the range.(does not specify the max value of the range).
    struct arg_str *addba_reject;           //string: "Enable", "Disable"
    struct arg_str *addba_req_bufsize;      //string
    struct arg_str *addba_rsp_bufsize;      //string
    struct arg_int *alloc_id;               //integer: 1-n
    struct arg_str *alloc_type;             //string: "CBAP"
    struct arg_str *ampdu;                  //string: "Enable", "Disable"
    struct arg_str *amsdu;                  //string: "Enable", "Disable"
    struct arg_int *ba_param_amsdu_support; //integer: 0-not set, 1-set
    struct arg_int *ba_recv_status;         //integer: max number of MSDUs/A-MSDUs that can be acknowledged for BA frame.

    struct arg_int *ba_recv_buf;            //integer: 2 or 4
    struct arg_str *band;                   //string: "2.4" or "5"
    struct arg_str *bcc;                    //string: "Enable", "Disable"
    struct arg_int *bcn_interval;           //integer
    struct arg_int *bfmee_sts;              //integer: only support 1
    struct arg_str *bsrp_bqrp_ampdu_aggr;   //string: "Enable", "Disable"
    struct arg_str *bw_signaling;           //string: "Enable", "Disable"
    struct arg_int *cbap_only;              //boolean: 1, 0
    struct arg_int *channel;                //integer
    struct arg_str *cts_bandwidth;          //string: "20", "40", "80

    struct arg_str *dev_role;               //string: "STA", "PCP", "P2P", "STA-CFON"
    struct arg_str *dyn_bw_signaling;       //string: "Enable", "Disable"
    struct arg_str *encrypt_type;           //string: "AES-GCMP"
    struct arg_str *ext_schedule_ie;        //string: "Enable", "Disable"
    struct arg_int *heartbeat;              //boolean: 1, 0
    struct arg_str *interface_id;           //string
    struct arg_str *ldpc;                   //string: "Enable", "Disable"
    struct arg_int *mcs_fixed_rate;         //integer: 0-7
    struct arg_int *msdu_size;              //integer
    struct arg_str *mu_edca;                //string: "Override", if "Override", configure STA to use MU_EDCA parameters as listed in the test plan.

    struct arg_str *mu_txbf;                //string: "Enable", "Disable", MU TxBF beamformee capability with explicit feedback.
    struct arg_str *nss_mcs_cap;            //string: "nss_mcs_cap,1;0-7"
    struct arg_str *ofdma;                  //string: "DL", "UL"
    struct arg_str *om_control;             //string: "Enable", "Disable"
    struct arg_str *opt_mode_notif_ie;      //string: "Opt_md_notif_ie,1;20" (NSS=1, BW=20MHz)
    struct arg_int *percent_bcn_intval;     //integer: 0-100, percentage of beacon interval
    struct arg_str *ppdu_tx_type;           //string: "SU" "MU" "ER" "TB"
    struct arg_str *program;                //string: "11n" "WPS" "HE"
    struct arg_str *psk;                    //string: encryption passphrase
    struct arg_str *radio;                  //string: "ON", "OFF", this is used start/stop the device frame transmission/reception. turns the device radio on/off.

    struct arg_str *rts_force;              //string: "Enable", "Disable", force a STA to send RTS.
    struct arg_str *rts_width;              //string: "20" "40" "80", set the bandwidth in the RTS request initiated by the DUT
    struct arg_int *rx_ss;                  //integer: 1
    struct arg_str *security;               //string: "Open" "WPA2-PSK"
    struct arg_str *sgi_80m;                //string: "Enable", "Disable"
    struct arg_str *ssid;                   //string
    struct arg_int *rx_stbc_streams;        //short integer
    struct arg_str *tm_support;             //string: "Enable", "Disable", set/reset Extended capability(BIT23) in the Timing Measurement Action frame
    struct arg_int *trig_mac_padding_dur;   //integer
    struct arg_str *twt_req_support;        //string: "Enable", "Disable"

    struct arg_str *tx_bandwidth;           //string: "20" "40" "Auto"
    struct arg_str *tx_cts;                 //string: "Enable", "Disable", allows a STA to send 80MHz or 40MHz CTS
    struct arg_int *tx_lgi_rate;            //integer: set the TX Highest Supported Long Gi Data Rate subfield
    struct arg_str *tx_rts;                 //string: "Enable", "Disable", initialize the STA to transmit RTS frame with specified bandwidth indicated in the subsequent field
    struct arg_str *tx_bf;                  //string: "Enable", "Disable", SU TxBF beamformee capability with explicit feedback
    struct arg_int *tx_ss;                  //integer: 1
    struct arg_str *tx_pwr_change;          //string: "STEPUP" "STEPDOWN" "RESET", increase/decrease/reset transmit power
    struct arg_str *ultx_before_trig;       //string: "Enable", "Disable", trigger-enabled and announced TWT: configure UL transmission before sending Trigger frames.
    struct arg_str *uph;                    //string: "Enable", "Disable"
    struct arg_str *vht_tkip;               //string: "Enable", "Disable", enable/disable TKIP in VHT mode

    struct arg_str *vht_wep;               //string: "Enable", "Disable", enable/disable WEP in VHT mode
    struct arg_str *width;                 //string: "20" "40" "80" "160" "Auto"
    struct arg_str *wps4pin;               //string: "Enable", "Disable", enable/disable the generation of 4-digit PIN
    struct arg_str *zero_crc;              //string: "Enable", "Disable", set CRC field to all 0s

    struct arg_end *end;
} capi_sta_set_wireless_t;

static capi_sta_set_wireless_t capi_sta_set_wireless_args = { 0, };

static const char *TAG = "sta_set_wireless";

static int capi_sta_set_wireless_cmd_func(int argc, char **argv)
{
    char capi_sta_set_wireless[512] = { 0, };
    char capi_sta_set_wireless_param[32] = { 0, };
    int nerrors = arg_parse(argc, argv, (void **) &capi_sta_set_wireless_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, capi_sta_set_wireless_args.end, argv[0]);
        return 0;
    }
    sprintf(capi_sta_set_wireless, "sta_set_wireless");

    if (capi_sta_set_wireless_args.abftl_range->count) {
        sprintf(capi_sta_set_wireless_param, ",ABFTLRang,%s", capi_sta_set_wireless_args.abftl_range->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.addba_reject->count) {
        sprintf(capi_sta_set_wireless_param, ",ADDBA_Reject,%s", capi_sta_set_wireless_args.addba_reject->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.addba_req_bufsize->count) {
        sprintf(capi_sta_set_wireless_param, ",ADDBAReq_BufSize,%s", capi_sta_set_wireless_args.addba_req_bufsize->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.addba_rsp_bufsize->count) {
        sprintf(capi_sta_set_wireless_param, ",ADDBAResp_BufSize,%s", capi_sta_set_wireless_args.addba_rsp_bufsize->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.alloc_id->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",AllocID,%d", capi_sta_set_wireless_args.alloc_id->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.alloc_type->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",AllocType,%s", capi_sta_set_wireless_args.alloc_type->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ampdu->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",AMPDU,%s", capi_sta_set_wireless_args.ampdu->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.amsdu->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",AMSDU,%s", capi_sta_set_wireless_args.amsdu->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ba_param_amsdu_support->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BA_Param_AMSDU_Support,%d", capi_sta_set_wireless_args.ba_param_amsdu_support->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ba_recv_status->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BA_Recv_Status,%d", capi_sta_set_wireless_args.ba_recv_status->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    /*******/
    if (capi_sta_set_wireless_args.ba_recv_buf->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BAckRcvBuf,%d", capi_sta_set_wireless_args.ba_recv_buf->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.band->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Band,%s", capi_sta_set_wireless_args.band->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.bcc->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BCC,%s", capi_sta_set_wireless_args.bcc->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.bcn_interval->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BcnInt,%d", capi_sta_set_wireless_args.bcn_interval->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.bfmee_sts->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BeamformeeSTS,%d", capi_sta_set_wireless_args.bfmee_sts->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.bsrp_bqrp_ampdu_aggr->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BSRP_BQRP_AMPDU_Aggr,%s", capi_sta_set_wireless_args.bsrp_bqrp_ampdu_aggr->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.bw_signaling->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",BW_Sgnl,%s", capi_sta_set_wireless_args.bw_signaling->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.cbap_only->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",CBAPOnly,%d", capi_sta_set_wireless_args.cbap_only->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.channel->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Channel,%d", capi_sta_set_wireless_args.channel->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.cts_bandwidth->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",CTS_Width,%s", capi_sta_set_wireless_args.cts_bandwidth->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    /*******/
    if (capi_sta_set_wireless_args.dev_role->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",DevRole,%s", capi_sta_set_wireless_args.dev_role->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.dyn_bw_signaling->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",DYN_BW_Sgnl,%s", capi_sta_set_wireless_args.dyn_bw_signaling->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.encrypt_type->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Encrypt,%s", capi_sta_set_wireless_args.encrypt_type->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ext_schedule_ie->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",ExtSchIE,%s", capi_sta_set_wireless_args.ext_schedule_ie->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.heartbeat->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Heartbeat,%d", capi_sta_set_wireless_args.heartbeat->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.interface_id->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Interface,%s", capi_sta_set_wireless_args.interface_id->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ldpc->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",LDPC,%s", capi_sta_set_wireless_args.ldpc->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.mcs_fixed_rate->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",MCS_FixedRate,%d", capi_sta_set_wireless_args.mcs_fixed_rate->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.msdu_size->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",MSDUSize,%d", capi_sta_set_wireless_args.msdu_size->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.mu_edca->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",MU_EDCA,%s", capi_sta_set_wireless_args.mu_edca->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    /*******/
    if (capi_sta_set_wireless_args.mu_txbf->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",MU_TxBF,%s", capi_sta_set_wireless_args.mu_txbf->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.nss_mcs_cap->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",NSS_MCS_Cap,%s", capi_sta_set_wireless_args.nss_mcs_cap->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ofdma->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",OFDMA,%s", capi_sta_set_wireless_args.ofdma->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.om_control->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",OMControl,%s", capi_sta_set_wireless_args.om_control->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.opt_mode_notif_ie->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Opt_Md_Notif_IE,%s", capi_sta_set_wireless_args.opt_mode_notif_ie->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.percent_bcn_intval->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",PercentBI,%d", capi_sta_set_wireless_args.percent_bcn_intval->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ppdu_tx_type->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",PPDUTxType,%s", capi_sta_set_wireless_args.ppdu_tx_type->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.program->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Program,%s", capi_sta_set_wireless_args.program->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.psk->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",PSK,%s", capi_sta_set_wireless_args.psk->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.radio->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Radio,%s", capi_sta_set_wireless_args.radio->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    /*******/
    if (capi_sta_set_wireless_args.rts_force->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",RTS_Force,%s", capi_sta_set_wireless_args.rts_force->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.rts_width->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",RTS_Width,%s", capi_sta_set_wireless_args.rts_width->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.rx_ss->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",RXSP_Stream,%d", capi_sta_set_wireless_args.rx_ss->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.security->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Security,%s", capi_sta_set_wireless_args.security->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.sgi_80m->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",SGI80,%s", capi_sta_set_wireless_args.sgi_80m->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ssid->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",SSID,%s", capi_sta_set_wireless_args.ssid->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.rx_stbc_streams->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",STBC_RX,%d", capi_sta_set_wireless_args.rx_stbc_streams->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.tm_support->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",TM_Support,%s", capi_sta_set_wireless_args.tm_support->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.trig_mac_padding_dur->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Trig_MAC_Padding_Dur,%d", capi_sta_set_wireless_args.trig_mac_padding_dur->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.twt_req_support->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",TWT_ReqSupport,%s", capi_sta_set_wireless_args.twt_req_support->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    /*******/
    if (capi_sta_set_wireless_args.tx_bandwidth->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",TxBandwidth,%s", capi_sta_set_wireless_args.tx_bandwidth->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.tx_cts->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Tx_CTS,%s", capi_sta_set_wireless_args.tx_cts->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.tx_lgi_rate->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Tx_LGI_Rate,%d", capi_sta_set_wireless_args.tx_lgi_rate->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.tx_rts->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Tx_RTS,%s", capi_sta_set_wireless_args.tx_rts->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.tx_bf->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",TxBF,%s", capi_sta_set_wireless_args.tx_bf->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.tx_ss->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",TXSP_Stream,%d", capi_sta_set_wireless_args.tx_ss->ival[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.tx_pwr_change->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",TxPowerChange,%s", capi_sta_set_wireless_args.tx_pwr_change->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.ultx_before_trig->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",ULTxBeforeTrig,%s", capi_sta_set_wireless_args.ultx_before_trig->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.uph->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",UPH,%s", capi_sta_set_wireless_args.uph->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.vht_tkip->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",VHT_TKIP,%s", capi_sta_set_wireless_args.vht_tkip->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    /*******/
    if (capi_sta_set_wireless_args.vht_wep->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",VHT_WEP,%s", capi_sta_set_wireless_args.vht_wep->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.width->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Width,%s", capi_sta_set_wireless_args.width->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.wps4pin->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",WPS4PIN,%s", capi_sta_set_wireless_args.wps4pin->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }
    if (capi_sta_set_wireless_args.zero_crc->count == 1) {
        sprintf(capi_sta_set_wireless_param, ",Zero_CRC,%s", capi_sta_set_wireless_args.zero_crc->sval[0]);
        strcat(capi_sta_set_wireless, capi_sta_set_wireless_param);
    }

    ESP_LOGW(TAG, "%s", capi_sta_set_wireless);
    return 0;
}

void register_sta_set_wireless(void)
{
    capi_sta_set_wireless_args.abftl_range            = arg_str0(NULL, "abftl_range", "[abftl_range]", "e.g. Gt1>=1, specify the min value of the range");
    capi_sta_set_wireless_args.addba_reject           = arg_str0(NULL, "addba_reject", "[addba_reject]", "Enable/Disable, the rejection any ADDBA request by sending ADDBA response with the status \"decline\"");
    capi_sta_set_wireless_args.addba_req_bufsize      = arg_str0(NULL, "addba_req_bufsize", "[addba_req_bufsize]", "Buffer Size subfield of Block Ack Parameter Set of ADDBA request frame");
    capi_sta_set_wireless_args.addba_rsp_bufsize      = arg_str0(NULL, "addba_rsp_bufsize", "[addba_rsp_bufsize]", "Buffer Size subfield of Block Ack Parameter Set of ADDBA response frame");
    capi_sta_set_wireless_args.alloc_id               = arg_int0(NULL, "alloc_id", "[alloc_id]", "allocation ID used, integer(1-n)");
    capi_sta_set_wireless_args.alloc_type             = arg_str0(NULL, "alloc_type", "[alloc_type]", "CBAP");
    capi_sta_set_wireless_args.ampdu                  = arg_str0(NULL, "ampdu", "[ampdu]", "Enable/Disable A-MPDU");
    capi_sta_set_wireless_args.amsdu                  = arg_str0(NULL, "amsdu", "[amsdu]", "Enable/Disable A-MSDU");
    capi_sta_set_wireless_args.ba_param_amsdu_support = arg_int0(NULL, "ba_param_amsdu_support", "[ba_param_amsdu_support]", "boolean: 0-not set, 1-set");
    capi_sta_set_wireless_args.ba_recv_status         = arg_int0(NULL, "ba_recv_status", "[ba_recv_status]", "max number of MSDUs/A-MSDUs that can be acknowledged for BA frame");

    capi_sta_set_wireless_args.ba_recv_buf            = arg_int0(NULL, "ba_recv_buf", "[ba_recv_buf]", "BlockAck receive buffer size in terms of number of MPDUs, integer: 2 or 4");
    capi_sta_set_wireless_args.band                   = arg_str0(NULL, "band", "[band]", "2.4 or 5");
    capi_sta_set_wireless_args.bcc                    = arg_str0(NULL, "bcc", "[bcc]", "Enable/Disable BCC");
    capi_sta_set_wireless_args.bcn_interval           = arg_int0(NULL, "bcn_interval", "[bcn_interval]", "integer");
    capi_sta_set_wireless_args.bfmee_sts              = arg_int0(NULL, "bfmee_sts", "[bfmee_sts]", "integer: only support 1");
    capi_sta_set_wireless_args.bsrp_bqrp_ampdu_aggr   = arg_str0(NULL, "bsrp_bqrp_ampdu_aggr", "[bsrp_bqrp_ampdu_aggr]", "Enable/Disable, accepts BSRP/BQRP Trigger frame aggregated with other control/data/mgmt frames");
    capi_sta_set_wireless_args.bw_signaling           = arg_str0(NULL, "bw_signaling", "[bw_signaling]", "Enable/Disable, send out RTS with bandwidth signaling");
    capi_sta_set_wireless_args.cbap_only              = arg_int0(NULL, "cbap_only", "[cbap_only]", "CBAP only, boolean: 0, 1");
    capi_sta_set_wireless_args.channel                = arg_int0(NULL, "channel", "[channel]", "integer: channel");
    capi_sta_set_wireless_args.cts_bandwidth          = arg_str0(NULL, "cts_bandwidth", "[cts_bandwidth]", "string: 20, 40, 80");

    capi_sta_set_wireless_args.dev_role               = arg_str0(NULL, "dev_role", "[dev_role]", "string: STA, PCP, P2P, STA-CFON");
    capi_sta_set_wireless_args.dyn_bw_signaling       = arg_str0(NULL, "dyn_bw_signaling", "[dyn_bw_signaling]", "Enable/Disable, send out RTS with dynamic bandwidth signaling");
    capi_sta_set_wireless_args.encrypt_type           = arg_str0(NULL, "encrypt_type", "[encrypt_type]", "AES-GCMP");
    capi_sta_set_wireless_args.ext_schedule_ie        = arg_str0(NULL, "ext_schedule_ie", "[ext_schedule_ie]", "Enable/Disable, Extended Schedule IE");
    capi_sta_set_wireless_args.heartbeat              = arg_int0(NULL, "heartbeat", "[heartbeat]", "boolean: 0, 1");
    capi_sta_set_wireless_args.interface_id           = arg_str0(NULL, "interface_id", "[interface_id]", "string");
    capi_sta_set_wireless_args.ldpc                   = arg_str0(NULL, "ldpc", "[ldpc]", "Enable/Disable LDPC");
    capi_sta_set_wireless_args.mcs_fixed_rate         = arg_int0(NULL, "mcs_fixed_rate", "[mcs_fixed_rate]", "integer: 0-7");
    capi_sta_set_wireless_args.msdu_size              = arg_int0(NULL, "msdu_size", "[msdu_size]", "integer");
    capi_sta_set_wireless_args.mu_edca                = arg_str0(NULL, "mu_edca", "[mu_edca]", "Override-configure STA to use MU_EDCA parameters as listed in the test plan");

    capi_sta_set_wireless_args.mu_txbf                = arg_str0(NULL, "mu_txbf", "[mu_txbf]", "Enable/Disable, MU TxBF beamformee capability with explicit feedback.");
    capi_sta_set_wireless_args.nss_mcs_cap            = arg_str0(NULL, "nss_mcs_cap", "[nss_mcs_cap]", "nss_mcs_cap,1;0-7");
    capi_sta_set_wireless_args.ofdma                  = arg_str0(NULL, "ofdma", "[ofdma]", "DL/UL");
    capi_sta_set_wireless_args.om_control             = arg_str0(NULL, "om_control", "[om_control]", "Enable/Disable OM control field in HE Capability IE");
    capi_sta_set_wireless_args.opt_mode_notif_ie      = arg_str0(NULL, "opt_mode_notif_ie", "[opt_mode_notif_ie]", "Opt_md_notif_ie,1;20 (NSS=1, BW=20MHz)");
    capi_sta_set_wireless_args.percent_bcn_intval     = arg_int0(NULL, "percent_bcn_intval", "[percent_bcn_intval]", "integer:0-100, percentage of beacon interval");
    capi_sta_set_wireless_args.ppdu_tx_type           = arg_str0(NULL, "ppdu_tx_type", "[ppdu_tx_type]", "SU, MU, ER, TB");
    capi_sta_set_wireless_args.program                = arg_str0(NULL, "program", "[program]", "11n, WPS, HE");
    capi_sta_set_wireless_args.psk                    = arg_str0(NULL, "psk", "[psk]", "encryption passphrase");
    capi_sta_set_wireless_args.radio                  = arg_str0(NULL, "radio", "[radio]", "ON/OFF");

    capi_sta_set_wireless_args.rts_force              = arg_str0(NULL, "rts_force", "[rts_force]", "Enable/Disable, force a STA to send RTS");
    capi_sta_set_wireless_args.rts_width              = arg_str0(NULL, "rts_width", "[rts_width]", "20, 40, 80");
    capi_sta_set_wireless_args.rx_ss                  = arg_int0(NULL, "rx_ss", "[rx_ss]", "integer: only support 1");
    capi_sta_set_wireless_args.security               = arg_str0(NULL, "security", "[security]", "Open, WPA2-PSK");
    capi_sta_set_wireless_args.sgi_80m                = arg_str0(NULL, "sgi_80m", "[sgi_80m]", "Enable/Disable, SGI at 80MHz");
    capi_sta_set_wireless_args.ssid                   = arg_str0(NULL, "ssid", "[ssid]", "ssid");
    capi_sta_set_wireless_args.rx_stbc_streams        = arg_int0(NULL, "rx_stbc_streams", "[rx_stbc_streams]", "short integer: 2");
    capi_sta_set_wireless_args.tm_support             = arg_str0(NULL, "tm_support", "[tm_support]", "Enable/Disable, set/reset Extended capability(BIT23) in the Timing Measurement Action frame");
    capi_sta_set_wireless_args.trig_mac_padding_dur   = arg_int0(NULL, "trig_mac_padding_dur", "[trig_mac_padding_dur]", "integer");
    capi_sta_set_wireless_args.twt_req_support        = arg_str0(NULL, "twt_req_support", "[twt_req_support]", "Enable/Disable, TWT requester on test bed STA");

    capi_sta_set_wireless_args.tx_bandwidth           = arg_str0(NULL, "tx_bandwidth", "[tx_bandwidth]", "20, 40, Auto");
    capi_sta_set_wireless_args.tx_cts                 = arg_str0(NULL, "tx_cts", "[tx_cts]", "Enable/Disable, allows a STA to send 80MHz or 40MHz CTS");
    capi_sta_set_wireless_args.tx_lgi_rate            = arg_int0(NULL, "tx_lgi_rate", "[tx_lgi_rate]", "integer: set the TX Highest Supported Long Gi Data Rate subfield");
    capi_sta_set_wireless_args.tx_rts                 = arg_str0(NULL, "tx_rts", "[tx_rts]", "Enable/Disable, initialize the STA to transmit RTS frame with specified bandwidth indicated in the subsequent field");
    capi_sta_set_wireless_args.tx_bf                  = arg_str0(NULL, "tx_bf", "[tx_bf]", "Enable/Disable, SU TxBF beamformee capability with explicit feedback");
    capi_sta_set_wireless_args.tx_ss                  = arg_int0(NULL, "tx_ss", "[tx_ss]", "integer: only support 1");
    capi_sta_set_wireless_args.tx_pwr_change          = arg_str0(NULL, "tx_pwr_change", "[tx_pwr_change]", "STEPUP/STEPDOWN/RESET, increase/decrease/reset transmit power");
    capi_sta_set_wireless_args.ultx_before_trig       = arg_str0(NULL, "ultx_before_trig", "[ultx_before_trig]", "Enable/Disable, trigger-enabled and announced TWT: configure UL transmission before sending Trigger frames");
    capi_sta_set_wireless_args.uph                    = arg_str0(NULL, "uph", "[uph]", "Enable/Disable UPH");
    capi_sta_set_wireless_args.vht_tkip               = arg_str0(NULL, "vht_tkip", "[vht_tkip]", "Enable/Disable TKIP in VHT mode");

    capi_sta_set_wireless_args.vht_wep                = arg_str0(NULL, "vht_wep", "[vht_wep]", "Enable/Disable WEP in VHT mode");
    capi_sta_set_wireless_args.width                  = arg_str0(NULL, "width", "[width]", "20, 40, 80, 160, Auto");
    capi_sta_set_wireless_args.wps4pin                = arg_str0(NULL, "wps4pin", "[wps4pin]", "Enable/Disable the generation of 4-digit PIN");
    capi_sta_set_wireless_args.zero_crc               = arg_str0(NULL, "zero_crc", "[zero_crc]", "Enable/Disable, set CRC field to all 0s");
    capi_sta_set_wireless_args.end                    = arg_end(1);

    const esp_console_cmd_t capi_sta_set_wireless_cmd = {
        .command = "sta_set_wireless",
        .help = "Configure the STA settings",
        .hint = NULL,
        .func = &capi_sta_set_wireless_cmd_func,
        .argtable = &capi_sta_set_wireless_args,
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&capi_sta_set_wireless_cmd));
}
