/* ESP Wi-Fi Allaince Sigma agent software
*
* Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
* Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.
*
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <stdio.h>
#include <string.h>
#include "esp_log.h"
#include <sys/socket.h>
#include "esp_console.h"
#include "argtable3/argtable3.h"
#include "cmd_decl.h"
#include "freertos/FreeRTOS.h"
#include "freertos/event_groups.h"
#include "esp_wifi.h"
#include "esp_netif.h"
#include "esp_event.h"
#include "lwip/inet.h"
#include "lwip/netdb.h"
#include "lwip/sockets.h"
#include "ping/ping_sock.h"
#include "esp_sigma.h"
#include "wifi_debug.h"
#include "esp_heap_trace.h"
//Copied from Quicktrack definition. Sent as a flag in sendto function
#define MSG_CONFIRM 0x800
#define STA_INTERFACE_KEY "WIFI_STA_DEF"
extern esp_netif_t *netif_sta;

void *loopback_task_handler;

static struct {
    struct arg_dbl *timeout;
    struct arg_dbl *interval;
    struct arg_int *data_size;
    struct arg_int *count;
    struct arg_int *tos;
    struct arg_str *host;
    struct arg_lit *abort;
    struct arg_end *end;
} ping_args;

typedef struct loopback_receive_message_t {
    struct sockaddr_in addr;
    int s;
    bool is_running;
} loopback_receive_message_t;

struct loopback_receive_message_t loopback_receive_message_socket;

typedef struct {
    struct arg_str *profile;
    struct arg_str *direction;
    struct arg_str *source;
    struct arg_str *destination;
    struct arg_int *destinationPort;
    struct arg_int *sourcePort;
    struct arg_int *frameRate;
    struct arg_int *duration;
    struct arg_int *payloadSize;
    struct arg_int *startDelay;
    struct arg_str *trafficClass;
    struct arg_str *hti;
    struct arg_int *maxCnt;
    struct arg_end *end;
} sigma_traffic_config_t;

static sigma_traffic_config_t traffic_config_args;

typedef struct {
    struct arg_str *streamID;
    struct arg_end *end;
} sigma_traffic_send_t;
static sigma_traffic_send_t traffic_send_args;

/**********************************************************/
typedef struct {
    struct arg_int *port;
    struct arg_end *end;
} sigma_start_loopback_server_t;
static sigma_start_loopback_server_t start_loopback_server_args;

typedef struct {
    struct arg_end *end;
} sigma_stop_loopback_server_t;
static sigma_stop_loopback_server_t stop_loopback_server_args;

/*********************************************************/

typedef struct {
    struct arg_int *streamID;
    struct arg_end *end;
} sigma_traffic_receive_start_t;
static sigma_traffic_receive_start_t traffic_receive_start_args;

typedef struct {
    struct arg_int *streamID;
    struct arg_end *end;
} sigma_traffic_receive_stop_t;
static sigma_traffic_receive_stop_t traffic_receive_stop_args;

static struct sigma_dut sigma_dut;
static int receive_stream_num = 0;

static int sigma_traffic_config_cmd(int argc, char** argv)
{
    sigma_dut.debug_level = DUT_MSG_INFO;
    char tg_capi_cmd[512], capi_cmd[44];
    int nerrors = arg_parse(argc, argv, (void**) &traffic_config_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, traffic_config_args.end, argv[0]);
        return 1;
    }

    sprintf(tg_capi_cmd, "traffic_agent_config");

    if (traffic_config_args.profile->count == 1) {
        sprintf(capi_cmd, ",profile,%s", traffic_config_args.profile->sval[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.direction->count == 1) {
        sprintf(capi_cmd, ",direction,%s", traffic_config_args.direction->sval[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.source->count == 1) {
        sprintf(capi_cmd, ",source,%s", traffic_config_args.source->sval[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.destination->count == 1) {
        sprintf(capi_cmd, ",destination,%s", traffic_config_args.destination->sval[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.destinationPort->count == 1) {
        sprintf(capi_cmd, ",destinationPort,%d",  traffic_config_args.destinationPort->ival[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.sourcePort->count == 1) {
        sprintf(capi_cmd, ",sourcePort,%d", traffic_config_args.sourcePort->ival[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.frameRate->count == 1) {
        sprintf(capi_cmd, ",frameRate,%d", traffic_config_args.frameRate->ival[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.duration->count == 1) {
        sprintf(capi_cmd, ",duration,%d", traffic_config_args.duration->ival[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.payloadSize->count == 1) {
        sprintf(capi_cmd, ",payloadSize,%d", traffic_config_args.payloadSize->ival[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.startDelay->count == 1) {
        sprintf(capi_cmd, ",startDelay,%d", traffic_config_args.startDelay->ival[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.trafficClass->count == 1) {
        sprintf(capi_cmd, ",trafficClass,%s", traffic_config_args.trafficClass->sval[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.hti->count == 1) {
        sprintf(capi_cmd, ",hti,%s", traffic_config_args.hti->sval[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    if (traffic_config_args.maxCnt->count == 1) {
        sprintf(capi_cmd, ",hti,%d", traffic_config_args.maxCnt->ival[0]);
        strcat(tg_capi_cmd, capi_cmd);
    }

    process_sigma_traffic_cmd(&sigma_dut, tg_capi_cmd, SIGMA_TRAFFIC_CONFIG);
    return 0;
}

static int sigma_traffic_send_cmd(int argc, char** argv)
{

    int nerrors = arg_parse(argc, argv, (void**) &traffic_send_args);
    char capi_cmd[64];
    char *tg_capi_cmd = (char*)malloc(sizeof(char) * 200);
    if (nerrors != 0) {
        arg_print_errors(stderr, traffic_send_args.end, argv[0]);
        return 1;
    }

    sprintf(capi_cmd, "traffic_agent_send");
    if (traffic_send_args.streamID->count == 1) {
        char *streamID = traffic_send_args.streamID->sval[0];
        char temp[strlen(streamID)];
        memset(temp, 0, sizeof(*streamID));
        /*case of single stream ID  i.e "{streamID}$"  Assuming single digit streamIDs*/
        if (strlen(streamID) == 2) {
            strncat(temp, streamID, 1);
        } else {
            for (int i = 0; i < strlen(streamID); i++) {
                if (!strncmp(streamID + i, "$", strlen("$"))) {
                    strcat(temp, " ");
                    i += strlen("$") - 1;
                } else {
                    strncat(temp, streamID + i, 1);
                }
            }
        }

        strcpy(streamID, temp);
        sprintf(tg_capi_cmd, "%s,streamID,%s", capi_cmd, streamID);
        printf("%s,streamID,%s\n", capi_cmd, streamID);
    }
    process_sigma_traffic_cmd(&sigma_dut, tg_capi_cmd, SIGMA_TRAFFIC_SEND);
    free(tg_capi_cmd);
    return 0;
}

static void loopback_server_receive_message_thread(void *sock_ctx)
{
    struct sockaddr_storage from;
    unsigned char buffer[1200];
    int len;
    socklen_t fromlen;

    fromlen = sizeof(from);
    while (1) {
        len = recvfrom(loopback_receive_message_socket.s, buffer, 2048, 0, (struct sockaddr *) &from, &fromlen);
        if (!loopback_receive_message_socket.is_running) {
            break;
        }
        printf("Got packet of len=%d\n", len);
        if (len < 0) {
            printf("\nLoopback server recvfrom[server] error");
            return ;
        }

        len = sendto(loopback_receive_message_socket.s, (const char *)buffer, len, MSG_CONFIRM, (struct sockaddr *)&from, sizeof(from));
        printf("sent back packet of len=%d\n", len);
    }
    vTaskDelete(NULL);
}

static int sigma_stop_loopback_server_cmd(int argc, char** argv)
{
    /* doing this should be suffient? */
    if (loopback_receive_message_socket.is_running == true) {
        loopback_receive_message_socket.is_running = false;
        close(loopback_receive_message_socket.s);
        loopback_receive_message_socket.s = -1;
    }
    return 0;
}

void wifi_get_local_ip(esp_netif_ip_info_t *ip_info);
static int sigma_start_loopback_server_cmd(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &start_loopback_server_args);
    loopback_receive_message_socket.s = 0;
    if (nerrors > 0) {
        printf("(sigma_start_loopback_server_cmd) : Failed to parse parameters\n");
        return -1;
    }

    socklen_t len = sizeof(loopback_receive_message_socket.addr);
    esp_netif_ip_info_t ip_info;

    if (esp_netif_get_ip_info(netif_sta, &ip_info) != ESP_OK) {
        printf("(sigma_start_loopback_server_cmd) : Failed to get ip info\n");
        return -1;
    }

    /* Open UDP socket */
    loopback_receive_message_socket.s = socket(PF_INET, SOCK_DGRAM, 0);
    if (loopback_receive_message_socket.s < 0) {
        printf("Failed to open server socket\n");
        return -1;
    }
    /* Bind specific IP */
    memset(&loopback_receive_message_socket.addr, 0, sizeof(loopback_receive_message_socket.addr));
    loopback_receive_message_socket.addr.sin_family = AF_INET;
    wifi_get_local_ip(&ip_info);
    printf("\tip: %d.%d.%d.%d\n", ip_info.ip.addr & 0xFF, (ip_info.ip.addr >> 8) & 0xFF, (ip_info.ip.addr >> 16) & 0xFF, (ip_info.ip.addr >> 24) & 0xFF);
    if (ip_info.ip.addr != 0) {
        loopback_receive_message_socket.addr.sin_addr.s_addr = ip_info.ip.addr;
    }
    printf("\tLoopback ip: %d.%d.%d.%d\n", loopback_receive_message_socket.addr.sin_addr.s_addr & 0xFF, (loopback_receive_message_socket.addr.sin_addr.s_addr >> 8) & 0xFF, (loopback_receive_message_socket.addr.sin_addr.s_addr >> 16) & 0xFF, (loopback_receive_message_socket.addr.sin_addr.s_addr >> 24) & 0xFF);
    loopback_receive_message_socket.addr.sin_port = htons(start_loopback_server_args.port->ival[0]);
    if (bind(loopback_receive_message_socket.s, (struct sockaddr *) &loopback_receive_message_socket.addr, sizeof(loopback_receive_message_socket.addr)) < 0) {
        printf("Failed to bind server socket\n");
        close(loopback_receive_message_socket.s);
        return -1;
    }
    loopback_receive_message_socket.is_running = true;
    int res = xTaskCreatePinnedToCore(loopback_server_receive_message_thread, "serve1", 5048, NULL, 7, NULL, 0);
    return 0;
}

static int sigma_traffic_receive_start_cmd(int argc, char** argv)
{

    int nerrors = arg_parse(argc, argv, (void**) &traffic_receive_start_args);
    char tg_capi_cmd[200], capi_cmd[44];

    if (nerrors != 0) {
        arg_print_errors(stderr, traffic_receive_start_args.end, argv[0]);
        return 1;
    }

    sprintf(capi_cmd, "traffic_agent_receive_start");
    if (traffic_receive_start_args.streamID->count == 1) {
        sprintf(tg_capi_cmd, "%s,streamID,%d", capi_cmd, traffic_receive_start_args.streamID->ival[0]);
    }

#if CONFIG_IDF_TARGET_ESP32C6 && CONFIG_ESP_WIFI_ENABLE_WIFI_RX_STATS
    wifi_cmd_clr_rx_statistics(0, NULL);
#endif
    process_sigma_traffic_cmd(&sigma_dut, tg_capi_cmd, SIGMA_TRAFFIC_RECEIVE_START);
    receive_stream_num += 1;

    return 0;
}

static int sigma_traffic_receive_stop_cmd(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &traffic_receive_stop_args);
    char tg_capi_cmd[200], capi_cmd[44];

    if (nerrors != 0) {
        arg_print_errors(stderr, traffic_receive_stop_args.end, argv[0]);
        return 1;
    }

    sprintf(capi_cmd, "traffic_agent_receive_stop");
    if (traffic_receive_stop_args.streamID->count == 1) {
        sprintf(tg_capi_cmd, "%s,streamID,%d", capi_cmd, traffic_receive_stop_args.streamID->ival[0]);
    }

    process_sigma_traffic_cmd(&sigma_dut, tg_capi_cmd, SIGMA_TRAFFIC_RECEIVE_STOP);
    //only get one time
    receive_stream_num -= 1;
    if (receive_stream_num == 0) {
#if CONFIG_IDF_TARGET_ESP32C6 && CONFIG_ESP_WIFI_ENABLE_WIFI_RX_STATS
        wifi_cmd_get_rx_statistics(0, NULL);
#endif
    }
    return 0;
}
static int used_flag;
static int sigma_traffic_reset_cmd(int argc, char** argv)
{
#if 0
    if (used_flag == 0) {
    heap_trace_record_t * buf11 = (heap_trace_record_t *)calloc(60, sizeof(heap_trace_record_t));
    heap_trace_init_standalone(buf11, 60);
    heap_trace_start(HEAP_TRACE_LEAKS);
    used_flag = 1;
    }
    else if (used_flag < 2) {
        used_flag++;
    } else if (used_flag == 2){
        heap_trace_stop();
        heap_trace_dump();
    }
#endif
    char tg_capi_cmd[44];
    sprintf(tg_capi_cmd, "traffic_agent_reset");

    process_sigma_traffic_cmd(&sigma_dut, tg_capi_cmd, SIGMA_TRAFFIC_RESET);
    return 0;
}

static void cmd_ping_on_ping_success(esp_ping_handle_t hdl, void *args)
{
    uint8_t ttl;
    uint16_t seqno;
    uint32_t elapsed_time, recv_len;
    ip_addr_t target_addr;
    esp_ping_get_profile(hdl, ESP_PING_PROF_SEQNO, &seqno, sizeof(seqno));
    esp_ping_get_profile(hdl, ESP_PING_PROF_TTL, &ttl, sizeof(ttl));
    esp_ping_get_profile(hdl, ESP_PING_PROF_IPADDR, &target_addr, sizeof(target_addr));
    esp_ping_get_profile(hdl, ESP_PING_PROF_SIZE, &recv_len, sizeof(recv_len));
    esp_ping_get_profile(hdl, ESP_PING_PROF_TIMEGAP, &elapsed_time, sizeof(elapsed_time));
    printf("%d bytes from %s icmp_seq=%d ttl=%d time=%d ms\n",
           recv_len, ipaddr_ntoa((ip_addr_t*)&target_addr), seqno, ttl, elapsed_time);
}

static void cmd_ping_on_ping_timeout(esp_ping_handle_t hdl, void *args)
{
    uint16_t seqno;
    ip_addr_t target_addr;
    esp_ping_get_profile(hdl, ESP_PING_PROF_SEQNO, &seqno, sizeof(seqno));
    esp_ping_get_profile(hdl, ESP_PING_PROF_IPADDR, &target_addr, sizeof(target_addr));
    printf("From %s icmp_seq=%d timeout\n", ipaddr_ntoa((ip_addr_t*)&target_addr), seqno);
}

static void cmd_ping_on_ping_end(esp_ping_handle_t hdl, void *args)
{
    ip_addr_t target_addr;
    uint32_t transmitted;
    uint32_t received;
    uint32_t total_time_ms;
    esp_ping_get_profile(hdl, ESP_PING_PROF_REQUEST, &transmitted, sizeof(transmitted));
    esp_ping_get_profile(hdl, ESP_PING_PROF_REPLY, &received, sizeof(received));
    esp_ping_get_profile(hdl, ESP_PING_PROF_IPADDR, &target_addr, sizeof(target_addr));
    esp_ping_get_profile(hdl, ESP_PING_PROF_DURATION, &total_time_ms, sizeof(total_time_ms));
    uint32_t loss = (uint32_t)((1 - ((float)received) / transmitted) * 100);
    if (IP_IS_V4(&target_addr)) {
        printf("\n--- %s ping statistics ---\n", inet_ntoa(*ip_2_ip4(&target_addr)));
    } else {
        printf("\n--- %s ping statistics ---\n", inet6_ntoa(*ip_2_ip6(&target_addr)));
    }
    printf("%d packets transmitted, %d received, %d%% packet loss, time %dms\n",
           transmitted, received, loss, total_time_ms);
    // delete the ping sessions, so that we clean up all resources and can create a new ping session
    // we don't have to call delete function in the callback, instead we can call delete function from other tasks
    esp_ping_delete_session(hdl);
}

static int do_ping_cmd(int argc, char **argv)
{
    esp_ping_config_t config = ESP_PING_DEFAULT_CONFIG();
    static esp_ping_handle_t ping;
    int nerrors = arg_parse(argc, argv, (void **)&ping_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, ping_args.end, argv[0]);
        return 1;
    }

    if (ping_args.timeout->count > 0) {
        config.timeout_ms = (uint32_t)(ping_args.timeout->dval[0] * 1000);
    }

    if (ping_args.interval->count > 0) {
        config.interval_ms = (uint32_t)(ping_args.interval->dval[0] * 1000);
    }

    if (ping_args.data_size->count > 0) {
        config.data_size = (uint32_t)(ping_args.data_size->ival[0]);
    }

    if (ping_args.count->count > 0) {
        config.count = (uint32_t)(ping_args.count->ival[0]);
    }

    if (ping_args.tos->count > 0) {
        config.tos = (uint32_t)(ping_args.tos->ival[0]);
    }

    if (ping_args.abort->count) {
        esp_ping_stop(ping);
        return 0;
    }

    // parse IP address
    ip_addr_t target_addr;
    struct addrinfo hint;
    struct sockaddr_in6 sock_addr6;
    struct addrinfo *res = NULL;
    memset(&hint, 0, sizeof(hint));
    memset(&target_addr, 0, sizeof(target_addr));
    /* convert domain name to IP address */
    if (inet_pton(AF_INET6, ping_args.host->sval[0], &sock_addr6.sin6_addr) == 1) {
        /* convert ip6 string to ip6 address */
        ipaddr_aton(ping_args.host->sval[0], &target_addr);
        if (netif_sta) {
            config.interface = esp_netif_get_netif_impl_index(netif_sta);
        }
    } else {
        if (getaddrinfo(ping_args.host->sval[0], NULL, &hint, &res) != 0) {
            printf("ping: unknown host %s\n", ping_args.host->sval[0]);
            return 1;
        }
        if (res->ai_family == AF_INET) {
            struct in_addr addr4 = ((struct sockaddr_in *)(res->ai_addr))->sin_addr;
            inet_addr_to_ip4addr(ip_2_ip4(&target_addr), &addr4);
        } else {
            struct in6_addr addr6 = ((struct sockaddr_in6 *)(res->ai_addr))->sin6_addr;
            inet6_addr_to_ip6addr(ip_2_ip6(&target_addr), &addr6);
        }
        freeaddrinfo(res);
    }
    config.target_addr = target_addr;
    config.task_stack_size = 2048 + 1024;

    /* set callback functions */
    esp_ping_callbacks_t cbs = {
        .on_ping_success = cmd_ping_on_ping_success,
        .on_ping_timeout = cmd_ping_on_ping_timeout,
        .on_ping_end = cmd_ping_on_ping_end,
        .cb_args = NULL
    };

    esp_ping_new_session(&config, &cbs, &ping);
    esp_ping_start(ping);

    return 0;
}

void register_traffic()
{

    ping_args.timeout = arg_dbl0("W", "timeout", "<t>", "Time to wait for a response, in seconds");
    ping_args.interval = arg_dbl0("i", "interval", "<t>", "Wait interval seconds between sending each packet");
    ping_args.data_size = arg_int0("s", "size", "<n>", "Specify the number of data bytes to be sent");
    ping_args.count = arg_int0("c", "count", "<n>", "Stop after sending count packets");
    ping_args.tos = arg_int0("Q", "tos", "<n>", "Set Type of Service related bits in IP datagrams");
    ping_args.host = arg_str0(NULL, NULL, "[host]", "Host address");
    ping_args.abort = arg_lit0("a", "abort", "abort");
    ping_args.end = arg_end(1);
    const esp_console_cmd_t ping_cmd = {
        .command = "ping",
        .help = "send ICMP ECHO_REQUEST to network hosts",
        .hint = NULL,
        .func = &do_ping_cmd,
        .argtable = &ping_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&ping_cmd));

    traffic_config_args.profile = arg_str0(NULL, "profile", "<profile>", "wfa sigma traffic profile");
    traffic_config_args.direction = arg_str0(NULL, "direction", "<direction>", "Sigma traffic direction");
    traffic_config_args.source = arg_str0(NULL, "source", "<source>", "Sigma traffic source");
    traffic_config_args.destination = arg_str0(NULL, "destination", "<destination>", "Sigma traffic destination");
    traffic_config_args.destinationPort = arg_int0(NULL, "destinationPort", "<destinationPort>", "Sigma traffic destination port");
    traffic_config_args.sourcePort = arg_int0(NULL, "sourcePort", "<sourcePort>", "Sigma traffic source port");
    traffic_config_args.frameRate = arg_int0(NULL, "frameRate", "<frameRate>", "Sigma traffic frame rate");
    traffic_config_args.duration = arg_int0(NULL, "duration", "<duration>", "Sigma traffic duration");
    traffic_config_args.payloadSize = arg_int0(NULL, "payloadSize", "<payloadSize>", "Sigma traffic payloadSize");
    traffic_config_args.startDelay = arg_int0(NULL, "startDelay", "<startDelay>", "Sigma traffic start delay");
    traffic_config_args.trafficClass = arg_str0(NULL, "trafficClass", "<trafficClass>", "Sigma traffic traffic class");
    traffic_config_args.hti = arg_str0(NULL, "HTI", "<HTI>", "Sigma traffic hti");
    traffic_config_args.maxCnt = arg_int0(NULL, "maxCnt", "<maxCnt>", "Sigma traffic maximum count");
    traffic_config_args.end = arg_end(1);

    const esp_console_cmd_t traffic_config_cmd = {
        .command = "sigma_traffic_config",
        .help = "Sigma capi traffic config arguments",
        .hint = NULL,
        .func = &sigma_traffic_config_cmd,
        .argtable = &traffic_config_args
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&traffic_config_cmd));

    traffic_send_args.streamID = arg_str0(NULL, "id", "<stream_id>", "Sigma send traffic stream");
    traffic_send_args.end = arg_end(1);

    const esp_console_cmd_t traffic_send_cmd = {
        .command = "sigma_traffic_send",
        .help = "Sigma capi traffic send",
        .hint = NULL,
        .func = &sigma_traffic_send_cmd,
        .argtable = &traffic_send_args
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&traffic_send_cmd));

    /******************************************************/

    start_loopback_server_args.port = arg_int0(NULL, "port", "<port>", "Port at which Sigma loopback server starts");
    start_loopback_server_args.end = arg_end(1);

    const esp_console_cmd_t start_loopback_server = {
        .command = "sigma_loopback_server_start",
        .help = "Sigma capi Star a loopback server at a specific port",
        .hint = NULL,
        .func = &sigma_start_loopback_server_cmd,
        .argtable = &start_loopback_server_args
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&start_loopback_server));

    stop_loopback_server_args.end = arg_end(1);

    const esp_console_cmd_t stop_loopback_server = {
        .command = "sigma_loopback_server_stop",
        .help = "Sigma stop loopback server",
        .hint = NULL,
        .func = &sigma_stop_loopback_server_cmd,
        .argtable = &stop_loopback_server_args
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&stop_loopback_server));

    /***************************************************/

    traffic_receive_start_args.streamID = arg_int0(NULL, "id", "<stream_id>", "Sigma receive traffic stream");
    traffic_receive_start_args.end = arg_end(1);

    const esp_console_cmd_t traffic_receive_cmd = {
        .command = "sigma_traffic_receive_start",
        .help = "Sigma capi traffic receive start",
        .hint = NULL,
        .func = &sigma_traffic_receive_start_cmd,
        .argtable = &traffic_receive_start_args
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&traffic_receive_cmd));

    traffic_receive_stop_args.streamID = arg_int0(NULL, "id", "<stream_id>", "Sigma receive traffic stream");
    traffic_receive_stop_args.end = arg_end(1);

    const esp_console_cmd_t traffic_receive_stop_cmd = {
        .command = "sigma_traffic_receive_stop",
        .help = "Sigma capi traffic receive stop",
        .hint = NULL,
        .func = &sigma_traffic_receive_stop_cmd,
        .argtable = &traffic_receive_stop_args
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&traffic_receive_stop_cmd));

    const esp_console_cmd_t traffic_reset_cmd = {
        .command = "sigma_traffic_reset",
        .help = "Sigma capi traffic reset",
        .hint = NULL,
        .func = &sigma_traffic_reset_cmd,
        .argtable = NULL
    };

    ESP_ERROR_CHECK(esp_console_cmd_register(&traffic_reset_cmd));
}
