/* ESP Wi-Fi Allaince Sigma agent software
*
* Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
* Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.
*
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <stdio.h>
#include <string.h>
#include "esp_idf_version.h"
#include "esp_log.h"
#include "esp_console.h"
#include "argtable3/argtable3.h"
#include "cmd_decl.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "freertos/event_groups.h"
#include "esp_wifi.h"
#ifdef CONFIG_WPA_MBO_SUPPORT
#include "esp_wnm.h"
#include "esp_rrm.h"
#include "esp_mbo.h"
#endif

/* IDF MR: 25233*/
#if (defined __has_include && __has_include("esp_eap_client.h"))
#include "esp_eap_client.h"
#else
#include "esp_wpa2.h" /* use old wpa2 functions */
#ifndef esp_eap_client_set_ca_cert
#define esp_eap_client_set_ca_cert esp_wifi_sta_wpa2_ent_set_ca_cert
#define esp_eap_client_set_certificate_and_key esp_wifi_sta_wpa2_ent_set_cert_key
#define esp_eap_client_set_identity esp_wifi_sta_wpa2_ent_set_identity
#define esp_eap_client_set_username esp_wifi_sta_wpa2_ent_set_username
#define esp_eap_client_set_password esp_wifi_sta_wpa2_ent_set_password
#define esp_eap_client_set_suiteb_192bit_certification esp_wifi_sta_wpa2_set_suiteb_192bit_certification
#define esp_wifi_sta_enterprise_enable esp_wifi_sta_wpa2_ent_enable
#endif
#endif

#include "esp_netif.h"
#include "esp_event.h"
#include "esp_partition.h"
// (20/40M)
#include "esp_private/esp_wifi_private.h"
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
#include "esp_private/esp_wifi_he_private.h"
#endif
#include "esp_private/wifi_types.h"
#include "esp_private/wifi.h"

#include "esp_pm.h"
#include "iperf.h"
#include "wifi_debug.h"
#if CONFIG_IDF_TARGET_ESP32C6
#include "soc/lp_aon_reg.h"
#include "soc/lp_io_reg.h"
#include "soc/lp_clkrst_reg.h"
#include "soc/lp_io_reg.h"
#include "soc/systimer_reg.h"
#include "esp32c6/rom/ets_sys.h"
#include "esp_wifi_he.h"
#endif
#include "soc/gpio_reg.h"
#include "driver/gpio.h"

#include "esp_wnm.h"
#include "esp_sigma.h"

/*******************************************************
 *                Macros
 *******************************************************/
#define ESP_WIFI_HAL_TEST                  (0)
#define HAL_MAC_ESP32AX_SIGMA_5_53_1       (0) //5.53.1, BSRP

#define MACSTR "%02x:%02x:%02x:%02x:%02x:%02x"
#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
bool str_2_mac(uint8_t *str, uint8_t *dest);
#define STR2MAC(in, out) str_2_mac((uint8_t *)in, (uint8_t *)out)

#define OPEN "open"
#define WPA2 "wpa2"
#define WPA2_EAP "wpa2-eap"
#define WEP "wep"
#define WPA3 "wpa3"

#define EAP_PEAP "peap"
#define EAP_TLS  "tls"
#define EAP_TTLS "ttls"

#define CERT_MAX_LENGTH 8192
#define CERT_EMBEDDED

#ifdef CERT_EMBEDDED

/* User certificates */
extern uint8_t wifiuser_crt_start[] asm("_binary_wifiuser_pem_start");
extern uint8_t wifiuser_crt_end[] asm("_binary_wifiuser_pem_end");

extern uint8_t user_p256_crt_start[] asm("_binary_user_p256_pem_start");
extern uint8_t user_p256_crt_end[] asm("_binary_user_p256_pem_end");

extern uint8_t rsa3072_user_crt_start[] asm("_binary_rsa3072_user_pem_start");
extern uint8_t rsa3072_user_crt_end[] asm("_binary_rsa3072_user_pem_end");

extern uint8_t rsa3072_user_rsa2048_crt_start[] asm("_binary_rsa3072_user_rsa2048_pem_start");
extern uint8_t rsa3072_user_rsa2048_crt_end[] asm("_binary_rsa3072_user_rsa2048_pem_end");

extern uint8_t ec2_user_crt_start[] asm("_binary_ec2_user_pem_start");
extern uint8_t ec2_user_crt_end[]   asm("_binary_ec2_user_pem_end");

//extern uint8_t ec2_user_p256_crt_start[] asm("_binary_ec2_user_p256_pem_start");
//extern uint8_t ec2_user_p256_crt_end[]   asm("_binary_ec2_user_p256_pem_end");

extern uint8_t user1_crt_start[] asm("_binary_rsa_user1_w1_fi_pem_start");
extern uint8_t user1_crt_end[]   asm("_binary_rsa_user1_w1_fi_pem_end");

extern uint8_t winuser_crt_start[] asm("_binary_wifiuserwin_pem_start");
extern uint8_t winuser_crt_end[]   asm("_binary_wifiuserwin_pem_end");
/* CA certificates */
extern uint8_t cas_pem_start[] asm("_binary_cas_pem_start");
extern uint8_t cas_pem_end[]   asm("_binary_cas_pem_end");

extern uint8_t ec2_ca_pem_start[] asm("_binary_ec2_ca_pem_start");
extern uint8_t ec2_ca_pem_end[]   asm("_binary_ec2_ca_pem_end");

extern uint8_t rsa3072_ca_pem_start[] asm("_binary_rsa3072_ca_pem_start");
extern uint8_t rsa3072_ca_pem_end[]   asm("_binary_rsa3072_ca_pem_end");

extern uint8_t ca2_pem_start[] asm("_binary_rsa_ca2_w1_fi_pem_start");
extern uint8_t ca2_pem_end[]   asm("_binary_rsa_ca2_w1_fi_pem_end");

extern uint8_t ca1_pem_start[] asm("_binary_rsa_ca1_w1_fi_pem_start");
extern uint8_t ca1_pem_end[]   asm("_binary_rsa_ca1_w1_fi_pem_end");

extern uint8_t caswin_pem_start[] asm("_binary_caswin_pem_start");
extern uint8_t caswin_pem_end[]   asm("_binary_caswin_pem_end");
#endif

/*******************************************************
 *                Structures
 *******************************************************/

typedef struct {
    struct arg_str *ssid;
    struct arg_str *security;
    struct arg_str *eap_method;
    struct arg_str *eap_id;
    struct arg_str *username;
    struct arg_str *password;
    struct arg_str *pmf;
    struct arg_str *trustedRootCA;
    struct arg_str *clientCertificate;
    struct arg_str *roaming;
    struct arg_str *authmode;
    struct arg_str *keymgmtType;
    struct arg_dbl *addba_reject;
    struct arg_str *bssid;
    struct arg_str *btm;
    struct arg_end *end;
} wifi_args_t;

typedef struct {
    struct arg_str *ssid;
    struct arg_str *password;
    struct arg_str *cbw;
    struct arg_str *chan;
    struct arg_end *end;
} wifi_ap_args_t;

typedef struct {
    struct arg_str *ip;
    struct arg_str *gw;
    struct arg_str *netmask;
    struct arg_end *end;
} static_ip_args_t;

#ifdef CONFIG_WPA_MBO_SUPPORT
typedef struct {
    struct arg_int *reset;
    struct arg_int *ch_pref_num1;
    struct arg_int *ch_pref1;
    struct arg_int *ch_op_class1;
    struct arg_int *ch_reason_code1;
    struct arg_int *ch_pref_num2;
    struct arg_int *ch_pref2;
    struct arg_int *ch_op_class2;
    struct arg_int *ch_reason_code2;
    struct arg_end *end;
} wifi_mbo_set_chan_pref_args_t;
#endif

typedef struct {
    struct arg_str *ssid;
    struct arg_end *end;
} wifi_scan_arg_t;
typedef struct {
    struct arg_int *mode;
    struct arg_end *end;
} wifi_powersave_arg_t;

typedef struct {
    struct arg_str *reason_code;
    struct arg_str *candidate_list;
    struct arg_end *end;
} wifi_btm_args_t;

typedef struct {
    struct arg_str *mac;
    struct arg_end *end;
} wifi_disconnect_arg_t;

typedef struct {
    struct arg_int *level;
    struct arg_end *end;
} loglevel_args_t;

typedef struct {
    struct arg_int *time;
    struct arg_end *end;
} settime_args_t;

typedef struct {
    struct arg_str *prog;
    struct arg_end *end;
} setprog_args_t;

typedef enum {
    PKI_CA = 1,
    PKI_CLIENT_CERT = 2,
    PKI_CLIENT_KEY = 3,
} wpa2_pki_type_t;

typedef struct {
    struct arg_dbl *enable;
    struct arg_end *end;
} test_rfo_dump_t;

#if ESP_WIFI_HAL_TEST
typedef struct {
    struct arg_dbl *enable;
    struct arg_end *end;
} test_su_dump_t;

typedef struct {
    struct arg_int *ull;
    struct arg_end *end;
} test_ull_args_t;
#endif

typedef struct {
    struct arg_int *gain;
    struct arg_end *end;
} test_force_gain_t;

typedef struct {
    struct arg_int *mcs;
    struct arg_int *bcc;
    struct arg_int *ltf;
    struct arg_int *gi;
    struct arg_int *nss;
    struct arg_int *ersu;
    struct arg_end *end;
} wifi_set_args_t;

typedef struct {
    struct arg_dbl *enable;
    struct arg_end *end;
} wifi_set_amsdu_t;

typedef struct {
    struct arg_dbl *enable;
    struct arg_end *end;
} wifi_set_obss_scan_t;

typedef struct {
    struct arg_str *ip;
    struct arg_lit *server;
    struct arg_lit *udp;
    struct arg_lit *version;
    struct arg_int *port;
    struct arg_int *length;
    struct arg_int *interval;
    struct arg_int *time;
    struct arg_int *bw_limit;
    struct arg_lit *abort;
    struct arg_end *end;
} wifi_iperf_t;

typedef struct {
    struct arg_int *tid;
    struct arg_end *end;
} wifi_addba_args_t;

typedef struct {
    struct arg_int *offset;
    struct arg_end *end;
} wifi_itwt_set_twt_offset_t;

/*******************************************************
 *                Variable Definitions
 *******************************************************/

static wifi_args_t sta_args;
static wifi_scan_arg_t scan_args;
static wifi_powersave_arg_t powersave_args;
static static_ip_args_t static_ip_args;
static wifi_btm_args_t btm_args;
#ifdef CONFIG_WPA_MBO_SUPPORT
static wifi_mbo_set_chan_pref_args_t wifi_mbo_set_chan_pref_args;
#endif
static wifi_ap_args_t ap_args;
static wifi_disconnect_arg_t disconnect_args;
static loglevel_args_t loglevel_args;
static settime_args_t settime_args;
static setprog_args_t setprog_args;
static bool reconnect = true;
static const char *TAG = "cmd_wifi";
esp_netif_t *netif_ap = NULL;
esp_netif_t *netif_sta = NULL;

EventGroupHandle_t wifi_event_group;
const int CONNECTED_BIT = BIT0;
const int DISCONNECTED_BIT = BIT1;
const int GOT_IP_BIT = BIT2;

static test_rfo_dump_t test_rfo_dump_args;
#if ESP_WIFI_HAL_TEST
static test_su_dump_t test_su_dump_args;
static test_ull_args_t test_ull_args;
#endif
static bool g_roaming = true;

static test_force_gain_t force_gain_args;
static wifi_set_args_t set_args;
static wifi_set_amsdu_t set_amsdu_args;
static wifi_set_obss_scan_t set_obss_args;
static wifi_iperf_t iperf_args;
static wifi_addba_args_t addba_args;
static wifi_itwt_set_twt_offset_t itwt_set_twt_offset_args;
#if CONFIG_SOC_WIFI_HE_SUPPORT
extern int esp_wifi_cert_tx_mcs; //mcs = [0, 7]
extern int esp_wifi_cert_tx_bcc; //enable
extern int esp_wifi_cert_tx_ltf;
extern int esp_wifi_cert_tx_gi;
extern int esp_wifi_cert_tx_nss; //nss = 1, esp32ax only supports one spatial stream
extern int esp_wifi_cert_tx_ersu;
extern int esp_test_su_rx_mcs7;
extern int esp_test_su_rx_mcs7_retries;
extern int esp_test_su_rx_mcs6;
extern int esp_test_su_rx_mcs6_retries;
#endif

#if CONFIG_IDF_TARGET_ESP32C6
extern int esp_test_mu_rx_mcs7;
extern int esp_test_mu_rx_mcs6;
extern int esp_test_he_rx_noeb;
#endif

#if CONFIG_IDF_TARGET_ESP32C6 && !CONFIG_IDF_ENV_FPGA
extern int esp_test_ethernet_input;
extern int esp_test_ip4_input;
extern int esp_test_ip6_input;
extern int esp_test_etharp_input;

extern int esp_test_ip_input;
extern int esp_test_ip_ip4_input;
extern int esp_test_ip_ip6_input;

extern int esp_test_tcpip_input;

extern int esp_test_ethernet_output;

extern int esp_test_ethernet_output_etharp_update_arp_entry;
extern int esp_test_ethernet_output_etharp_output_to_arp_index;
extern int esp_test_ethernet_output_etharp_output;
extern int esp_test_ethernet_output_etharp_query;
extern int esp_test_ethernet_output_etharp_raw;
#endif

#if ESP_WIFI_HAL_TEST
extern bool esp_test_dump_su;
extern bool esp_test_dump_mu;
extern bool esp_test_dump_rfo;
extern int esp_test_tb_enable_ac;
#endif

#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
extern int g_update_cfo_threshold;

extern esp_err_t esp_wifi_cert_set_mcs(int mcs);
extern esp_err_t esp_wifi_cert_set_bcc(bool enable);
extern esp_err_t esp_wifi_cert_set_ltf(float ltf_us); //only accept 3.2, 6.4 and 12.8
extern esp_err_t esp_wifi_cert_set_gi(float gi_us); //only accept 0.8, 1.6 and 3.2
extern esp_err_t esp_wifi_cert_set_nss(int nss);
extern esp_err_t esp_wifi_cert_set_ersu(int ersu);
#endif
/*******************************************************
 *                Function Declarations
 *******************************************************/
#if CONFIG_IDF_TARGET_ESP32C6
extern esp_err_t esp_wifi_cert_powersave(bool on);
#endif
// /* debug */
// #ifdef CONFIG_SOC_WIFI_HE_SUPPORT
// extern void register_wifi_cmd(void);
// extern void register_wifi_stats(void);
// extern void register_wifi_itwt(void);
// #endif
/*******************************************************
 *                Function Definitions
 *******************************************************/

bool str_2_mac(uint8_t *str, uint8_t *dest)
{
    uint8_t loop = 0;
    uint8_t tmp = 0;
    uint8_t *src_p = str;

    if (strlen((char *)src_p) != 17) {
        // must be like 12:34:56:78:90:AB
        return false;
    }

    for (loop = 0; loop < 17 ; loop++) {
        if (loop % 3 == 2) {
            if (src_p[loop] != ':') {
                return false;
            }
            continue;
        }

        if ((src_p[loop] >= '0') && (src_p[loop] <= '9')) {
            tmp = tmp * 16 + src_p[loop] - '0';
        } else if ((src_p[loop] >= 'A') && (src_p[loop] <= 'F')) {
            tmp = tmp * 16 + src_p[loop] - 'A' + 10;
        } else if ((src_p[loop] >= 'a') && (src_p[loop] <= 'f')) {
            tmp = tmp * 16 + src_p[loop] - 'a' + 10;
        } else {
            return false;
        }

        if (loop % 3 == 1) {
            *dest++ = tmp;
            tmp = 0;
        }
    }

    return true;
}

static void scan_done_handler(void* arg, esp_event_base_t event_base,
                              int32_t event_id, void* event_data)
{
    uint16_t sta_number = 0;
    uint8_t i;
    wifi_ap_record_t *ap_list_buffer;

    esp_wifi_scan_get_ap_num(&sta_number);
    if (!sta_number) {
        ESP_LOGE(TAG, "No AP found");
        return;
    }
    ap_list_buffer = malloc(sta_number * sizeof(wifi_ap_record_t));
    if (ap_list_buffer == NULL) {
        ESP_LOGE(TAG, "Failed to malloc buffer to print scan results");
        return;
    }

    if (esp_wifi_scan_get_ap_records(&sta_number, (wifi_ap_record_t *) ap_list_buffer) == ESP_OK) {
        for (i = 0; i < sta_number; i++) {
#if CONFIG_IDF_TARGET_ESP32C6
            if (ap_list_buffer[i].phy_11ax) {
                ESP_LOGW(TAG,
                         "[%2d][%32s][rssi=%2d]authmode:0x%x, channel:%2d[%d], phymode:%4s, "MACSTR", bssid-index:%d, bss_color:%d, disabled:%d",
                         i, ap_list_buffer[i].ssid, ap_list_buffer[i].rssi, ap_list_buffer[i].authmode,
                         ap_list_buffer[i].primary, ap_list_buffer[i].second,
                         ap_list_buffer[i].phy_11ax ? "11ax" : (ap_list_buffer[i].phy_11n ? "11n" :
                                                                (ap_list_buffer[i].phy_11g ? "11g" : (ap_list_buffer[i].phy_11b ? "11b" : ""))),
                         MAC2STR(ap_list_buffer[i].bssid), ap_list_buffer[i].he_ap.bssid_index,
                         ap_list_buffer[i].he_ap.bss_color, ap_list_buffer[i].he_ap.bss_color_disabled);
            } else {
                ESP_LOGI(TAG,
                         "[%2d][%32s][rssi=%2d]authmode:0x%x, channel:%2d[%d], phymode:%4s, "MACSTR"",
                         i, ap_list_buffer[i].ssid, ap_list_buffer[i].rssi, ap_list_buffer[i].authmode,
                         ap_list_buffer[i].primary, ap_list_buffer[i].second,
                         ap_list_buffer[i].phy_11ax ? "11ax" : (ap_list_buffer[i].phy_11n ? "11n" :
                                                                (ap_list_buffer[i].phy_11g ? "11g" : (ap_list_buffer[i].phy_11b ? "11b" : ""))),
                         MAC2STR(ap_list_buffer[i].bssid));
            }
#else //5.72.1, report scan results
//             ESP_LOGI(TAG,
//                     "[%32s][rssi=%2d]authmode:0x%x, channel:%2d, "MACSTR", phymode:%4s, bss_color:%d, disabled:%d",
//                     ap_list_buffer[i].ssid, ap_list_buffer[i].rssi, ap_list_buffer[i].authmode,
//                     ap_list_buffer[i].primary, MAC2STR(ap_list_buffer[i].bssid),
//                     ap_list_buffer[i].phy_11ax ? "11ax" : (ap_list_buffer[i].phy_11n ? "11n" : (ap_list_buffer[i].phy_11g ? "11g" : (ap_list_buffer[i].phy_11b ? "11b" : ""))),
//                     ap_list_buffer[i].bss_color, ap_list_buffer[i].bss_color_disabled);
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
            ESP_LOGI(TAG,
                     "[%32s][rssi=%2d]authmode:0x%x, channel:%2d, "MACSTR", phymode:%4s",
                     ap_list_buffer[i].ssid, ap_list_buffer[i].rssi, ap_list_buffer[i].authmode,
                     ap_list_buffer[i].primary, MAC2STR(ap_list_buffer[i].bssid),
                     ap_list_buffer[i].phy_11ax ? "11ax" : (ap_list_buffer[i].phy_11n ? "11n" : (ap_list_buffer[i].phy_11g ? "11g" : (ap_list_buffer[i].phy_11b ? "11b" : ""))));
#else
            ESP_LOGI(TAG,
                     "[%32s][rssi=%2d]authmode:0x%x, channel:%2d, "MACSTR", phymode:%4s",
                     ap_list_buffer[i].ssid, ap_list_buffer[i].rssi, ap_list_buffer[i].authmode,
                     ap_list_buffer[i].primary, MAC2STR(ap_list_buffer[i].bssid),
                     ap_list_buffer[i].phy_11n ? "11n" : (ap_list_buffer[i].phy_11g ? "11g" : (ap_list_buffer[i].phy_11b ? "11b" : "")));
#endif
#endif
        }
    }
    free(ap_list_buffer);
    ESP_LOGI(TAG, "sta scan done");
}

static void disconnect_handler(void* arg, esp_event_base_t event_base,
                               int32_t event_id, void* event_data)
{
    if (reconnect && (g_roaming == true)) {
        ESP_LOGI(TAG, "sta disconnect, reconnect...");
        esp_wifi_connect();
    } else {
        ESP_LOGI(TAG, "sta disconnect");
    }
    xEventGroupClearBits(wifi_event_group, CONNECTED_BIT);
    xEventGroupClearBits(wifi_event_group, GOT_IP_BIT);
    xEventGroupSetBits(wifi_event_group, DISCONNECTED_BIT);
}

static void connect_handler(void* arg, esp_event_base_t event_base,
                            int32_t event_id, void* event_data)
{
    uint8_t protocol_bitmap = 0;
    esp_wifi_get_protocol(0, &protocol_bitmap);
    wifi_event_sta_connected_t *sta_connected = (wifi_event_sta_connected_t *) event_data;
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
    ESP_LOGW(TAG, "SYSTEM_EVENT_STA_CONNECTED, "MACSTR", channel:%d, protocol:0x%x, cfo_threshold:%d",
             MAC2STR(sta_connected->bssid), sta_connected->channel, protocol_bitmap,
             g_update_cfo_threshold);
#else
    ESP_LOGW(TAG, "SYSTEM_EVENT_STA_CONNECTED, "MACSTR", channel:%d, protocol:0x%x",
             MAC2STR(sta_connected->bssid), sta_connected->channel, protocol_bitmap);
#endif
#if CONFIG_PM_ENABLE
    int itwt_flow_id = 0xff;
    esp_err_t err = esp_wifi_sta_itwt_setup(TWT_SUGGEST, 1, 0, 255, 10, 512, &itwt_flow_id);
    ESP_LOGW(TAG, "(itwt)setup(suggest), flow_type:%d, flow_id:%d, err:0x%x", 0, itwt_flow_id, err);
    // esp_test_disable_edca_tx(2); //TXQ2
#endif

#if HAL_MAC_ESP32AX_SIGMA_5_53_1
    /*
     * case 5.53.1 BSRP, create 8 agreements
     */
    esp_test_sta_create_addba(0); //BE
    esp_test_sta_create_addba(1);
    esp_test_sta_create_addba(2);
    esp_test_sta_create_addba(3);
    esp_test_sta_create_addba(4);
    esp_test_sta_create_addba(5); //VI
    esp_test_sta_create_addba(6);
    esp_test_sta_create_addba(7); //VO
#endif

#if 0
#if HAL_MAC_ESP32AX_760
    // Force Gain
    SET_PERI_REG_MASK(0x6001c02c, (BIT(23)));
    // -rssi -15
    SET_PERI_REG_BITS(0x6001c02c, 0xFF, 15, 24);//20
#endif
#endif
#if HAL_MAC_ESP32AX_760
    // BB ABORT when UNSUPPORT RATE and SIGB CRC ERR
    SET_PERI_REG_MASK(0x6001cd78, (BIT(8)) | (BIT(9)));
    // Ignore LDPC ERR
    CLEAR_PERI_REG_MASK(0x6001cd78, (BIT(22)));

    CLEAR_PERI_REG_MASK(0x6001cd80, (BIT(25)));
    SET_PERI_REG_MASK(0x6001d0a8, (BIT(20)));
#endif

    xEventGroupClearBits(wifi_event_group, DISCONNECTED_BIT);
    xEventGroupSetBits(wifi_event_group, CONNECTED_BIT);
}

uint16_t bf_matrices_intel[192] = {

    /* 0   - 191 SU */
    /* 192 - 223 MU, delta SNR */
    /* 224 - 228 CQI */
    /* 229       */
    0x8621, 0x71d7, 0x616a, 0x8a20, 0x71d6, 0x616a, 0x8a20, 0x6fd4, 0x6368, 0x8c21,
    0x71d4, 0x6169, 0x8c21, 0x71d3, 0x6168, 0x8a21, 0x6dd2, 0x6167, 0x8c1f,
    0x6fd3, 0x6168, 0x8e22, 0x6fd4, 0x636a, 0x8e20, 0x71d3, 0x6168, 0x8e20,
    0x71d1, 0x6368, 0x8e20, 0x71d1, 0x6168, 0x8c20, 0x71d2, 0x6168, 0x9021,
    0x71d4, 0x6369, 0x8e1f, 0x75d0, 0x6166, 0x9022, 0x73d0, 0x6366, 0x9021,
    0x75cf, 0x6165, 0x8e21, 0x71cf, 0x6166, 0x9021, 0x71cf, 0x6164, 0x9220,
    0x71ce, 0x6165, 0x9420, 0x6fcc, 0x6361, 0x9621, 0x6fcc, 0x6164, 0x9220,
    0x6fcd, 0x6163, 0x961f, 0x6fcd, 0x5f63, 0x941f, 0x6fcd, 0x6164, 0x961f,
    0x6fcd, 0x6363, 0x941f, 0x6fcc, 0x6162, 0x961f, 0x6fcc, 0x6162, 0x961d,
    0x6fcc, 0x6161, 0x961d, 0x6fcd, 0x6161, 0x981e, 0x71cc, 0x6160, 0x981f,
    0x6fcb, 0x615f, 0x981f, 0x6fcb, 0x6161, 0x961d, 0x6fcb, 0x615f, 0x961d,
    0x6fcb, 0x615d, 0x981d, 0x6dca, 0x615d, 0x9a1c, 0x6dcb, 0x615e, 0x9c1d,
    0x6dca, 0x615c, 0x981d, 0x6bca, 0x615a, 0x9a1b, 0x6bca, 0x5f5c, 0x9c1a,
    0x6bca, 0x5f59, 0x9c1a, 0x6bcb, 0x615a, 0x9e19, 0x6bca, 0x6158, 0x9c1b,
    0x6bca, 0x6159, 0x9c1a, 0x6bca, 0x5f56, 0x9c19, 0x6bcb, 0x5f58, 0x9c1b,
    0x6bcc, 0x5f57, 0x9e1a, 0x6bcb, 0x6156, 0x9e19, 0x6bca, 0x6153, 0x9c19,
    0x6bc9, 0x5f52, 0x9c19, 0x69c9, 0x5f51, 0x9c17, 0x6bca, 0x5f50, 0xa018,
    0x69cb, 0x5f50, 0xa018, 0x6bca, 0x5f4f, 0xa017, 0x69c9, 0x5f4d, 0xa019,
    0x69cb, 0x5d4e, 0xa017, 0x69ca, 0x5d4e, 0xa016, 0x67cb, 0x5f4d, 0xa218,
    0x69cc, 0x5d4e, 0xa018, 0x69cc, 0x5d4c, 0xa017, 0x6bcc, 0x5d4a, 0xa017,
    0x6bcc, 0x5b4b, 0xa016, 0x6bcd, 0x5d4a, 0xa216, 0x6bcd, 0x5d48, 0xa014,
    0x6dcd, 0x5d47,
};

#if HAL_MAC_ESP32AX_760
#define BB_BF_CTRL1_REG        (0x6001d0a8)
#define MAC_BF_CONFIG_REG      (0x60033c74)
#define BF_START               (0x60034c00)
#else
#define BB_BF_CTRL1_REG        (0x6004a0a8)
#define MAC_BF_CONFIG_REG      (0x60047c74)
#define BF_START               (0x60048c00)
#endif

static int wifi_sw_write_rpt(int argc, char **argv)
{
    ESP_LOGI(TAG, "BB_BF_CTRL1_REG  (0x%08x): 0x%lx", BB_BF_CTRL1_REG, REG_READ(BB_BF_CTRL1_REG));
    ESP_LOGI(TAG, "MAC_BF_CONFIG_REG(0x%08x): 0x%lx", MAC_BF_CONFIG_REG, REG_READ(MAC_BF_CONFIG_REG));
    ESP_LOGI(TAG, "BF_START         (0x%08x): 0x%lx", BF_START, REG_READ(BF_START));
    SET_PERI_REG_MASK(BB_BF_CTRL1_REG, (BIT(15)));
    SET_PERI_REG_MASK(MAC_BF_CONFIG_REG, (BIT(0)));
    int i = 0;
    int offset = 0;
    for (i = 0; i < 192; i++) {
        REG_WRITE(BF_START + (offset + 4 * i), bf_matrices_intel[i]);
    }
    ESP_LOGW(TAG, "BB_BF_CTRL1_REG  (0x%08x): 0x%lx", BB_BF_CTRL1_REG, REG_READ(BB_BF_CTRL1_REG));
    ESP_LOGW(TAG, "MAC_BF_CONFIG_REG(0x%08x): 0x%lx", MAC_BF_CONFIG_REG, REG_READ(MAC_BF_CONFIG_REG));
    ESP_LOGW(TAG, "BF_START         (0x%08x): 0x%lx", BF_START, REG_READ(BF_START));
    return 0;
}

static int wifi_hw_write_rpt(int argc, char **argv)
{
    CLEAR_PERI_REG_MASK(BB_BF_CTRL1_REG, (BIT(15)));
    ESP_LOGW(TAG, "BB_BF_CTRL1_REG  (0x%08x): 0x%lx", BB_BF_CTRL1_REG, REG_READ(BB_BF_CTRL1_REG));
    ESP_LOGW(TAG, "MAC_BF_CONFIG_REG(0x%08x): 0x%lx", MAC_BF_CONFIG_REG, REG_READ(MAC_BF_CONFIG_REG));
    ESP_LOGW(TAG, "BF_START         (0x%08x): 0x%lx", BF_START, REG_READ(BF_START));
    return 0;
}

static void got_ip_handler(void* arg, esp_event_base_t event_base,
                           int32_t event_id, void* event_data)
{
    xEventGroupClearBits(wifi_event_group, DISCONNECTED_BIT);
    xEventGroupSetBits(wifi_event_group, GOT_IP_BIT);
}

static void connected_handler(void* esp_netif, esp_event_base_t event_base,
                              int32_t event_id, void* event_data)
{
    esp_netif_create_ip6_linklocal(esp_netif);
}

/* types of ipv6 addresses to be displayed on ipv6 events */
static const char *s_ipv6_addr_types[] = {
    "ESP_IP6_ADDR_IS_UNKNOWN",
    "ESP_IP6_ADDR_IS_GLOBAL",
    "ESP_IP6_ADDR_IS_LINK_LOCAL",
    "ESP_IP6_ADDR_IS_SITE_LOCAL",
    "ESP_IP6_ADDR_IS_UNIQUE_LOCAL",
    "ESP_IP6_ADDR_IS_IPV4_MAPPED_IPV6"
};

static void got_ip6_handler(void* arg, esp_event_base_t event_base,
                            int32_t event_id, void* event_data)
{
    ip_event_got_ip6_t *event = (ip_event_got_ip6_t *)event_data;

    if (event->esp_netif == netif_sta) {
        esp_ip6_addr_type_t ipv6_type = esp_netif_ip6_get_addr_type(&event->ip6_info.ip);
        ESP_LOGI(TAG, "STA Interface ready [IPv6 - "IPV6STR", type - %s]", IPV62STR(event->ip6_info.ip), s_ipv6_addr_types[ipv6_type]);
    }
}

static int str_compare(uint8_t* str, int str_len, uint8_t str_cmp)
{
    int i = 0, cnt = 0;
    if (str_len == 0 && str == NULL) {
        return cnt;
    }

    for(i = str_len; i > 0; i--) {
        if (str[i - 1] == str_cmp) {
            ++cnt;
        } else {
            return cnt;
        }
    }

    return cnt;
}
static void bcn_timeout_handler(void *arg, esp_event_base_t event_base, int32_t event_id,
        void *event_data)
{
    static int bcn_to_count = 0;
    uint16_t secs = 0;
    ESP_ERROR_CHECK(esp_wifi_get_inactive_time(ESP_IF_WIFI_STA, &secs));
    ESP_LOGW(TAG, "(count: %d)WIFI_EVENT_STA_BEACON_TIMEOUT, %d secs", bcn_to_count ++, secs);
}
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
const char* itwt_probe_status_to_str(wifi_itwt_probe_status_t status)
{
    switch (status) {
        case ITWT_PROBE_FAIL:                 return "itwt probe fail";
        case ITWT_PROBE_SUCCESS:              return "itwt probe success";
        case ITWT_PROBE_TIMEOUT:              return "itwt probe timeout";
        case ITWT_PROBE_STA_DISCONNECTED:     return "Sta disconneced";
        default:                              return "Unknown status";
    }
}

static void itwt_probe_handler(void *arg, esp_event_base_t event_base, int32_t event_id,
        void *event_data)
{
    wifi_event_sta_itwt_probe_t *evt = (wifi_event_sta_itwt_probe_t *) event_data;
    ESP_LOGW(TAG, "(itwt)WIFI_EVENT_ITWT_PROBE, status: %s, reason: %d", itwt_probe_status_to_str(evt->status), evt->reason);
}

static void itwt_information_handler(void *arg, esp_event_base_t event_base, int32_t event_id,
        void *event_data)
{
    wifi_event_sta_itwt_suspend_t *evt = (wifi_event_sta_itwt_suspend_t *) event_data;
    ESP_LOGW(TAG, "(itwt)WIFI_EVENT_ITWT_SUSPEND, status:%d, flow_id_bitmap:0x%x, actual_suspend_time_ms:[%lu %lu %lu %lu %lu %lu %lu %lu]",
             evt->status, evt->flow_id_bitmap,
             evt->actual_suspend_time_ms[0], evt->actual_suspend_time_ms[1], evt->actual_suspend_time_ms[2], evt->actual_suspend_time_ms[3],
             evt->actual_suspend_time_ms[4], evt->actual_suspend_time_ms[5], evt->actual_suspend_time_ms[6], evt->actual_suspend_time_ms[7]);
}
#endif
#if CONFIG_IDF_TARGET_ESP32C6
static void s_pmu_internal_signal_map_to_io(void)
{
    /**
     * No.      Name
     * 26       PMU HP MODEM
     * 28       PMU HP ACTIVE
     * 45       REGDMA work
     * 59       PMU SOC wakeup
     * 92       Modem SOC Wakeup
     * 93       Modem Sleep REQ
     * 94       RF flag
     * 95       RX Frame
     */
    /* map PMU MODEM signal to GPIO0 */
    *(volatile uint32_t *)LP_AON_GPIO_MUX_REG |= BIT(0);
    *(volatile uint32_t *)LP_IO_DEBUG_SEL0_REG |= (26 & 0x7f);
    *(volatile uint32_t *)LP_IO_GPIO0_REG |= ((2 & 0x7) << 12);
    *(volatile uint32_t *)LP_CLKRST_LP_CLK_PO_EN_REG = 0xffffffff;

    /* map PMU ACTIVE signal to GPIO1 */
    *(volatile uint32_t *)LP_AON_GPIO_MUX_REG |= BIT(1);
    *(volatile uint32_t *)LP_IO_DEBUG_SEL0_REG |= (28 & 0x7f) << 7;
    *(volatile uint32_t *)LP_IO_GPIO1_REG |= ((2 & 0x7) << 12);
    *(volatile uint32_t *)LP_CLKRST_LP_CLK_PO_EN_REG = 0xffffffff;

    /* map PMU SOC wakeup signal to GPIO2 */
    *(volatile uint32_t *)LP_AON_GPIO_MUX_REG |= BIT(2);
    *(volatile uint32_t *)LP_IO_DEBUG_SEL0_REG |= (27 & 0x7f) << 14;
    *(volatile uint32_t *)LP_IO_GPIO2_REG |= ((2 & 0x7) << 12);
    *(volatile uint32_t *)LP_CLKRST_LP_CLK_PO_EN_REG = 0xffffffff;

    /* map REGDMA work signal to GPIO3 */
    *(volatile uint32_t *)LP_AON_GPIO_MUX_REG |= BIT(3);
    *(volatile uint32_t *)LP_IO_DEBUG_SEL0_REG |= (95 & 0x7f) << 21;
    *(volatile uint32_t *)LP_IO_GPIO3_REG |= ((2 & 0x7) << 12);
    *(volatile uint32_t *)LP_CLKRST_LP_CLK_PO_EN_REG = 0xffffffff;

    extern void tsf_hal_map_tbtt_target_to_rx_frame(void);
    tsf_hal_map_tbtt_target_to_rx_frame();
}
#endif

void initialise_light_sleep(void)
{
#if CONFIG_IDF_TARGET_ESP32C6
    extern void sleep_peripherals_retention_init(void);
    sleep_peripherals_retention_init();
#endif

#if CONFIG_PM_ENABLE
    // Configure dynamic frequency scaling:
    // maximum and minimum frequencies are set in sdkconfig,
    // automatic light sleep is enabled if tickless idle support is enabled.
#if CONFIG_IDF_TARGET_ESP32
    esp_pm_config_esp32_t pm_config = {
#elif CONFIG_IDF_TARGET_ESP32S2
    esp_pm_config_esp32s2_t pm_config = {
#elif CONFIG_IDF_TARGET_ESP32C3
    esp_pm_config_esp32c3_t pm_config = {
#elif CONFIG_IDF_TARGET_ESP32S3
    esp_pm_config_esp32s3_t pm_config = {
#elif CONFIG_IDF_TARGET_ESP32C2
    esp_pm_config_esp32c2_t pm_config = {
#elif CONFIG_IDF_TARGET_ESP32C6
    esp_pm_config_esp32c6_t pm_config = {
#endif
            .max_freq_mhz = 80,
#if CONFIG_IDF_ENV_FPGA
            .min_freq_mhz = 80,
#else
            .min_freq_mhz = 10,
#endif
#if CONFIG_FREERTOS_USE_TICKLESS_IDLE
            .light_sleep_enable = true
#endif
    };
    ESP_ERROR_CHECK( esp_pm_configure(&pm_config) );
#endif // CONFIG_PM_ENABLE

#if CONFIG_IDF_TARGET_ESP32C6
    // s_soft_gpio_toggle_signal_to_io();
    s_pmu_internal_signal_map_to_io();
#endif
}


static esp_err_t find_cert_in_partition(const char *cert_name, uint8_t **cert_pem, unsigned int *cert_len)
{
    esp_err_t ret = ESP_OK;
    const esp_partition_t* wpa2_part = esp_partition_find_first(ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_ANY, cert_name);

    uint8_t *tem_buf = NULL;
    uint32_t wpa2_pki_len = 0;

    if (!wpa2_part) {
        ESP_LOGE(TAG, "failed to find partition");
        return ESP_FAIL;
    }
    tem_buf = malloc(CERT_MAX_LENGTH);
    if (!tem_buf) {
        ESP_LOGE(TAG, "Malloc failure");
        return ESP_FAIL;
    }
    if (ESP_OK != esp_partition_read(wpa2_part, 0, tem_buf, CERT_MAX_LENGTH)) {
        ESP_LOGE(TAG,"read parition fail");
        return ESP_FAIL;
    }
    wpa2_pki_len = CERT_MAX_LENGTH - str_compare(tem_buf, CERT_MAX_LENGTH, 0xff);
    *cert_pem = malloc(wpa2_pki_len + 1);
    if (*cert_pem == NULL) {
        ESP_LOGE(TAG, "Malloc failure");
        ret = ESP_FAIL;
    } else {
        memcpy(*cert_pem, tem_buf, wpa2_pki_len);
        (*cert_pem)[wpa2_pki_len] = '\0';
        ESP_LOGI(TAG, "certs: %s\n", *cert_pem);
    }
    free(tem_buf);
    *cert_len = wpa2_pki_len;

    return ret;
}

static esp_err_t set_ca_certs(const char* cert_name)
{
    uint8_t *ca_pem;
    unsigned int ca_len = 0;
    esp_err_t ret = ESP_FAIL;
    /* Before adding to this if-else ladder ensure that the string used to denote
     *  to the certificate does not overlap with an existing one.*/
    if(strstr(cert_name,"caswin")){
        ca_pem = caswin_pem_start;
        ca_len = caswin_pem_end - caswin_pem_start; 
        ESP_LOGI(TAG, "CA cert: caswin");
    } else if (strstr(cert_name, "cas")) {
        ca_pem = cas_pem_start;
        ca_len = cas_pem_end - cas_pem_start;
                ESP_LOGI(TAG, "CA cert: cas");
    } else if (strstr(cert_name, "rsa_ca1_w1_fi")) {
        ca_pem = ca1_pem_start;
        ca_len = ca1_pem_end - ca1_pem_start;
        ESP_LOGI(TAG, "CA cert: rsa_ca1_w1_fi");
    } else if (strstr(cert_name, "rsa_ca2_w1_fi")) {
        ca_pem = ca2_pem_start;
        ca_len = ca2_pem_end - ca2_pem_start;
        ESP_LOGI(TAG, "CA cert: rsa_ca2_w1_fi");
    } else if (strstr(cert_name, "rsa3072-ca")) {
        ca_pem = rsa3072_ca_pem_start;
        ca_len = rsa3072_ca_pem_end - rsa3072_ca_pem_start;
        ESP_LOGI(TAG, "CA cert: rsa3072-ca");
    } else if (strstr(cert_name, "ec2-ca")) {
        ca_pem = ec2_ca_pem_start;
        ca_len = ec2_ca_pem_end - ec2_ca_pem_start;
        ESP_LOGI(TAG, "CA cert: ec2-ca");
    } else {
        ESP_LOGE(TAG, "Failed to find certificate embedded, trying to search in partitions");
	    if (find_cert_in_partition(cert_name, &ca_pem, &ca_len) != ESP_OK) {
            ESP_LOGE(TAG, "Failed to find certificate partition");
        }
    }

    if (ca_len)
        ret = esp_eap_client_set_ca_cert(ca_pem, ca_len);

    ESP_LOGI(TAG, "SetCA Status: %d", ret);

    return ret;
}

static esp_err_t set_user_certs(const char* cert_name)
{
    uint8_t *user_pem;
    unsigned int user_len = 0;
    esp_err_t ret = ESP_FAIL;
    /* Before adding to this if-else ladder ensure that the string used to denote
     *  to the certificate does not overlap with an existing one.*/
    if(strstr(cert_name, "wifiuserwin")){
        user_pem = winuser_crt_start;
        user_len = winuser_crt_end - winuser_crt_start;
        ESP_LOGI(TAG, "Client cert: wifiuserwin");
    } else if (strstr(cert_name,"wifiuser")) {
        user_pem = wifiuser_crt_start;
        user_len = wifiuser_crt_end - wifiuser_crt_start;
        ESP_LOGI(TAG, "Client cert: wifiuser");
    }/* else if (strstr(cert_name, "ec2-user-p256")) {
        user_pem = ec2_user_p256_crt_start;
        user_len = ec2_user_p256_crt_end - ec2_user_p256_crt_start;
    } */else if (strstr(cert_name, "user-p256")) {
        user_pem = user_p256_crt_start;
        user_len = user_p256_crt_end - user_p256_crt_start;
        ESP_LOGI(TAG, "Client cert: user-p256");
    } else if (strstr(cert_name, "rsa_user1_w1_fi")) {
        user_pem = user1_crt_start;
        user_len = user1_crt_end - user1_crt_start;
        ESP_LOGI(TAG, "Client cert: rsa_user1_w1_fi");
    } else if (strstr(cert_name, "rsa3072-user-rsa2048")) {
        user_pem = rsa3072_user_rsa2048_crt_start;
        user_len = rsa3072_user_rsa2048_crt_end - rsa3072_user_rsa2048_crt_start;
        ESP_LOGI(TAG, "Client cert: rsa3072-user-rsa2048");
    } else if (strstr(cert_name, "rsa3072-user")) {
        user_pem = rsa3072_user_crt_start;
        user_len = rsa3072_user_crt_end - rsa3072_user_crt_start;
        ESP_LOGI(TAG, "Client cert: rsa3072-user");
    } else if (strstr(cert_name, "ec2-user")) {
        user_pem = ec2_user_crt_start;
        user_len = ec2_user_crt_end - ec2_user_crt_start;
        ESP_LOGI(TAG, "Client cert: ec2-user");
    } else {
        ESP_LOGE(TAG, "Failed to find certificate embedded, trying to search in partitions");
	if (find_cert_in_partition(cert_name, &user_pem, &user_len) != ESP_OK) {
            ESP_LOGE(TAG, "Failed to find certificate partition");
        }
    }
    printf("\nCert_name : %s\n",cert_name);
    ret = esp_eap_client_set_certificate_and_key(user_pem, user_len, user_pem, user_len, NULL, 0);
    ESP_LOGI(TAG, "SetClient Status: %d", ret);

    return ret;
}

void initialise_wifi(void)
{
    //esp_log_level_set("wifi", ESP_LOG_WARN);
    static bool initialized = false;

    if (initialized) {
        return;
    }
    ESP_ERROR_CHECK(esp_netif_init());
    wifi_event_group = xEventGroupCreate();
    ESP_ERROR_CHECK( esp_event_loop_create_default() );
    netif_ap = esp_netif_create_default_wifi_ap();
    assert(netif_ap);
    netif_sta = esp_netif_create_default_wifi_sta();
    assert(netif_sta);
    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    ESP_ERROR_CHECK( esp_wifi_init(&cfg) );
    ESP_ERROR_CHECK( esp_event_handler_register(WIFI_EVENT, WIFI_EVENT_SCAN_DONE, &scan_done_handler, NULL) );
    ESP_ERROR_CHECK( esp_event_handler_register(WIFI_EVENT, WIFI_EVENT_STA_CONNECTED, &connected_handler, netif_sta) );
    ESP_ERROR_CHECK( esp_event_handler_register(WIFI_EVENT, WIFI_EVENT_STA_DISCONNECTED, &disconnect_handler, NULL) );
    ESP_ERROR_CHECK( esp_event_handler_register(WIFI_EVENT, WIFI_EVENT_STA_CONNECTED, &connect_handler, NULL) );
    ESP_ERROR_CHECK(esp_event_handler_register(WIFI_EVENT,WIFI_EVENT_STA_BEACON_TIMEOUT,&bcn_timeout_handler,NULL));
    ESP_ERROR_CHECK( esp_event_handler_register(IP_EVENT, IP_EVENT_STA_GOT_IP, &got_ip_handler, NULL) );
    ESP_ERROR_CHECK( esp_event_handler_register(IP_EVENT, IP_EVENT_GOT_IP6, &got_ip6_handler, NULL) );
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT    
    ESP_ERROR_CHECK(esp_event_handler_register(WIFI_EVENT,WIFI_EVENT_ITWT_PROBE,&itwt_probe_handler,NULL));
    ESP_ERROR_CHECK(esp_event_handler_register(WIFI_EVENT,WIFI_EVENT_ITWT_SUSPEND,&itwt_information_handler,NULL));
#endif
    ESP_ERROR_CHECK( esp_wifi_set_storage(WIFI_STORAGE_RAM) );
    ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_STA));
    ESP_ERROR_CHECK( esp_wifi_internal_set_spp_amsdu(WIFI_IF_STA, false, false) );
#if 1
    ESP_ERROR_CHECK(esp_wifi_set_ps(WIFI_PS_NONE));
#else
    // 5.72.1 and 5.60.1 enable power save
    ESP_ERROR_CHECK(esp_wifi_set_ps(WIFI_PS_MIN_MODEM));
#endif

#if 0
    //wpa2 5.2.3
    extern esp_err_t esp_wifi_internal_set_spp_amsdu(wifi_interface_t ifidx, bool spp_cap, bool spp_req);
    esp_wifi_internal_set_spp_amsdu(0, 0, 0);
#endif

#if 0
    //ESP_ERROR_CHECK(esp_wifi_set_protocol(0, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N | WIFI_PROTOCOL_11AX));
//#else
    ESP_ERROR_CHECK(esp_wifi_set_protocol(0, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N));
    ESP_ERROR_CHECK(esp_wifi_set_bandwidth(0, WIFI_BW_HT40));
#endif
#if CONFIG_IDF_ENV_FPGA
    // DUT
#if CONFIG_IDF_TARGET_ESP32C6
    uint8_t dut_mac[6] = { 0x00, 0x11, 0x22, 0x22, 0x11, 0x33 };
#else
    uint8_t dut_mac[6] = { 0x00, 0x11, 0x22, 0x22, 0x11, 0x34 };
#endif
    ESP_ERROR_CHECK(esp_wifi_set_mac(ESP_IF_WIFI_STA, dut_mac));
#endif

#if 0// STA1
    uint8_t sta1_mac[6] = { 0x00, 0x11, 0x22, 0x33, 0x44, 0x67 };
    ESP_ERROR_CHECK(esp_wifi_set_mac(ESP_IF_WIFI_STA, sta1_mac));
#endif

    ESP_ERROR_CHECK( esp_wifi_start() );

#if 0 // This will be set by Sigma at run time
    wifi_ioctl_config_t cfg1;
    cfg1.data.ht2040_coex.enable = 1;
    esp_wifi_internal_ioctl(WIFI_IOCTL_SET_STA_HT2040_COEX, &cfg1);
#endif
    
#if 0
    // 5.66.1 disable UPH by default
    esp_wifi_enable_htc_uph(0);
#endif

    // create 2 agreements, tid=0 and tid=7
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
    esp_test_tx_addba_request = 1;
    esp_test_baparas_support_amsdu = 1;
#endif
#if ESP_WIFI_HAL_TEST
    esp_test_dump_mu = 0;
    esp_test_dump_rfo = 0;
#if 0
    /*
     * 5.64.1 MU EDCA Parameter Set, create tid=5(VI)
     */
    esp_test_tb_enable_ac = 1;
#endif
#endif

    // 5.44.1 and 5.61.1, u may want to disable su beamforming
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
    esp_wifi_enable_su_bmfmee(1);
#endif
#if 0
    // 5.39.1: SU Beamformee
    esp_test_set_beamforming_feedback(SU_CODEBOOK_0_NR_2);
    // 5.58.1: SU Beamformee
    wifi_sw_write_rpt(0, NULL);
    esp_test_set_beamforming_feedback(MU_CODEBOOK_1_NR_4);
#endif

#if HAL_MAC_ESP32AX_760
    /*
     * 7.6.0 workaround
     * TB ignore CCA(ED-CCA and NAV)
     * except 5.62.1: dut correctly follows the CS Required, ED sensing and NAV consideration)
     */
    CLEAR_PERI_REG_MASK(0x60033c78, (BIT(12))); //WDEV_TX_TB_CARE_CCA
    ESP_LOGW(TAG, "(mac)read 0x%x: 0x%x(tb ignore cca and nav)", 0x60033c78, REG_READ(0x60033c78));
    /*
     * 7.6.0 workaround
     * 5.61.1 Intel AP EDCA ignore NAVs
     */
    ESP_LOGW(TAG, "(mac)read 0x%x: 0x%x(edca disable intra-nav and share-nav)", 0x60033c10, REG_READ(0x60033c10));
#endif

#if !CONFIG_IDF_ENV_FPGA
    // 5.59.1, 5.58.1 CFO
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
    g_update_cfo_threshold = 0xffff; //0xffff: disable SW compensation
#endif
#endif
#if 0
    /*
     * 5.31.1 rx mcs7
     * rssi=-20, good
     * rssi=-47, too many retries
     */
    ESP_LOGW(TAG, "(bb)read 0x%x: 0x%x", 0x6001cd64, REG_READ(0x6001cd64));
    REG_WRITE(0x6001cd64, 0xd8d8d8d8);
    ESP_LOGW(TAG, "(bb)read 0x%x: 0x%x", 0x6001cd64, REG_READ(0x6001cd64));
#endif
#if 0
    dbg_dump_rx_errors();
    ESP_LOGW(TAG, "(test)Dump RX errors");
#endif
#if CONFIG_IDF_TARGET_ESP32C6
#ifdef CONFIG_ESP_WIFI_ENABLE_WIFI_RX_STATS
#ifdef CONFIG_ESP_WIFI_ENABLE_WIFI_RX_MU_STATS
    esp_wifi_enable_rx_statistics(true, true);
#else
    esp_wifi_enable_rx_statistics(true, false);
#endif
#endif
#ifdef CONFIG_ESP_WIFI_ENABLE_WIFI_TX_STATS
    esp_wifi_enable_tx_statistics(2, true);
#endif
#endif
    initialized = true;
}

static bool wifi_cmd_sta_join(const char* ssid, const char* pass, const char* pmf, const char *authmode, const char *bssid,const char* btm)
{
    int bits = xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT, 0, 1, 0);
    wifi_config_t wifi_config = { 0 };

    memcpy(wifi_config.sta.ssid, ssid, strlen(ssid));
    if (pass) {
        memcpy(wifi_config.sta.password, pass, strlen(pass));
    }
    if (authmode) {
        if ( strcasecmp(authmode,"wpa3") == 0) {
            wifi_config.sta.threshold.authmode = WIFI_AUTH_WPA3_PSK;
        }
        if ( strcasecmp(authmode,"wpa") == 0) {
            wifi_config.sta.threshold.authmode = WIFI_AUTH_WPA_PSK;
	    }
    }
if (bssid) {
        uint8_t address[6];
        STR2MAC(bssid, &address);
        ESP_LOGW(TAG, "sta set to connect bssid:" MACSTR, MAC2STR(address));
        wifi_config.sta.bssid_set = 1;
        memcpy(&wifi_config.sta.bssid, &address, sizeof(address));
    }
    #ifdef CONFIG_WPA_MBO_SUPPORT
    wifi_config.sta.rm_enabled =1;
    wifi_config.sta.btm_enabled =1;
    if ( strcasecmp(authmode,"wpa") == 0) {
		wifi_config.sta.mbo_enabled =0;
	}
    else{
        wifi_config.sta.mbo_enabled =1;
    }
    wifi_config.sta.pmf_cfg.capable = 1;
    
    if(btm){
        if(strcasecmp(btm,"disable") == 0){
            wifi_config.sta.mbo_enabled = 0;
            wifi_config.sta.rm_enabled = 0;
            wifi_config.sta.btm_enabled = 0;
        }
    }
    #endif 

    wifi_config.sta.pmf_cfg.capable = 1;
    if (pmf) {
        if ( strcasecmp(pmf,"optional") == 0) {
            wifi_config.sta.pmf_cfg.capable = true;
            wifi_config.sta.pmf_cfg.required = false;        
        } 
        if ( strcasecmp(pmf,"required") == 0) {
            wifi_config.sta.pmf_cfg.capable = true;
            wifi_config.sta.pmf_cfg.required = true;
        }
        if ( strcasecmp(pmf,"disabled") == 0) {
            wifi_config.sta.pmf_cfg.capable = false;
            wifi_config.sta.pmf_cfg.required = false;
        }
    }
    wifi_config.sta.transition_disable = 1;
    if (bits & CONNECTED_BIT) {
        reconnect = false;
        xEventGroupClearBits(wifi_event_group, CONNECTED_BIT);
        ESP_ERROR_CHECK( esp_wifi_disconnect() );
        xEventGroupWaitBits(wifi_event_group, DISCONNECTED_BIT, 0, 1, portTICK_PERIOD_MS);
    }

    reconnect = true;
    ESP_LOGI(TAG, "issuing disconnect to clean previous states");
    esp_wifi_disconnect();
    ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_STA) );
    ESP_ERROR_CHECK( esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config) );
    if (pmf) {
        if (strcasecmp(pmf,"disabled") == 0) {
            esp_wifi_disable_pmf_config(ESP_IF_WIFI_STA);
        }
    }

    ESP_ERROR_CHECK( esp_wifi_start() );
    // sleep(2);
    vTaskDelay(2000 / portTICK_PERIOD_MS);
#if CONFIG_IDF_ENV_FPGA
    uint8_t protocol_bitmap = 0;
    ESP_ERROR_CHECK( esp_wifi_get_protocol(WIFI_IF_STA, &protocol_bitmap) );
    if (protocol_bitmap & WIFI_PROTOCOL_11AX) {
        /* 11ax mode */
        uint8_t mcs_set = 0;
        ESP_ERROR_CHECK( esp_wifi_set_he_mcs(mcs_set) );
        ESP_LOGW(TAG, "(11ax)support tx/rx %s", (mcs_set == 1) ? "HE-MCS[0,9]" : "HE-MCS[0,7]");
        bool dcm_en = true;
        esp_wifi_enable_dcm(dcm_en);
        ESP_LOGW(TAG, "(11ax)enable dcm:%d", dcm_en);
    }
#endif
    ESP_LOGI(TAG, "esp_wifi_connect start, h2e:%d", wifi_config.sta.sae_pwe_h2e);
    esp_wifi_connect();
    xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT, 0, 1, 5000/portTICK_PERIOD_MS);
    return true;
}

static int wifi_cmd_sta(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &sta_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, sta_args.end, argv[0]);
        return 1;
    }
    
    if ( sta_args.pmf->count == 1 ) {
        if (strcasecmp(sta_args.pmf->sval[0],"optional") && strcasecmp(sta_args.pmf->sval[0],"required") && strcasecmp(sta_args.pmf->sval[0],"disabled")) {
             arg_print_errors(stderr, sta_args.end, argv[0]);
             return 1;
        }
    }

    g_roaming = true;
    if(sta_args.roaming->count != 0){
        if(strcasecmp(sta_args.roaming->sval[0],"disable") == 0){
            g_roaming = false;
        }
    }

    ESP_LOGI(TAG, "sta connecting to '%s'", sta_args.ssid->sval[0]);
            
   
    ESP_LOGI(TAG, "security %s", sta_args.security->sval[0]);
    if ( strcasecmp( sta_args.security->sval[0], WPA2_EAP) == 0 ) { 

#if 1
#if CONFIG_SOC_WIFI_HE_SUPPORT
    if (sta_args.addba_reject->count == 1) {
        esp_test_tx_addba_request = 0;
        ESP_LOGI(TAG, "ADDBA Reject");
    }
#endif
#endif

        wifi_init_config_t config = WIFI_INIT_CONFIG_DEFAULT();
        ESP_ERROR_CHECK( esp_wifi_init(&config) );
        
	    wifi_config_t wifi_config = { 0 };
	    strlcpy((char*) wifi_config.sta.ssid, sta_args.ssid->sval[0], sizeof(wifi_config.sta.ssid));
        
	#ifdef CONFIG_WPA_MBO_SUPPORT
	    wifi_config.sta.rm_enabled =1;
    	    wifi_config.sta.btm_enabled =1;
            wifi_config.sta.mbo_enabled =1;
            wifi_config.sta.pmf_cfg.capable = 1;
        

        #endif
	 
        wifi_config.sta.pmf_cfg.capable = 1;
	if ( sta_args.pmf->count == 1 ) {
            if ( strcasecmp(sta_args.pmf->sval[0],"optional") == 0) {
                wifi_config.sta.pmf_cfg.capable = true;
                wifi_config.sta.pmf_cfg.required = false;        
            } 
            if ( strcasecmp(sta_args.pmf->sval[0],"required") == 0) {
                wifi_config.sta.pmf_cfg.capable = true;
                wifi_config.sta.pmf_cfg.required = true;
            }
            if ( strcasecmp(sta_args.pmf->sval[0],"disabled") == 0) {
                wifi_config.sta.pmf_cfg.capable = false;
                wifi_config.sta.pmf_cfg.required = false;
            } 
        }

            ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_STA) );
            ESP_ERROR_CHECK( esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config) );
            ESP_LOGI(TAG, "EAP ID: '%s'", sta_args.eap_id->sval[0]);
            ESP_LOGI(TAG, "EAP METHOD: '%s'", sta_args.eap_method->sval[0]);
            ESP_LOGI(TAG, "EAP USERNAME: '%s'", sta_args.username->sval[0]);
            ESP_LOGI(TAG, "EAP PASSWORD: '%s'", sta_args.password->sval[0]);

	    /* Set CA certs */
	    if (sta_args.trustedRootCA->sval[0]) {
                set_ca_certs(sta_args.trustedRootCA->sval[0]);
	    }

	    /* set user certs */
	   if (strcasecmp( sta_args.eap_method->sval[0], EAP_TLS) == 0) {
	       if (sta_args.clientCertificate->sval[0]) {
		   set_user_certs(sta_args.clientCertificate->sval[0]);
	       }
        }
        if (sta_args.eap_id->count == 1 ) {
            esp_eap_client_set_identity((uint8_t *)sta_args.eap_id->sval[0], strlen(sta_args.eap_id->sval[0]));
        }
	    if ( strcasecmp( sta_args.eap_method->sval[0], EAP_PEAP) == 0  || strcasecmp( sta_args.eap_method->sval[0], EAP_TTLS) == 0 ) { 
			esp_eap_client_set_username((uint8_t *)sta_args.username->sval[0], strlen(sta_args.username->sval[0]));
			ESP_ERROR_CHECK( esp_eap_client_set_password((uint8_t *)sta_args.password->sval[0], strlen(sta_args.password->sval[0])) );
	    }

        if (strcasecmp( sta_args.keymgmtType->sval[0], "suiteb") == 0) {
		    ESP_ERROR_CHECK(esp_eap_client_set_suiteb_192bit_certification(true));
        }

	    reconnect = true;
	    ESP_ERROR_CHECK( esp_wifi_sta_enterprise_enable() );
	    ESP_ERROR_CHECK( esp_wifi_start() );

            ESP_LOGI(TAG, "esp_wifi_connect start, h2e:%d", wifi_config.sta.sae_pwe_h2e);
	    esp_wifi_connect();		    
        }
        else{
	    if(sta_args.bssid->count){
                if(sta_args.password->count == 1){
                    if(sta_args.authmode->count == 1){
		        if (strcasecmp( sta_args.authmode->sval[0], WPA2) == 0 ) {
                            printf("disabling wpa3\n");
		            void esp_wifi_unregister_wpa3_cb(void);
                            esp_wifi_unregister_wpa3_cb();
                        }
	            }
                wifi_cmd_sta_join(sta_args.ssid->sval[0], sta_args.password->sval[0], sta_args.pmf->sval[0], sta_args.authmode->sval[0],sta_args.bssid->sval[0],sta_args.btm->sval[0]);
	    }
                else{
                    wifi_cmd_sta_join(sta_args.ssid->sval[0], NULL, sta_args.pmf->sval[0], sta_args.authmode->sval[0],sta_args.bssid->sval[0],sta_args.btm->sval[0]);
                }
            }
	    else{
		if(sta_args.password->count == 1){
                    if(sta_args.authmode->count == 1){
		        if (strcasecmp( sta_args.authmode->sval[0], WPA2) == 0 ) {
                           printf("disabling wpa3\n");
		           void esp_wifi_unregister_wpa3_cb(void);
                           esp_wifi_unregister_wpa3_cb();
                        }
	            }
                    wifi_cmd_sta_join(sta_args.ssid->sval[0], sta_args.password->sval[0], sta_args.pmf->sval[0], sta_args.authmode->sval[0],NULL,sta_args.btm->sval[0]);
	        }
                else{
                    wifi_cmd_sta_join(sta_args.ssid->sval[0], NULL, sta_args.pmf->sval[0], sta_args.authmode->sval[0],NULL,sta_args.btm->sval[0]);
                }
            }
    }
    return 0;
}

static bool wifi_cmd_sta_scan(const char* ssid)
{
    wifi_scan_config_t scan_config = { 0 };
    scan_config.ssid = (uint8_t *) ssid;

    ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_STA) );
    ESP_ERROR_CHECK( esp_wifi_scan_start(&scan_config, false) );
    return true;
}

static int wifi_cmd_scan(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &scan_args);

    if (nerrors != 0) {
        arg_print_errors(stderr, scan_args.end, argv[0]);
        return 1;
    }
	
	ESP_LOGI(TAG, "sta start to scan");
	if ( scan_args.ssid->count == 1 ) {
		wifi_cmd_sta_scan(scan_args.ssid->sval[0]);
	} else {
		wifi_cmd_sta_scan(NULL);
	}
    return 0;
}


static int wifi_cmd_set_ip(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &static_ip_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, static_ip_args.end, argv[0]);
        return 0;
    }
    
    esp_netif_ip_info_t ip;
    
    if (static_ip_args.ip->count != 0) {
	ip.ip.addr = esp_ip4addr_aton(static_ip_args.ip->sval[0]);    
    }

    if (static_ip_args.netmask->count != 0) {
	ip.netmask.addr = esp_ip4addr_aton(static_ip_args.netmask->sval[0]);    
    } else {
	ip.netmask.addr = esp_ip4addr_aton("255.255.255.0");
    }

    if (static_ip_args.gw->count != 0) {
        ip.gw.addr = esp_ip4addr_aton(static_ip_args.gw->sval[0]);
    } else {
	ip.gw.addr = ((ip.ip.addr & ~(0xFF<<24)) | (0x01<<24));
    }
  
    /* set static IP settings */
    esp_netif_dhcpc_stop(netif_sta);   
    esp_netif_set_ip_info(netif_sta, &ip); 
    return 0;
}

#ifdef CONFIG_WPA_MBO_SUPPORT
static int wifi_mbo_set_chan_pref(int argc, char** argv)
{
	
    int nerrors = arg_parse(argc, argv, (void**) &wifi_mbo_set_chan_pref_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, wifi_mbo_set_chan_pref_args.end, argv[0]);
        return 0;
    }

    if (wifi_mbo_set_chan_pref_args.reset->count != 0) {
        esp_mbo_update_non_pref_chan(NULL);
        return 0;
    }

    struct non_pref_chan_s *chans = malloc(sizeof(struct non_pref_chan_s) + 2 * sizeof(struct non_pref_chan));
    chans->non_pref_chan_num = 2;

    /* first */
    struct non_pref_chan *chan = &chans->chan[0];
    chan->reason = wifi_mbo_set_chan_pref_args.ch_reason_code1->ival[0];
    chan->oper_class = wifi_mbo_set_chan_pref_args.ch_op_class1->ival[0];
    chan->chan = wifi_mbo_set_chan_pref_args.ch_pref_num1->ival[0];
    chan->preference = wifi_mbo_set_chan_pref_args.ch_pref1->ival[0];;

    /* second */
    chan = &chans->chan[1];
    chan->reason = wifi_mbo_set_chan_pref_args.ch_reason_code2->ival[0];
    chan->oper_class = wifi_mbo_set_chan_pref_args.ch_op_class2->ival[0];
    chan->chan = wifi_mbo_set_chan_pref_args.ch_pref_num2->ival[0];
    chan->preference = wifi_mbo_set_chan_pref_args.ch_pref2->ival[0];;

    esp_mbo_update_non_pref_chan(chans);
    free(chans);

    return 0;
}
#endif 

static bool wifi_cmd_ap_set(const char* ssid, const char* pass, int channel)
{
    wifi_config_t wifi_config = {
        .ap = {
            .ssid = "",
            .ssid_len = 0,
            .max_connection = 4,
            .password = "",
            .channel = 1,
            .authmode = WIFI_AUTH_WPA_WPA2_PSK
        },
    };

    reconnect = false;
    strncpy((char*) wifi_config.ap.ssid, ssid, sizeof(wifi_config.ap.ssid));
    wifi_config.ap.channel = channel;
    if (pass) {
        if (strlen(pass) != 0 && strlen(pass) < 8) {
            reconnect = true;
            ESP_LOGE(TAG, "password less than 8");
            return false;
        }
        strncpy((char*) wifi_config.ap.password, pass, sizeof(wifi_config.ap.password));
    }

    if (strlen(pass) == 0) {
        wifi_config.ap.authmode = WIFI_AUTH_OPEN;
    }

    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_AP));
    uint8_t dut_mac[6] = { 0x18, 0x11, 0x22, 0x22, 0x33, 0x78 };
    ESP_ERROR_CHECK(esp_wifi_set_mac(ESP_IF_WIFI_AP, dut_mac));
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
    ESP_ERROR_CHECK(esp_wifi_set_protocol(ESP_IF_WIFI_AP, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N | WIFI_PROTOCOL_11AX));
#else
    ESP_ERROR_CHECK(esp_wifi_set_protocol(ESP_IF_WIFI_AP, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N));
#endif    
    ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_AP, &wifi_config));
    return true;
}

static int wifi_cmd_ap(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &ap_args);
    int chan =1;
    
    if (nerrors != 0) {
        arg_print_errors(stderr, ap_args.end, argv[0]);
        return 1;
    }

    if ( ap_args.chan->count == 1 ) {
		chan = atoi(ap_args.chan->sval[0]);
	} 
    
    wifi_cmd_ap_set(ap_args.ssid->sval[0], ap_args.password->sval[0], chan);
    ESP_LOGI(TAG, "AP mode, ssid: '%s', password '%s', channel: '%d' ", ap_args.ssid->sval[0], ap_args.password->sval[0], chan);
            
    if ( ap_args.cbw->count == 1 ) {
		int cbw = atoi(ap_args.cbw->sval[0]);
		ESP_LOGI(TAG, "Configuring %d MHz on AP", cbw);
		if (cbw == 20) {
			esp_wifi_set_bandwidth(WIFI_IF_AP, WIFI_BW_HT20);					
		} else if (cbw == 40) {
			esp_wifi_set_bandwidth(WIFI_IF_AP, WIFI_BW_HT40);
			
		} else {
			ESP_LOGI(TAG, "%d MHz is not supported", cbw); 
		}
    }
    return 0;
}

void wifi_get_local_ip(esp_netif_ip_info_t *ip_info)
{
    int bits = xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT, 0, 1, 0);
    esp_netif_t *ifx = netif_ap;
    //tcpip_adapter_ip_info_t ip_info;
    wifi_mode_t mode;

    esp_wifi_get_mode(&mode);
    if (WIFI_MODE_STA == mode) {
        bits = xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT, 0, 1, 0);
        if (bits & CONNECTED_BIT) {
            ifx = netif_sta;
        } else {
            ESP_LOGE(TAG, "sta has no IP");
            return;
        }
     }
     esp_netif_get_ip_info(ifx, ip_info);
}

static int device_reset(int argc, char** argv){
    usleep(5000);
    printf("Reseting Device :");

    esp_restart();
    return 0;
}

int esp_wifi_get_assoc_bssid_internal(uint8_t *bssid);
static int wifi_cmd_query(int argc, char** argv)
{
    wifi_config_t cfg;
    wifi_bandwidth_t cbw;
    uint8_t mac[6];
    esp_netif_ip_info_t ip_info;
    
    wifi_mode_t mode;
    esp_wifi_get_mode(&mode);
    wifi_get_local_ip(&ip_info);
    
     printf("\nFree memory in quer is : %d \n", heap_caps_get_free_size(MALLOC_CAP_8BIT));
    printf("Wireless info:");
    if (WIFI_MODE_AP == mode) {
        esp_wifi_get_config(WIFI_IF_AP, &cfg);
        esp_wifi_get_bandwidth(WIFI_IF_AP, &cbw);
             
        printf("\n");
        printf("\tmode: ap\n");
        printf("\tssid: %s\n", cfg.ap.ssid);
        printf("\tpassword: %s\n", cfg.ap.password);
        printf("\tchannel: %d\n", cfg.ap.channel);
        if ( cbw == WIFI_BW_HT20 )
			printf("\tcbw: 20 MHz\n");
		else if ( cbw == WIFI_BW_HT40 )
		    printf("\tcbw: 40 MHz\n");
		    
		if (esp_wifi_get_mac(WIFI_IF_AP,mac) == ESP_OK) {
            printf("\tap mac: "MACSTR , MAC2STR(mac));
            printf("\n");
        } 
        printf("\tip: %d.%d.%d.%d\n", ip_info.ip.addr&0xFF, (ip_info.ip.addr>>8)&0xFF, (ip_info.ip.addr>>16)&0xFF, (ip_info.ip.addr>>24)&0xFF);
        printf("\tnetmask: %d.%d.%d.%d\n", ip_info.netmask.addr&0xFF, (ip_info.netmask.addr>>8)&0xFF, (ip_info.netmask.addr>>16)&0xFF, (ip_info.netmask.addr>>24)&0xFF);
        printf("\tgateway: %d.%d.%d.%d\n", ip_info.gw.addr&0xFF, (ip_info.gw.addr>>8)&0xFF, (ip_info.gw.addr>>16)&0xFF, (ip_info.gw.addr>>24)&0xFF); 
        printf("\n");
    } else if (WIFI_MODE_STA == mode) {
        int bits = xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT, 0, 1, 0);
        if (bits & CONNECTED_BIT) {
            esp_wifi_get_config(WIFI_IF_STA, &cfg);
            esp_wifi_get_bandwidth(WIFI_IF_STA, &cbw);
            printf("\n");
            printf("\tmode: station\n");
            printf("\tstatus: connected\n");
            printf("\tssid: %s\n", cfg.ap.ssid);
            esp_wifi_get_assoc_bssid_internal(cfg.sta.bssid);
            printf("\tbssid: "MACSTR, MAC2STR(cfg.sta.bssid));
            printf("\n");
            wifi_second_chan_t secondary;
            esp_wifi_get_channel(&cfg.sta.channel, &secondary);
            printf("\tchannel: %d\n", cfg.sta.channel);
            if ( cfg.sta.pmf_cfg.capable ) {
                if ( cfg.sta.pmf_cfg.required ) {
                    printf("\tpmf: required\n");
                } else {
                    printf("\tpmf: optional\n");
                }
            } else {
                printf("\tpmf: disabled\n");
            }
            if (cbw == WIFI_BW_HT20)
                printf("\tcbw: 20 MHz\n");
            else if ( cbw == WIFI_BW_HT40 )
                printf("\tcbw: 40 MHz\n");
            
            if (esp_wifi_get_mac(WIFI_IF_STA,mac) == ESP_OK) {
				printf("\tsta mac: "MACSTR , MAC2STR(mac));
				printf("\n");
            }

            if (bits & GOT_IP_BIT) {
                printf("\tsta_ip: %d.%d.%d.%d\n", ip_info.ip.addr&0xFF, (ip_info.ip.addr>>8)&0xFF, (ip_info.ip.addr>>16)&0xFF, (ip_info.ip.addr>>24)&0xFF);
                printf("\tnetmask: %d.%d.%d.%d\n", ip_info.netmask.addr&0xFF, (ip_info.netmask.addr>>8)&0xFF, (ip_info.netmask.addr>>16)&0xFF, (ip_info.netmask.addr>>24)&0xFF);
                printf("\tgateway: %d.%d.%d.%d\n", ip_info.gw.addr&0xFF, (ip_info.gw.addr>>8)&0xFF, (ip_info.gw.addr>>16)&0xFF, (ip_info.gw.addr>>24)&0xFF);
                printf("\n");
            } else {
                printf("\tsta_ip: NA\n\tnetmask: NA\n\tgateway: NA\n");
            }
           
        } else {
            printf("\n");
            printf("\tmode: disconnected\n");
        }
    } else {		
        printf("\n");
        printf("\tmode: null\n");
    }
        if (esp_wifi_get_mac(WIFI_IF_AP,mac) == ESP_OK) {
            printf("\tap mac: "MACSTR , MAC2STR(mac));
            printf("\n");
        }         

        if (esp_wifi_get_mac(WIFI_IF_STA,mac) == ESP_OK) {
            printf("\tsta mac: "MACSTR , MAC2STR(mac));
            printf("\n");
        }          
        return 0;

    return 0;
}

static int wifi_disconnect(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &disconnect_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, btm_args.end, argv[0]);
        return 1;
    }

    reconnect = false;
    esp_wifi_disconnect();
				
    return 0;
}

static int sta_send_btm_query(int argc, char** argv){

    int nerrors = arg_parse(argc, argv, (void**) &btm_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, btm_args.end, argv[0]);
        return 1;
    }
    int reason_code = atoi((char*)btm_args.reason_code);
    int candidate_list = atoi((char*)btm_args.candidate_list);
    ESP_LOGI(TAG, "Calling the wifi supplicant BTM Query with the following values:\n Reason Code : %d \ncandidate_list : %d\n",reason_code,candidate_list);
    return esp_wnm_send_bss_transition_mgmt_query(reason_code,NULL,candidate_list);
}
static int set_wifi_log_level(int argc, char** argv)
{
	int nerrors = arg_parse(argc, argv, (void**) &loglevel_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, loglevel_args.end, argv[0]);
        return 1;
    }
    esp_log_level_set("wifi", loglevel_args.level->ival[0]);
	return 0;
}

static int set_prog_cmd(int argc, char** argv)
{
	int nerrors = arg_parse(argc, argv, (void**) &setprog_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, setprog_args.end, argv[0]);
        return 1;
    }

	if ( setprog_args.prog->count == 1 ) {
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
		if(strcasecmp(setprog_args.prog->sval[0],"HE") == 0){
            ESP_ERROR_CHECK(esp_wifi_set_protocol(0, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N | WIFI_PROTOCOL_11AX));
            ESP_ERROR_CHECK(esp_wifi_set_bandwidth(0, WIFI_BW_HT20));
            printf("cmd_wifi.c set_prog: Sta set BW to 20 Mhz and Prog to HE\n");
            return 0;
        }
#endif //CONFIG_SOC_WIFI_HE_SUPPORT
#ifdef CONFIG_WPA_MBO_SUPPORT
       if(strcasecmp(setprog_args.prog->sval[0],"MBO") == 0){
#ifdef CONFIG_IDF_TARGET_ESP32C6
            ESP_ERROR_CHECK(esp_wifi_set_protocol(0, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N | WIFI_PROTOCOL_11AX));
#else //CONFIG_IDF_TARGET_ESP32C6
            ESP_ERROR_CHECK(esp_wifi_set_protocol(0, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N));
#endif //CONFIG_IDF_TARGET_ESP32C6
            ESP_ERROR_CHECK(esp_wifi_set_bandwidth(0, WIFI_BW_HT20));
            printf("cmd_wifi.c set_prog: Sta set BW to 20 Mhz and Prog to MBO\n");
            return 0;
       }
#endif //CONFIG_WPA_MBO_SUPPORT
            ESP_ERROR_CHECK(esp_wifi_set_protocol(0, WIFI_PROTOCOL_11B | WIFI_PROTOCOL_11G | WIFI_PROTOCOL_11N));
            ESP_ERROR_CHECK(esp_wifi_set_bandwidth(0, WIFI_BW_HT40));
            printf("cmd_wifi.c set_prog: Sta set BW to 40 Mhz\n");
    }
    else{
        return 1;
    }
	return 0;
}

static int set_timestamp(int argc, char** argv)
{
	int nerrors = arg_parse(argc, argv, (void**) &settime_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, settime_args.end, argv[0]);
        return 1;
    }

    struct timeval now = { .tv_sec = settime_args.time->ival[0] };
    settimeofday(&now, NULL);

    gettimeofday(&now, NULL);
    printf("set time: %d\n", now.tv_sec);


	return 0;
}

static int wifi_cmd_tb_power(int argc, char** argv)
{
#if HAL_MAC_ESP32AX_760
#define DR_REG_WDEV_TX_BASE  0x60033400

#define WDEVPWR_TB_TAB0                         (DR_REG_WDEV_TX_BASE + 0x0030)
#define WDEV_PWR_MCS3_TB_V 0x0000003F
#define WDEV_PWR_MCS3_TB_S 24
#define WDEV_PWR_MCS3_TB   (WDEV_PWR_MCS3_TB_V << WDEV_PWR_MCS3_TB_S)

#define WDEV_PWR_MCS2_TB_V 0x0000003F
#define WDEV_PWR_MCS2_TB_S 16
#define WDEV_PWR_MCS2_TB   (WDEV_PWR_MCS2_TB_V << WDEV_PWR_MCS2_TB_S)

#define WDEV_PWR_MCS1_TB_V 0x0000003F
#define WDEV_PWR_MCS1_TB_S 8
#define WDEV_PWR_MCS1_TB   (WDEV_PWR_MCS1_TB_V << WDEV_PWR_MCS1_TB_S)

#define WDEV_PWR_MCS0_TB_V 0x0000003F
#define WDEV_PWR_MCS0_TB_S 0
#define WDEV_PWR_MCS0_TB   (WDEV_PWR_MCS0_TB_V << WDEV_PWR_MCS0_TB_S)

#define WDEVPWR_TB_TAB1                         (DR_REG_WDEV_TX_BASE + 0x0034)
#define WDEV_PWR_MCS7_TB_V 0x0000003F
#define WDEV_PWR_MCS7_TB_S 24
#define WDEV_PWR_MCS7_TB   (WDEV_PWR_MCS7_TB_V << WDEV_PWR_MCS7_TB_S)

#define WDEV_PWR_MCS6_TB_V 0x0000003F
#define WDEV_PWR_MCS6_TB_S 16
#define WDEV_PWR_MCS6_TB   (WDEV_PWR_MCS6_TB_V << WDEV_PWR_MCS6_TB_S)

#define WDEV_PWR_MCS5_TB_V 0x0000003F
#define WDEV_PWR_MCS5_TB_S 8
#define WDEV_PWR_MCS5_TB   (WDEV_PWR_MCS5_TB_V << WDEV_PWR_MCS5_TB_S)

#define WDEV_PWR_MCS4_TB_V 0x0000003F
#define WDEV_PWR_MCS4_TB_S 0
#define WDEV_PWR_MCS4_TB   (WDEV_PWR_MCS4_TB_V << WDEV_PWR_MCS4_TB_S)

    ESP_LOGW(TAG, "%d, %d, %d, %d, %d, %d, %d, %d",
            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB0, WDEV_PWR_MCS0_TB),
            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB0, WDEV_PWR_MCS1_TB),
            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB0, WDEV_PWR_MCS2_TB),
            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB0, WDEV_PWR_MCS3_TB),

            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB1, WDEV_PWR_MCS4_TB),
            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB1, WDEV_PWR_MCS5_TB),
            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB1, WDEV_PWR_MCS6_TB),
            (int8_t) REG_GET_FIELD(WDEVPWR_TB_TAB1, WDEV_PWR_MCS7_TB));
#endif
    return 0;
}

static int wifi_cmd_rfo_dump(int argc, char** argv)
{
#if ESP_WIFI_HAL_TEST
    int nerrors = arg_parse(argc, argv, (void**) &test_rfo_dump_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, test_rfo_dump_args.end, argv[0]);
        return 1;
    }
    if (test_rfo_dump_args.enable->count) {
        if (test_rfo_dump_args.enable->dval[0] == 1) {
            esp_test_dump_rfo = 1;
        } else {
            esp_test_dump_rfo = 0;
        }
    }
#endif
    return 0;
}

static int wifi_dump_rpt(int argc, char** argv)
{
    int addr;
    int value;
    int i = 0;

    SET_PERI_REG_MASK(BB_BF_CTRL1_REG, (BIT(15)));
    SET_PERI_REG_MASK(MAC_BF_CONFIG_REG, (BIT(0)));
#if 0
    int rev_val;
    for (i = 0; i < 192; i++) {
        addr = base + (i << 2);
        value = REG_READ(addr);
        rev_val = (value & 0xFE00) | (((~(value & 0x1FF) + 1) & 0X1FF));
        ESP_LOGW(TAG, "rev: 0x%x, i: %d, addr 0x%x, 0x%x", rev_val, i, addr, value);
        printf("0x%x,\n", value);
        REG_WRITE(addr, rev_val);
    }
    printf("\n\n");
#endif
    ESP_LOGW(TAG, "*******************************");
    for (i = 0; i < 192; i++) {
        addr = BF_START + (i << 2);
        value = REG_READ(addr);
        printf("[%d]0x%x\n", i, value);
    }
    printf("addr:0x%08x, avgSNR:%ld\n", BF_START + (229 << 2), REG_READ(BF_START + (229 << 2)));
    printf("\n");
    return 0;

}

static int wifi_cmd_powersave(int argc, char **argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &powersave_args);

    if (nerrors != 0) {
        arg_print_errors(stderr, powersave_args.end, argv[0]);
        return 1;
    }
    if(powersave_args.mode->count == 1)
    {
        if(powersave_args.mode->ival[0]){
            ESP_LOGW(TAG, "set to WIFI_PS_MIN_MODEM");
            ESP_ERROR_CHECK(esp_wifi_set_ps(WIFI_PS_MIN_MODEM));
        }
        else{
            ESP_LOGW(TAG, "set to WIFI_PS_NONE");
            ESP_ERROR_CHECK(esp_wifi_set_ps(WIFI_PS_NONE));
        }
    }
    return 0;
}

#if ESP_WIFI_HAL_TEST
static int wifi_cmd_su_dump(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &test_su_dump_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, test_su_dump_args.end, argv[0]);
        return 1;
    }
    if (test_su_dump_args.enable->count) {
        if (test_su_dump_args.enable->dval[0] == 1) {
            esp_test_dump_su = 1;
            s_dump_su_state = true;
        } else {
            esp_test_dump_su = 0;
            s_dump_su_state = false;
        }
    } else {
        esp_test_dump_su = 1 - s_dump_su_state;
        s_dump_su_state = esp_test_dump_su;
    }
    if (s_dump_su_state) {
        printf("dump su enabled\n");
    } else {
        printf("dump su disabled\n");
    }
    return 0;
}

static int wifi_cmd_set_ull(int argc, char **argv)
{
    int nerrors = arg_parse(argc, argv, (void **) &test_ull_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, test_ull_args.end, argv[0]);
        return 1;
    }

    dbg_set_tb_ul_len(test_ull_args.ull->ival[0]);
    ESP_LOGW(TAG, "(tb)set ul_length:%d", test_ull_args.ull->ival[0]);
    return 0;
}
#endif

static int wifi_cmd_set_force_gain(int argc, char **argv)
{
    int nerrors = arg_parse(argc, argv, (void **)&force_gain_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, force_gain_args.end, argv[0]);
        return 1;
    }
    if (force_gain_args.gain->count) {
#if HAL_MAC_ESP32AX_760
        SET_PERI_REG_MASK(0x6001c02c, (BIT(23)));
        SET_PERI_REG_BITS(0x6001c02c, 0xFF, force_gain_args.gain->ival[0], 24);
#else
        SET_PERI_REG_MASK(0x6004902c, (BIT(23)));
        SET_PERI_REG_BITS(0x6004902c, 0xFF, force_gain_args.gain->ival[0], 24);
#endif
        printf("force gain set to %d\n", force_gain_args.gain->ival[0]);
    }
    return 0;
}


static int wifi_cmd_set_wireless(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &set_args);

    if (nerrors != 0) {
        arg_print_errors(stderr, set_args.end, argv[0]);
        return 1;
    }
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
    if (set_args.mcs->count != 0) {
        ESP_ERROR_CHECK(esp_wifi_cert_set_mcs(set_args.mcs->ival[0]));
    }
    if (set_args.bcc->count != 0) {
        ESP_ERROR_CHECK(esp_wifi_cert_set_bcc(set_args.bcc->ival[0]));
    }
    if (set_args.ltf->count != 0) {
        ESP_ERROR_CHECK(esp_wifi_cert_set_ltf((float) (float) set_args.ltf->ival[0] / (float) 10));
    }
    if (set_args.gi->count != 0) {
        ESP_ERROR_CHECK(esp_wifi_cert_set_gi((float) (float) set_args.gi->ival[0] / (float) 10));
    }
    if (set_args.nss->count != 0) {
        ESP_ERROR_CHECK(esp_wifi_cert_set_nss(set_args.nss->ival[0]));
    }
    if (set_args.ersu->count != 0) {
        ESP_ERROR_CHECK(esp_wifi_cert_set_ersu(set_args.ersu->ival[0]));
    }
    /*
     * esp_wifi_cert_tx_ltf = 1, 2, 4
     *   1 - 3.2 us
     *   2 - 6.4 us
     *   4 - 12.8 us
     * esp_wifi_cert_tx_gi = 1, 2, 4
     *   1 - 0.8 us
     *   2 - 1.6 us
     *   4 - 3.2 us
     * esp32ax only supports the following settings:
     * esp_wifi_cert_tx_ltf=2 and esp_wifi_cert_tx_gi=1, HE_SU_ERSU_2_LTF_0_8_US_GI
     * esp_wifi_cert_tx_ltf=2 and esp_wifi_cert_tx_gi=2, HE_SU_ERSU_2_LTF_1_6_US_GI
     * esp_wifi_cert_tx_ltf=4 and esp_wifi_cert_tx_gi=4, HE_SU_ERSU_4_LTF_3_2_US_GI
     *
     * note:

     * esp_wifi_cert_tx_ltf=1 and esp_wifi_cert_tx_gi=1, HE_SU_ERSU_1_LTF_0_8_US_GI(not support)
     *
     */
    ESP_LOGW(TAG, "mcs:%d, bcc:%d, ltf:%d, gi:%d, nss:%d, ersu:%d", esp_wifi_cert_tx_mcs,
            esp_wifi_cert_tx_bcc, esp_wifi_cert_tx_ltf, esp_wifi_cert_tx_gi, esp_wifi_cert_tx_nss, esp_wifi_cert_tx_ersu);
#endif
    return 0;
}

static int wifi_cmd_set_amsdu(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &set_amsdu_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, set_amsdu_args.end, argv[0]);
        return 1;
    }
#ifdef CONFIG_SOC_WIFI_HE_SUPPORT
    if (set_amsdu_args.enable->count) {
        if (set_amsdu_args.enable->dval[0] == 1) {
            esp_test_baparas_support_amsdu = 1;
        } else {
            esp_test_baparas_support_amsdu = 0;
        }
    }
    ESP_LOGW(TAG, "amsdu:%d", esp_test_baparas_support_amsdu);
#endif
    return 0;
}

static int wifi_cmd_set_obss(int argc, char** argv)
{
    wifi_ioctl_config_t cfg1 = {0};
    int nerrors = arg_parse(argc, argv, (void**) &set_obss_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, set_obss_args.end, argv[0]);
        return 1;
    }

    if (set_obss_args.enable->count) {
        if (set_obss_args.enable->dval[0] == 1) {
            cfg1.data.ht2040_coex.enable = 1;
            esp_wifi_internal_ioctl(WIFI_IOCTL_SET_STA_HT2040_COEX, &cfg1);
        } else {
            cfg1.data.ht2040_coex.enable = 0;
            esp_wifi_internal_ioctl(WIFI_IOCTL_SET_STA_HT2040_COEX, &cfg1);
        }
    }
    ESP_LOGW(TAG, "obss scan enable:%d", cfg1.data.ht2040_coex.enable);

    return 0;
}


static int wifi_cmd_iperf(int argc, char **argv)
{
    int nerrors = arg_parse(argc, argv, (void **) &iperf_args);
    iperf_cfg_t cfg;

    if (nerrors != 0) {
        arg_print_errors(stderr, iperf_args.end, argv[0]);
        return 0;
    }

    memset(&cfg, 0, sizeof(cfg));

    // now wifi iperf only support IPV4 address
    cfg.type = IPERF_IP_TYPE_IPV4;

    if ( iperf_args.abort->count != 0) {
        iperf_stop();
        return 0;
    }

    if ( ((iperf_args.ip->count == 0) && (iperf_args.server->count == 0)) ||
            ((iperf_args.ip->count != 0) && (iperf_args.server->count != 0)) ) {
        ESP_LOGE(TAG, "should specific client/server mode");
        return 0;
    }

    if (iperf_args.ip->count == 0) {
        cfg.flag |= IPERF_FLAG_SERVER;
    } else {
        cfg.destination_ip4 = esp_ip4addr_aton(iperf_args.ip->sval[0]);
        cfg.flag |= IPERF_FLAG_CLIENT;
    }
    int bits = xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT, 0, 1, 0);
    esp_netif_t *netif = netif_ap;
    esp_netif_ip_info_t ip_info;
    wifi_mode_t mode;

    esp_wifi_get_mode(&mode);
    if (WIFI_MODE_STA == mode) {
        bits = xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT, 0, 1, 0);
        if (bits & CONNECTED_BIT) {
            netif = netif_sta;
        } else {
            ESP_LOGE(TAG, "sta has no IP");
             cfg.source_ip4 = 0;
             return 0;
        }
    }
    esp_netif_get_ip_info(netif, &ip_info);
    cfg.source_ip4 = ip_info.ip.addr;

    if (iperf_args.udp->count == 0) {
        cfg.flag |= IPERF_FLAG_TCP;
    } else {
        cfg.flag |= IPERF_FLAG_UDP;
    }

    if (iperf_args.length->count == 0) {
        cfg.len_send_buf = 0;
    } else {
        cfg.len_send_buf = iperf_args.length->ival[0];
    }

    if (iperf_args.port->count == 0) {
        cfg.sport = IPERF_DEFAULT_PORT;
        cfg.dport = IPERF_DEFAULT_PORT;
    } else {
        if (cfg.flag & IPERF_FLAG_SERVER) {
            cfg.sport = iperf_args.port->ival[0];
            cfg.dport = IPERF_DEFAULT_PORT;
        } else {
            cfg.sport = IPERF_DEFAULT_PORT;
            cfg.dport = iperf_args.port->ival[0];
        }
    }

    if (iperf_args.interval->count == 0) {
        cfg.interval = IPERF_DEFAULT_INTERVAL;
    } else {
        cfg.interval = iperf_args.interval->ival[0];
        if (cfg.interval <= 0) {
            cfg.interval = IPERF_DEFAULT_INTERVAL;
        }
    }

    if (iperf_args.time->count == 0) {
        cfg.time = IPERF_DEFAULT_TIME;
    } else {
        cfg.time = iperf_args.time->ival[0];
        if (cfg.time <= cfg.interval) {
            cfg.time = cfg.interval;
        }
    }

    /* iperf -b */
    if (iperf_args.bw_limit->count == 0) {
        cfg.bw_lim = IPERF_DEFAULT_NO_BW_LIMIT;
    } else {
        cfg.bw_lim = iperf_args.bw_limit->ival[0];
        if (cfg.bw_lim <= 0) {
            cfg.bw_lim = IPERF_DEFAULT_NO_BW_LIMIT;
        }
    }

    ESP_LOGI(TAG, "mode=%s-%s sip=%ld.%ld.%ld.%ld:%d, dip=%ld.%ld.%ld.%ld:%d, interval=%ld, time=%ld",
             cfg.flag & IPERF_FLAG_TCP ? "tcp" : "udp",
             cfg.flag & IPERF_FLAG_SERVER ? "server" : "client",
             cfg.source_ip4 & 0xFF, (cfg.source_ip4 >> 8) & 0xFF, (cfg.source_ip4 >> 16) & 0xFF,
             (cfg.source_ip4 >> 24) & 0xFF, cfg.sport,
             cfg.destination_ip4 & 0xFF, (cfg.destination_ip4 >> 8) & 0xFF,
             (cfg.destination_ip4 >> 16) & 0xFF, (cfg.destination_ip4 >> 24) & 0xFF, cfg.dport,
             cfg.interval, cfg.time);

    iperf_start(&cfg);

    return 0;
}

static int wifi_cmd_addba(int argc, char **argv)
{
    int nerrors = arg_parse(argc, argv, (void **)&addba_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, addba_args.end, argv[0]);
        return 1;
    }
    if (addba_args.tid->count) {
        if (addba_args.tid->ival[0] >=0 && addba_args.tid->ival[0] <= 7) {
#if ESP_WIFI_HAL_TEST
            esp_test_sta_create_addba(addba_args.tid->ival[0]);
#endif
        }
    }
    return 0;
}


static int wifi_cmd_itwt_set_target_wake_time_offset(int argc, char** argv)
{
    int nerrors = arg_parse(argc, argv, (void**) &itwt_set_twt_offset_args);
    if (nerrors != 0) {
        arg_print_errors(stderr, itwt_set_twt_offset_args.end, argv[0]);
        return 1;
    }
    esp_err_t err = ESP_OK;
    if (itwt_set_twt_offset_args.offset->count) {
        ESP_LOGW(TAG, "(itwt)set target wake time offset:%d us", itwt_set_twt_offset_args.offset->ival[0]);
#ifdef CONFIG_IDF_TARGET_ESP32C6
        err = esp_wifi_sta_itwt_set_target_wake_time_offset(itwt_set_twt_offset_args.offset->ival[0]);
#endif
    }

    ESP_LOGW(TAG, "err:%d", err);
    return err;
}


void register_wifi()
{
    sta_args.ssid = arg_str1(NULL, NULL, "<ssid>", "SSID of AP");
    sta_args.security = arg_str0(NULL, "security", "<security>", "security type (ie. OPEN | WEP| WPA2 | WPA2-EAP)");
    sta_args.eap_method = arg_str0(NULL, "eap_method", "<eap_method>", "EAP Method (ie. TLS | PEAP | TTLS)");
    sta_args.eap_id = arg_str0(NULL, "eap_id", "<eap_id>", "EAP Id");
    sta_args.username = arg_str0(NULL, "username", "<username>", "Enterprise User name");    
    sta_args.password = arg_str0(NULL, "password", "<password>", "password of AP"); 
    sta_args.trustedRootCA = arg_str0(NULL, "trustedRootCA", "<trustedRootCA>", "trustedRootCA of Certs");
    sta_args.clientCertificate = arg_str0(NULL, "clientCertificate", "<clientCertificate>", "trustedRootCA of Certs");
    sta_args.pmf = arg_str0(NULL, "pmf", "<optional|required|disabled>", "PMF capability (ie. optional | required | disabled)"); 
    sta_args.roaming = arg_str0(NULL, "roaming", "<enable|disable>", "Roaming capability (ie. enable | disable)");
    sta_args.authmode = arg_str0(NULL, "authmode", "<wpa2|wpa3>", "minimum authmode to be used in connection");
    sta_args.keymgmtType = arg_str0(NULL, "keymgmtType", "<any|suiteb>", "key mgmt type");
    sta_args.addba_reject = arg_dbl0(NULL, "addba_reject", "<addba reject enabled>", "addba reject enable(ie. 1)");
    sta_args.bssid = arg_str0(NULL, "bssid", "<bssid>", "bssid of AP");
    sta_args.btm = arg_str0(NULL, "btm", "<enable|disable>", "decide whether or not to enable or disable btm");
    sta_args.end = arg_end(2);

    const esp_console_cmd_t sta_cmd = {
        .command = "connect_cmd",
        .help = "WiFi is station mode, join specified soft-AP",
        .hint = NULL,
        .func = &wifi_cmd_sta,
        .argtable = &sta_args
    };

    ESP_ERROR_CHECK( esp_console_cmd_register(&sta_cmd) );

    
    //Functionality to send a btm query

    btm_args.reason_code = arg_str0(NULL, "reason_code", "<reason_code>", "Reason Code 1-9");
    btm_args.candidate_list = arg_str0(NULL, "candidate_list", "<candidate_list>", "If a Candidate list is needed.");
    btm_args.end = arg_end(2);

    const esp_console_cmd_t sta_btm_query_cmd = {
        .command = "sta_btm_query",
        .help = "Sends a BSS Transition Management Query to the AP",
        .hint = NULL,
        .func = &sta_send_btm_query,
        .argtable = &btm_args
    };

    ESP_ERROR_CHECK( esp_console_cmd_register(&sta_btm_query_cmd) );


    scan_args.ssid = arg_str0(NULL, "ssid", "<ssid>", "SSID of AP want to be scanned");
    scan_args.end = arg_end(1);

    const esp_console_cmd_t scan_cmd = {
        .command = "scan",
        .help = "WiFi is station mode, start scan ap",
        .hint = NULL,
        .func = &wifi_cmd_scan,
        .argtable = &scan_args
    };

    ESP_ERROR_CHECK( esp_console_cmd_register(&scan_cmd) );
    
    ap_args.ssid = arg_str1(NULL, NULL, "<ssid>", "SSID of AP");
    ap_args.password = arg_str0(NULL, "password", "<password>", "password of AP");
    ap_args.cbw = arg_str0(NULL, "cbw", "<channel bandwidth>", "Channel Bandwidth 20 MHz | 40 MHz");
    ap_args.chan = arg_str0(NULL, "channel", "<channel>", "Channel of AP");
    ap_args.end = arg_end(2);

    const esp_console_cmd_t ap_cmd = {
        .command = "ap",
        .help = "AP mode, configure ssid, password and channel bandwidth",
        .hint = NULL,
        .func = &wifi_cmd_ap,
        .argtable = &ap_args
    };

    ESP_ERROR_CHECK( esp_console_cmd_register(&ap_cmd));

    const esp_console_cmd_t query_cmd = {
        .command = "query",
        .help = "query WiFi info",
        .hint = NULL,
        .func = &wifi_cmd_query,
    };
    ESP_ERROR_CHECK( esp_console_cmd_register(&query_cmd) );

    const esp_console_cmd_t device_reset_cmd = {
        .command = "device_reset",
        .help = "Restart Device",
        .hint = NULL,
        .func = &device_reset,
    };
    ESP_ERROR_CHECK( esp_console_cmd_register(&device_reset_cmd) );

    static_ip_args.ip = arg_str0("i", "ip", "<ip>", "ip address");
    static_ip_args.gw = arg_str0("g", "gateway", "<gw>", "gateway address");
    static_ip_args.netmask = arg_str0("n", "netmask", "<netmask>", "netmask addess");
    static_ip_args.end = arg_end(1);
    const esp_console_cmd_t static_ip_cmd = {
        .command = "ip",
        .help = "ip settings",
        .hint = NULL,
        .func = &wifi_cmd_set_ip,
        .argtable = &static_ip_args,
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&static_ip_cmd));
    
    #ifdef CONFIG_WPA_MBO_SUPPORT
    wifi_mbo_set_chan_pref_args.reset = arg_int0(NULL, "reset", "<0 | 1>", "reset");
    wifi_mbo_set_chan_pref_args.ch_pref_num1 = arg_int0(NULL, "ch_pref_num1", "ch_pref_num1", "");
    wifi_mbo_set_chan_pref_args.ch_pref1 = arg_int0(NULL, "ch_pref1", "ch_pref1", "");
    wifi_mbo_set_chan_pref_args.ch_op_class1 = arg_int0(NULL, "ch_op_class1", "ch_op_class1", "");
    wifi_mbo_set_chan_pref_args.ch_reason_code1 = arg_int0(NULL, "ch_reason_code1", "ch_reason_code1", "");
    wifi_mbo_set_chan_pref_args.ch_pref_num2 = arg_int0(NULL, "ch_pref_num2", "ch_pref_num2", "");
    wifi_mbo_set_chan_pref_args.ch_pref2 = arg_int0(NULL, "ch_pref2", "ch_pref2", "");
    wifi_mbo_set_chan_pref_args.ch_op_class2 = arg_int0(NULL, "ch_op_class2", "ch_op_class2", "");
    wifi_mbo_set_chan_pref_args.ch_reason_code2 = arg_int0(NULL, "ch_reason_code2", "ch_reason_code2", "");
    wifi_mbo_set_chan_pref_args.end = arg_end(1);
    const esp_console_cmd_t wifi_mbo_set_chan_pref_cmd = {
        .command = "mbo",
        .help = "wifi set mbo channel preference commands",
        .hint = NULL,
        .func = &wifi_mbo_set_chan_pref,
        .argtable = &wifi_mbo_set_chan_pref_args,
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&wifi_mbo_set_chan_pref_cmd));
    #endif 

    disconnect_args.mac = arg_str0(NULL, "mac", "<mac>", "Mac address of station");
    disconnect_args.end = arg_end(1);    
    
    const esp_console_cmd_t disconnect_cmd = {
        .command = "disconnect_cmd",
        .help = "disconnect connection",
        .hint = NULL,
        .func = &wifi_disconnect,
        .argtable = &disconnect_args
    };
    ESP_ERROR_CHECK( esp_console_cmd_register(&disconnect_cmd) );

    const esp_console_cmd_t tbpwrcmd = {
            .command = "tbp",
            .help = "tb power",
            .hint = NULL,
            .func = &wifi_cmd_tb_power,
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&tbpwrcmd));

    test_rfo_dump_args.enable = arg_dbl0("e", "enable", "[enable]", "dump cfo");
    test_rfo_dump_args.end = arg_end(1);
    const esp_console_cmd_t cfocmd = {
            .command = "cfo",
            .help = "dump rfo",
            .hint = NULL,
            .func = &wifi_cmd_rfo_dump,
            .argtable = &test_rfo_dump_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&cfocmd));

    const esp_console_cmd_t beamcmd = {
            .command = "rpt",
            .help = "dump beamforming rpt",
            .hint = NULL,
            .func = &wifi_dump_rpt,
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&beamcmd));

    const esp_console_cmd_t swrcmd = {
            .command = "sw",
            .help = "sw write beamforming rpt",
            .hint = NULL,
            .func = &wifi_sw_write_rpt,
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&swrcmd));

    const esp_console_cmd_t hwrcmd = {
            .command = "hw",
            .help = "hw write beamforming rpt",
            .hint = NULL,
            .func = &wifi_hw_write_rpt,
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&hwrcmd));

    powersave_args.mode = arg_int0(NULL, "mode", "<1-MIN_MODE/0-NONE>", "Powersave Mode");
    powersave_args.end = arg_end(1);
    const esp_console_cmd_t powerave_cmd = {
            .command = "powersave",
            .help = "enable powersave",
            .hint = NULL,
            .func = &wifi_cmd_powersave,
            .argtable = &powersave_args,
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&powerave_cmd));
#if ESP_WIFI_HAL_TEST
    test_su_dump_args.enable = arg_dbl0("e", "enable", "[enable]", "dump su");
    test_su_dump_args.end = arg_end(1);
    const esp_console_cmd_t su_dump_cmd = {
            .command = "dumpsu",
            .help = "enable dump su",
            .hint = NULL,
            .func = &wifi_cmd_su_dump,
            .argtable = &test_su_dump_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&su_dump_cmd));

    test_ull_args.ull = arg_int0("l", "ull", "[ull]", "ul length");
    test_ull_args.end = arg_end(1);
    const esp_console_cmd_t ull_cmd = {
            .command = "ull",
            .help = "set ul_len for tx tb ppdu",
            .hint = NULL,
            .func = &wifi_cmd_set_ull,
            .argtable = &test_ull_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&ull_cmd));
#endif
    force_gain_args.gain = arg_int0(NULL, NULL, "[gain]", "force gain");
    force_gain_args.end = arg_end(1);
    const esp_console_cmd_t force_gain_cmd = {
        .command = "force",
        .help = "set force gain",
        .hint = NULL,
        .func = &wifi_cmd_set_force_gain,
        .argtable = &force_gain_args};
    ESP_ERROR_CHECK(esp_console_cmd_register(&force_gain_cmd));

    set_args.mcs = arg_int0("m", "mcs", "[mcs]", "MCS [0, 7]");
    set_args.bcc = arg_int0("b", "bcc", "[bcc]", "enable/disable BCC");
    set_args.ltf = arg_int0("l", "ltf", "[ltf]", "LTF, 3.2, 6.4, 12.8, unit: us");
    set_args.gi = arg_int0("g", "gi", "[gi]", "GI, 0.8, 1.6, 3.2, unit: us");
    set_args.nss = arg_int0("s", "nss", "[nss]", "NSS");
    set_args.ersu = arg_int0("e", "ersu", "[ersu]", "ERSU");
    set_args.end = arg_end(1);
    const esp_console_cmd_t set_cmd = {
            .command = "set",
            .help = "set wireless settings",
            .hint = NULL,
            .func = &wifi_cmd_set_wireless,
            .argtable = &set_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&set_cmd));

    set_amsdu_args.enable = arg_dbl0("e", "enable", "[enable]", "set amsdu");
    set_amsdu_args.end = arg_end(1);
    const esp_console_cmd_t set_amsdu_cmd = {
            .command = "amsdu",
            .help = "enable amsdu",
            .hint = NULL,
                .func = &wifi_cmd_set_amsdu,
            .argtable = &set_amsdu_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&set_amsdu_cmd));

    set_obss_args.enable = arg_dbl0("e", "enable", "[enable]", "set amsdu");
    set_obss_args.end = arg_end(1);
    const esp_console_cmd_t set_obss_cmd = {
            .command = "obss_scan",
            .help = "enable obss_scan",
            .hint = NULL,
                .func = &wifi_cmd_set_obss,
            .argtable = &set_obss_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&set_obss_cmd));

    /* iperf */
    iperf_args.ip = arg_str0("c", "client", "<ip>", "run in client mode, connecting to <host>");
    iperf_args.server = arg_lit0("s", "server", "run in server mode");
    iperf_args.udp = arg_lit0("u", "udp", "use UDP rather than TCP");
    iperf_args.version = arg_lit0("V", "ipv6_domain", "use IPV6 address rather than IPV4");
    iperf_args.port = arg_int0("p", "port", "<port>", "server port to listen on/connect to");
    iperf_args.length = arg_int0("l", "len", "<length>", "Set read/write buffer size");
    iperf_args.interval = arg_int0("i", "interval", "<interval>", "seconds between periodic bandwidth reports");
    iperf_args.time = arg_int0("t", "time", "<time>", "time in seconds to transmit for (default 10 secs)");
    iperf_args.bw_limit = arg_int0("b", "bandwidth", "<bandwidth>", "bandwidth to send at in Mbits/sec");
    iperf_args.abort = arg_lit0("a", "abort", "abort running iperf");
    iperf_args.end = arg_end(1);
    const esp_console_cmd_t iperf_cmd = {
        .command = "iperf",
        .help = "iperf command",
        .hint = NULL,
        .func = &wifi_cmd_iperf,
        .argtable = &iperf_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&iperf_cmd));
    /* addba */
    addba_args.tid = arg_int0("t", "tid", "[tid]", "tid [0, 7]");
    addba_args.end = arg_end(1);
    const esp_console_cmd_t addba_cmd = {
        .command = "addba",
        .help = "create addba agreement",
        .hint = NULL,
        .func = &wifi_cmd_addba,
        .argtable = &addba_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&addba_cmd));

    /* set offset of target wake time */
    itwt_set_twt_offset_args.offset = arg_int0("o", NULL, "[offset]", "set target wake time offset, unit us");
    itwt_set_twt_offset_args.end = arg_end(1);
    const esp_console_cmd_t itwt_set_twt_offset_cmd = {
            .command = "offset",
            .help = "itwt target wake time offset",
            .hint = NULL,
            .func = &wifi_cmd_itwt_set_target_wake_time_offset,
            .argtable = &itwt_set_twt_offset_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&itwt_set_twt_offset_cmd));

    loglevel_args.level = arg_int1(NULL, NULL, "<level>", "log level: 1-error, 2-warning, 3-info");
    loglevel_args.end = arg_end(1);

    const esp_console_cmd_t loglevel_cmd = {
        .command = "loglevel",
        .help = "set log level",
        .hint = NULL,
        .func = &set_wifi_log_level,
        .argtable = &loglevel_args
    };
    ESP_ERROR_CHECK( esp_console_cmd_register(&loglevel_cmd) );

    settime_args.time = arg_int1(NULL, NULL, "<timestamp>", "timestamp");
    settime_args.end = arg_end(1);

    const esp_console_cmd_t settime_cmd = {
        .command = "settime",
        .help = "set time",
        .hint = NULL,
        .func = &set_timestamp,
        .argtable = &settime_args
    };
    ESP_ERROR_CHECK( esp_console_cmd_register(&settime_cmd) );

    setprog_args.prog = arg_str0("p", "program", "[program]", "Program [HE, PMF,N,G]");
    setprog_args.end = arg_end(1);
    const esp_console_cmd_t setprog_cmd = {
        .command = "set_prog",
        .help = "Set the bandwidth of the mode of operation. 20Mhz or 40Mhz",
        .hint = NULL,
        .func = &set_prog_cmd,
        .argtable = &setprog_args
    };
    ESP_ERROR_CHECK(esp_console_cmd_register(&setprog_cmd));

// #ifdef CONFIG_SOC_WIFI_HE_SUPPORT
//     register_wifi_cmd();
//     register_wifi_itwt();
//     register_wifi_stats();                
// #endif
}



