/* ESP Wi-Fi Allaince Sigma agent software
*
* Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
* Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.
*
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include "esp_sigma.h"
#include <signal.h>
#include <lwip/tcp.h>
#include "wifi_debug.h"
#include <pthread.h>
#include "esp_heap_trace.h"
pthread_mutex_t send_response_mutex;

void sigma_dut_print(struct sigma_dut *dut, int level, const char *fmt, ...)
{
    va_list ap;
    struct timeval tv;
    if (level < dut->debug_level) {
        return;
    }
    pthread_mutex_lock(&send_response_mutex);

    gettimeofday(&tv, NULL);

    printf("\n");
    va_start(ap, fmt);
    //printf("%ld: ", (long) tv.tv_sec);
    vprintf(fmt, ap);
    printf("\n");
    pthread_mutex_unlock(&send_response_mutex);
}

void sigma_dut_dev_print(const char *fmt, ...)
{
    va_list ap;
    struct timeval tv;

    pthread_mutex_lock(&send_response_mutex);

    gettimeofday(&tv, NULL);

    va_start(ap, fmt);
    vprintf(fmt, ap);
    pthread_mutex_unlock(&send_response_mutex);

}

void sigma_dut_summary(struct sigma_dut *dut, const char *fmt, ...)
{
    va_list ap;
    if (!dut->summary_log) {
        return;
    }

    va_start(ap, fmt);
    va_end(ap);
}

void send_resp(struct sigma_dut *dut,
               enum sigma_status status, const char *buf)
{
    struct msghdr msg;
    struct iovec iov[4];
    size_t elems;

    sigma_dut_print(dut, DUT_MSG_INFO, "\nresp: status=%d buf=%s\n", status, buf ? buf : "N/A");
    iov[0].iov_base = "status,";
    iov[0].iov_len = 7;
    switch (status) {
    case SIGMA_RUNNING:
        iov[1].iov_base = "RUNNING,";
        iov[1].iov_len = 8;
        break;
    case SIGMA_INVALID:
        iov[1].iov_base = "INVALID,";
        iov[1].iov_len = 8;
        break;
    case SIGMA_ERROR:
        iov[1].iov_base = "ERROR,";
        iov[1].iov_len = 6;
        break;
    case SIGMA_COMPLETE:
        iov[1].iov_base = "COMPLETE,";
        iov[1].iov_len = 9;
        break;
    }

    if (buf) {
        iov[2].iov_base = (void *) buf;
        iov[2].iov_len = strlen(buf);
        iov[3].iov_base = "\r\n";
        iov[3].iov_len = 2;
        elems = 4;
    } else {
        iov[1].iov_len--;
        iov[2].iov_base = "\r\n";
        iov[2].iov_len = 2;
        elems = 3;
    }

    memset(&msg, 0, sizeof(msg));
    msg.msg_iov = iov;
    msg.msg_iovlen = elems;

//    sigma_dut_dev_print("\nResponse sent !!!\n");

    dut->response_sent++;
}

const char * get_params(struct sigma_cmd *cmd, const char *name)
{
    int i;
    for (i = 0; i < cmd->count; i++) {
        if (strcasecmp(name, cmd->params[i]) == 0) {
            return cmd->values[i];
        }
    }
    return NULL;
}

static void task_traffic(void *arg)
{
    enum sigma_cmd_result res;
    struct sigma_dut *dut = arg;
    struct sigma_cmd c = dut->c;
    res = cmd_traffic_agent_send(dut, &c);
    //printf("\nQweQwe Send task result %d\n", res);
    switch (res) {
    case ERROR_SEND_STATUS:
        send_resp(dut, SIGMA_ERROR, NULL);
        break;
    case INVALID_SEND_STATUS:
        send_resp(dut, SIGMA_INVALID, NULL);
        break;
    case STATUS_SENT:
    case STATUS_SENT_ERROR:
        break;
    case SUCCESS_SEND_STATUS:
        send_resp(dut, SIGMA_COMPLETE, NULL);
        break;
    }
    //printf("\nDeleting the task\n");
    vTaskDelete(NULL);
}

static struct sigma_cmd c;
void process_sigma_traffic_cmd(struct sigma_dut *dut, char *buf, int traffic_capi_cmd)
{
    //struct sigma_cmd_handler *h;

    char *cmd, *pos = NULL, *pos2 = NULL;
    int len;
    char txt[300];
    enum sigma_cmd_result res = 0;

    while (*buf == '\r' || *buf == '\n' || *buf == '\t' || *buf == ' ') {
        buf++;
    }
    len = strlen(buf);
    while (len > 0 && buf[len - 1] == ' ') {
        buf[len - 1] = '\0';
        len--;
    }
#if 0
    if (dut->debug_level < DUT_MSG_INFO) {
        pos = strchr(buf, ',');
        if (pos == NULL) {
            pos = buf + len;
        }
        if (pos - buf > 50) {
            pos = buf + 50;
        }
        memcpy(txt, "/====[ ", 7);
        pos2 = txt + 7;
        memcpy(pos2, buf, pos - buf);
        pos2 += pos - buf;
        *pos2++ = ' ';
        *pos2++ = ']';
        while (pos2 - txt < 70) {
            *pos2++ = '=';
        }
        *pos2++ = '/';
        *pos2 = '\0';
        printf("\n%s\n\n", txt);
    }
#endif
    //sigma_dut_print(dut, DUT_MSG_INFO, "cmd: %s", buf);
    sigma_dut_summary(dut, "CAPI cmd: %s", buf);
    snprintf(txt, sizeof(txt), "NOTE CAPI:%s", buf);

    memset(&c, 0, sizeof(c));
    cmd = buf;
    pos = strchr(cmd, ',');
    if (pos) {
        *pos++ = '\0';
        while (pos) {
            pos2 = strchr(pos, ',');
            if (pos2 == NULL) {
                sigma_dut_print(dut, DUT_MSG_INFO, "Invalid "
                                "parameters");
                return;
            }
            *pos2++ = '\0';
            if (c.count == MAX_PARAMS) {
                sigma_dut_print(dut, DUT_MSG_INFO, "Too many "
                                "parameters");
                return;
            }
            c.params[c.count] = pos;
            c.values[c.count] = pos2;
            c.count++;
            pos = strchr(pos2, ',');
            if (pos) {
                *pos++ = '\0';
            }

        }
    }
    if (traffic_capi_cmd == SIGMA_TRAFFIC_SEND) {
        dut->c = c;
        // wifi_cmd_dump_ctrl(0, 0); //dump control frames
#if CONFIG_IDF_TARGET_ESP32C6 && CONFIG_ESP_WIFI_ENABLE_WIFI_TX_STATS
        wifi_cmd_clr_tx_statistics(0, 0); //clear tx statistics
#endif
        //printf("\nStarting the task!\n");
        int ret = xTaskCreatePinnedToCore(task_traffic, "traffic_start", 4096, dut, 5, NULL, portNUM_PROCESSORS - 1);
        if(ret != pdPASS) {
            printf("\nMemory allocation failed for creating send task\n");
        }
    } else {
        switch (traffic_capi_cmd) {
        case SIGMA_TRAFFIC_CONFIG:
            res = cmd_traffic_agent_config(dut, &c);
            break;
        case SIGMA_TRAFFIC_RECEIVE_START:
            res = cmd_traffic_agent_receive_start(dut, &c);
            break;
        case SIGMA_TRAFFIC_RECEIVE_STOP:
            res = cmd_traffic_agent_receive_stop(dut, &c);
            break;
        case SIGMA_TRAFFIC_RESET:
            res = cmd_traffic_agent_reset(dut, &c);
            break;
        default:
            res = SIGMA_INVALID;
            break;
        }
        switch (res) {
        case ERROR_SEND_STATUS:
            send_resp(dut, SIGMA_ERROR, NULL);
            break;
        case INVALID_SEND_STATUS:
            send_resp(dut, SIGMA_INVALID, NULL);
            break;
        case STATUS_SENT:
        case STATUS_SENT_ERROR:
            break;
        case SUCCESS_SEND_STATUS:
            if (traffic_capi_cmd == SIGMA_TRAFFIC_RECEIVE_START) {
                for(int f = 0; f < 5; f++){
                    send_resp(dut, SIGMA_COMPLETE, NULL);
                    usleep(100000);
                }
            } else  {
                send_resp(dut, SIGMA_COMPLETE, NULL);
            }
            break;
        }
    }
}
