/* ESP Wi-Fi Allaince Sigma agent software
*
* Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
* Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.
*
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <errno.h>
#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_wifi.h"
#include "esp_log.h"
#include "esp_err.h"
#include "nvs_flash.h"

#include "esp_console.h"
#include "esp_vfs_dev.h"
#include "driver/uart.h"
#include "linenoise/linenoise.h"
#include "argtable3/argtable3.h"
#include "cmd_decl.h"
#include <pthread.h>

#include "esp_mac.h"

extern void register_sta_set_wireless(void);

#define WIFI_CONNECTED_BIT BIT0

extern pthread_mutex_t send_response_mutex;
void app_main(void)
{
    esp_err_t ret = nvs_flash_init();
    if (ret == ESP_ERR_NVS_NO_FREE_PAGES || ret == ESP_ERR_NVS_NEW_VERSION_FOUND) {
        ESP_ERROR_CHECK(nvs_flash_erase());
        ret = nvs_flash_init();
    }
    uint8_t base_mac[6] = {0x10, 0x97, 0xbd, 0xf1, 0x9e, 0xa4};
    esp_base_mac_addr_set(base_mac);
    ESP_ERROR_CHECK(ret);
    initialise_wifi();
    int ret_mut = pthread_mutex_init(&send_response_mutex, NULL);

    printf("\n **************************************************\n");
    printf(" ***    Espressif Wi-Fi Sigma Agent Software    ***\n");
    printf(" ***                                            ***\n");
    printf(" ************************************************** \n\n");
    printf(" Enter 'help' for supported commands \n");

#if CONFIG_IDF_TARGET_ESP32C6
    esp_console_repl_t *repl = NULL;
    esp_console_repl_config_t repl_config = ESP_CONSOLE_REPL_CONFIG_DEFAULT();
    repl_config.task_stack_size += 1024;
    repl_config.prompt = "espsigma>";
    // init console REPL environment
#if CONFIG_ESP_CONSOLE_UART
    esp_console_dev_uart_config_t uart_config = ESP_CONSOLE_DEV_UART_CONFIG_DEFAULT();
    uart_config.baud_rate = 115200;
    // init console REPL environment
    ESP_ERROR_CHECK(esp_console_new_repl_uart(&uart_config, &repl_config, &repl));
#elif CONFIG_ESP_CONSOLE_USB_CDC
    esp_console_dev_usb_cdc_config_t cdc_config = ESP_CONSOLE_DEV_CDC_CONFIG_DEFAULT();
    ESP_ERROR_CHECK(esp_console_new_repl_usb_cdc(&cdc_config, &repl_config, &repl));
#endif
#else
    esp_console_repl_t *repl = NULL;
    esp_console_repl_config_t repl_config = ESP_CONSOLE_REPL_CONFIG_DEFAULT();
    repl_config.task_stack_size += 2048;
    repl_config.task_stack_size += 512;
    esp_console_dev_uart_config_t uart_config = ESP_CONSOLE_DEV_UART_CONFIG_DEFAULT();
    uart_config.baud_rate = 115200;
    repl_config.prompt = "espsigma>";
    // init console REPL environment
    ESP_ERROR_CHECK(esp_console_new_repl_uart(&uart_config, &repl_config, &repl));

#endif

    /* Register commands */
    register_system();
    register_wifi();
    register_traffic();

#if 0
    register_sta_set_wireless();
#endif

    // start console REPL
    ESP_ERROR_CHECK(esp_console_start_repl(repl));

}
