 /* ESP Wi-Fi Allaince Sigma agent software
 *
 * Original Copyright (C) 2006-2016, ARM Limited, All Rights Reserved, Apache 2.0 License.
 * Additions Copyright (C) Copyright 2015-2016 Espressif Systems (Shanghai) PTE LTD, Apache 2.0 License.
 *
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "esp_sigma.h"
#include "wifi_debug.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

#define USE_PTHREAD 0
#define PRINT_TASK_STATS 0 //change the value in cmd_wifi.c

#define TG_MAX_CLIENTS_CONNECTIONS 3
static int simultaneous_tx_streams_count;
static int simultaneous_rx_streams_count;
void print_task_list()
{
    char ptrTaskList[250];
    vTaskList(ptrTaskList);
    printf("\n**********************************\n");
    printf("Task  State   Prio    Stack    Num\n");
    printf("**********************************\n");
    printf("%s\n",ptrTaskList);
    printf("**********************************\n");
}


size_t strlcpy(char *dest, const char *src, size_t siz)
{
#if 1
	strncpy(dest, src, siz);
	return 0;
#else
	const char *s = src;
	size_t left = siz;

	if (left) {
		/* Copy string up to the maximum size of the dest buffer */
		while (--left != 0) {
			if ((*dest++ = *s++) == '\0')
				break;
		}
	}

	if (left == 0) {
		/* Not enough room for the string; force NUL-termination */
		if (siz != 0)
			*dest = '\0';
		while (*s++)
			; /* determine total src string length */
	}

	return s - src - 1;
#endif
}


size_t strlcat(char *dst, const char *str, size_t size)
{
	char *pos;
	size_t dstlen, srclen, copy;

	srclen = strlen(str);
	for (pos = dst; pos - dst < size && *dst; pos++)
		;
	dstlen = pos - dst;
	if (*dst)
		return dstlen + srclen;
	if (dstlen + srclen + 1 > size)
		copy = size - dstlen - 1;
	else
		copy = srclen;
	memcpy(pos, str, copy);
	pos[copy] = '\0';
	return dstlen + srclen;
}


int is_ipv6_addr(const char *str)
{
	struct sockaddr_in6 addr;

	return inet_pton(AF_INET6, str, &(addr.sin6_addr));
}

#if PRINT_TASK_STATS
// ---------------------- Real Time Stats Function ------------------

#define ARRAY_SIZE_OFFSET   5   //Increase this if print_real_time_stats returns ESP_ERR_INVALID_SIZE

typedef struct {
    TaskStatus_t *start_array;
    TaskStatus_t *end_array;
    UBaseType_t start_array_size;
    UBaseType_t end_array_size;
    uint32_t start_run_time;
    uint32_t end_run_time;
} rts_info_t;

rts_info_t *info = NULL;

void print_real_time_stats_start()
{
    if (info != NULL) {
        return;
    }
    info = malloc(sizeof(rts_info_t));
    //Allocate array to store current task states
    int num_tasks = uxTaskGetNumberOfTasks();
    info->start_array_size = num_tasks + ARRAY_SIZE_OFFSET;
    info->start_array = calloc(1, sizeof(TaskStatus_t) * info->start_array_size);
    //Get current task states
    //info->start_array_size = uxTaskGetSystemState(info->start_array, info->start_array_size, &info->start_run_time);
    //Print start array tasks
    sigma_dut_dev_print("Start tasks:\n");
    for (int i = 0; i < num_tasks; i++) {
        sigma_dut_dev_print("%s\n", info->start_array[i].pcTaskName);
    }
}

void print_real_time_stats_end()
{
    if (info == NULL) {
        return;
    }
    //Allocate array to store tasks states post delay
    info->end_array_size = uxTaskGetNumberOfTasks() + ARRAY_SIZE_OFFSET;
    info->end_array = calloc(1, sizeof(TaskStatus_t) * info->end_array_size);
    //Get post delay task states
    //info->end_array_size = uxTaskGetSystemState(info->end_array, info->end_array_size, &info->end_run_time);

    //Calculate total_elapsed_time in units of run time stats clock period.
    uint32_t total_elapsed_time = (info->end_run_time - info->start_run_time);
    sigma_dut_dev_print("Total elapse time %d\n", total_elapsed_time);

    sigma_dut_dev_print("|       Task       |     Run Time     |    Percentage    \n");
    //Match each task in start_array to those in the end_array
    for (int i = 0; i < info->start_array_size; i++) {
        int k = -1;
        for (int j = 0; j < info->end_array_size; j++) {
            if (info->start_array[i].xHandle == info->end_array[j].xHandle) {
                k = j;
                //Mark that task have been matched by overwriting their handles
                info->start_array[i].xHandle = NULL;
                info->end_array[j].xHandle = NULL;
                break;
            }
        }
        //Check if matching task found
        if (k >= 0) {
            uint32_t task_elapsed_time = info->end_array[k].ulRunTimeCounter - info->start_array[i].ulRunTimeCounter;
            //uint32_t percentage_time = (task_elapsed_time * 100UL) / (total_elapsed_time * portNUM_PROCESSORS);
            uint32_t percentage_time = (task_elapsed_time * 100UL) / total_elapsed_time;
            sigma_dut_dev_print("| %16s | %16d | %16d%%| %16d | %16d | %16d \n", info->start_array[i].pcTaskName, task_elapsed_time, percentage_time,
                    info->start_array[i].uxCurrentPriority,
                    info->start_array[i].uxBasePriority,
                    info->start_array[i].usStackHighWaterMark);
        }
    }
    sigma_dut_dev_print("\n");

    //Print unmatched tasks
    for (int i = 0; i < info->start_array_size; i++) {
        if (info->start_array[i].xHandle != NULL) {
            sigma_dut_dev_print("| %16s | Deleted\n", info->start_array[i].pcTaskName);
        }
	}
	for (int i = 0; i < info->end_array_size; i++) {
		if (info->end_array[i].xHandle != NULL) {
			sigma_dut_dev_print("| %16s | Created\n", info->end_array[i].pcTaskName);
        }
    }
    free(info->start_array);
    free(info->end_array);
    free(info);
    info = NULL;
}

// ------------------------------------------------------------------
#endif

enum sigma_cmd_result cmd_traffic_agent_config(struct sigma_dut *dut,
						      struct sigma_cmd *cmd)
{
	struct sigma_stream *s;
	const char *val;
	char buf[100];

	if (dut->num_streams == MAX_SIGMA_STREAMS) {
		send_resp(dut, SIGMA_ERROR, "errorCode,No more "
			  "concurrent traffic streams supported");
		return STATUS_SENT;
	}

	//printf("free traffic stats : %d\n", dut->num_streams);
	s = &dut->streams[dut->num_streams];
	free(s->stats);
	memset(s, 0, sizeof(*s));
	s->sock = -1;
	s->no_timestamps = dut->no_timestamps;
    
	val = get_params(cmd, "profile");

	if (!val)
		return INVALID_SEND_STATUS;

	if (strcasecmp(val, "File_Transfer") == 0)
		s->profile = SIGMA_PROFILE_FILE_TRANSFER;
	else if (strcasecmp(val, "Multicast") == 0)
		s->profile = SIGMA_PROFILE_MULTICAST;
	else if (strcasecmp(val, "IPTV") == 0)
		s->profile = SIGMA_PROFILE_IPTV;
	else if (strcasecmp(val, "Transaction") == 0)
		s->profile = SIGMA_PROFILE_TRANSACTION;
	else if (strcasecmp(val, "Start_Sync") == 0)
		s->profile = SIGMA_PROFILE_START_SYNC;
	else if (strcasecmp(val, "Uapsd") == 0)
		s->profile = SIGMA_PROFILE_UAPSD;
	else {
		send_resp(dut, SIGMA_INVALID, "errorCode,Unsupported "
			  "profile");
		return STATUS_SENT;
	}


	val = get_params(cmd, "direction");
	if (!val)
		return INVALID_SEND_STATUS;
	if (strcasecmp(val, "send") == 0)
		s->sender = 1;
	else if (strcasecmp(val, "receive") == 0)
		s->sender = 0;
	else
		return INVALID_SEND_STATUS;

	val = get_params(cmd, "destination");
	if (val) {
		if (!is_ipv6_addr(val)) {
			if (inet_aton(val, &s->dst) == 0)
				return INVALID_SEND_STATUS;
		} else {
			if (inet_pton(AF_INET6, val, &s->dst) != 1)
				return INVALID_SEND_STATUS;
		}
	}

	val = get_params(cmd, "source");
	if (val) {
		if (!is_ipv6_addr(val)) {
			if (inet_aton(val, &s->src) == 0)
				return INVALID_SEND_STATUS;
		} else {
			if (inet_pton(AF_INET6, val, &s->src) != 1)
				return INVALID_SEND_STATUS;
		}
	}

	val = get_params(cmd, "destinationPort");
	if (val)
		s->dst_port = atoi(val);

	val = get_params(cmd, "sourcePort");
	if (val)
		s->src_port = atoi(val);

	val = get_params(cmd, "frameRate");
	if (val)
		s->frame_rate = atoi(val);

	val = get_params(cmd, "duration");
	if (val)
		s->duration = atoi(val);

	val = get_params(cmd, "payloadSize");
	if (val)
		s->payload_size = atoi(val);

	val = get_params(cmd, "startDelay");
	if (val)
		s->start_delay = atoi(val);

	val = get_params(cmd, "maxCnt");
	if (val)
		s->max_cnt = atoi(val);

	val = get_params(cmd, "trafficClass");
	if (val) {
		if (strstr("Voice", val))
			s->tc = SIGMA_TC_VOICE;
		else if (strstr("Video", val))
			s->tc = SIGMA_TC_VIDEO;
		else if (strstr("Background", val))
			s->tc = SIGMA_TC_BACKGROUND;
		else if (strstr("BestEffort", val))
			s->tc = SIGMA_TC_BEST_EFFORT;
		else
			return INVALID_SEND_STATUS;
	}

	val = get_params(cmd, "userpriority");
	if (val) {
		s->user_priority_set = 1;
		s->user_priority = atoi(val);
	}

	val = get_params(cmd, "tagName");
	if (val) {
		strlcpy(s->test_name, val, sizeof(s->test_name));
		sigma_dut_print(dut, DUT_MSG_DEBUG,
				"Traffic agent: U-APSD console tagname %s",
				s->test_name);
	}

	if (dut->throughput_pktsize && s->frame_rate == 0 && s->sender &&
	    dut->throughput_pktsize != s->payload_size &&
	    (s->profile == SIGMA_PROFILE_FILE_TRANSFER ||
	     s->profile == SIGMA_PROFILE_IPTV ||
	     s->profile == SIGMA_PROFILE_UAPSD)) {
		sigma_dut_print(dut, DUT_MSG_INFO,
				"Traffic agent: Override throughput test payload size %u -> %u",
				s->payload_size, dut->throughput_pktsize);
		s->payload_size = dut->throughput_pktsize;
	}

	val = get_params(cmd, "transProtoType");
	if (val) {
		if (strcmp(val, "1") == 0)
			s->trans_proto = IPPROTO_TCP;
		else if (strcmp(val, "0") == 0)
			s->trans_proto = IPPROTO_UDP;
		else
			return INVALID_SEND_STATUS;
	} else {
		s->trans_proto = IPPROTO_UDP;
	}

	if (s->profile == SIGMA_PROFILE_IPTV && !s->sender && !s->no_timestamps)
	{

		//printf("1111 calloc traffic stats : %d\n", dut->num_streams);
		s->stats = calloc(MAX_SIGMA_STATS,
				  sizeof(struct sigma_frame_stats));
		if (s->stats == NULL)
			return ERROR_SEND_STATUS;
	}

	dut->stream_id++;
	dut->num_streams++;

	s->stream_id = dut->stream_id;
	snprintf(buf, sizeof(buf), "streamID,%d", s->stream_id);
	send_resp(dut, SIGMA_COMPLETE, buf);
	return STATUS_SENT;
}


static void stop_stream(struct sigma_stream *s)
{
	if (s && s->started) {
#if 1
		pthread_join(s->thr, NULL);
		if (s->sock != -1) {
			close(s->sock);
			s->sock = -1;
		}
#endif
		s->started = 0;
	}
}


enum sigma_cmd_result cmd_traffic_agent_reset(struct sigma_dut *dut,
						     struct sigma_cmd *cmd)
{
	int i;
	for (i = 0; i < dut->num_streams; i++) {
		struct sigma_stream *s = &dut->streams[i];
		s->stop = 1;
		stop_stream(s);
	}
	dut->num_streams = 0;
	simultaneous_tx_streams_count = 0;
	simultaneous_rx_streams_count = 0;
	memset(&dut->streams, 0, sizeof(dut->streams));
	return SUCCESS_SEND_STATUS;
}


static int get_stream_id(char *str, int streams[MAX_SIGMA_STREAMS])
{
	int count;
	count = 0;
	for (;;) {
		if (count == MAX_SIGMA_STREAMS)
			return -1;
		streams[count] = atoi(str);
		if (streams[count] == 0)
			return -1;
		count++;
		str = strchr(str, ' ');
		if (str == NULL)
			break;
		while (*str == ' ')
			str++;
	}
	return count;
}


static int open_socket_file_transfer(struct sigma_dut *dut,
				     struct sigma_stream *s)
{
	struct sockaddr_in addr;
	int sock_opt_val = 1;

	s->sock = socket(PF_INET, IPPROTO_UDP == s->trans_proto ? SOCK_DGRAM :
			 SOCK_STREAM, s->trans_proto);
	if (s->sock < 0) {
		perror("socket");
		return -1;
	}

	if (setsockopt(s->sock, SOL_SOCKET, SO_REUSEADDR, &sock_opt_val,
		       sizeof(sock_opt_val)) < 0) {
		perror("setsockopt");
		close(s->sock);
		s->sock = -1;
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(s->sender ? s->src_port : s->dst_port);
//	sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: sender=%d "
//			"bind port %d", s->sender, ntohs(addr.sin_port));
	if (bind(s->sock, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		perror("bind");
		close(s->sock);
		s->sock = -1;
		return -1;
	}
	
	if (s->profile == SIGMA_PROFILE_MULTICAST && !s->sender)
		return 0;

	if (s->trans_proto == IPPROTO_TCP && s->sender == 0) {
		if (listen(s->sock, TG_MAX_CLIENTS_CONNECTIONS ) < 0) {
			sigma_dut_print(dut, DUT_MSG_INFO,
					"Listen failed with error %d: %s",
					errno, strerror(errno));
			close(s->sock);
			s->sock = -1;
			return -1;
		}
	} else {
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = s->sender ? s->dst.s_addr :
			s->src.s_addr;
		addr.sin_port = htons(s->sender ? s->dst_port : s->src_port);
//		sigma_dut_print(dut, DUT_MSG_DEBUG,
//				"Traffic agent: connect %s:%d",
//				inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
		if (connect(s->sock, (struct sockaddr *) &addr, sizeof(addr)) <
		    0) {
			perror("connect");
			close(s->sock);
			s->sock = -1;
			return -1;
		}	
	}

	return 0;
}


static int open_socket_multicast(struct sigma_dut *dut, struct sigma_stream *s)
{
	if (open_socket_file_transfer(dut, s) < 0)
		return -1;

	if (!s->sender) {
		struct ip_mreq mr;
		memset(&mr, 0, sizeof(mr));
		mr.imr_multiaddr.s_addr = s->dst.s_addr;
		mr.imr_interface.s_addr = htonl(INADDR_ANY);
		sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: "
				"IP_ADD_MEMBERSHIP %s", inet_ntoa(s->dst));
		if (setsockopt(s->sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			       (void *) &mr, sizeof(mr)) < 0) {
			sigma_dut_print(dut, DUT_MSG_INFO,
					"setsockopt[IP_ADD_MEMBERSHIP]: %s",
					strerror(errno));
			/*
			 * Continue anyway since this can happen, e.g., if the
			 * default route is missing. This is not critical for
			 * multicast RX testing.
			 */
		}
	}

	return 0;
}


static int set_socket_prio(struct sigma_stream *s)
{
	int tos = 0x00;

	switch (s->tc) {
	case SIGMA_TC_VOICE:
		if (s->user_priority_set) {
			if (s->user_priority == 6)
				tos = 48 << 2;
			else if (s->user_priority == 7)
				tos = 56 << 2;
			else
				return -1;
		} else
			tos = 0xe0; /* DSCP = 56 */
		break;
	case SIGMA_TC_VIDEO:
		if (s->user_priority_set) {
			if (s->user_priority == 4)
				tos = 32 << 2;
			else if (s->user_priority == 5)
				tos = 40 << 2;
			else
				return -1;
		} else
			tos = 0xa0; /* DSCP = 40 */
		break;
	case SIGMA_TC_BACKGROUND:
		if (s->user_priority_set) {
			if (s->user_priority == 1)
				tos = 8 << 2;
			else if (s->user_priority == 2)
				tos = 16 << 2;
			else
				return -1;
		} else
			tos = 0x20; /* DSCP = 8 */
		break;
	case SIGMA_TC_BEST_EFFORT:
		if (s->user_priority_set) {
			if (s->user_priority == 0)
				tos = 0 << 2;
			else if (s->user_priority == 3)
				tos = 20 << 2;
			else
				return -1;
		} else
			tos = 0x00; /* DSCP = 0 */
		break;
	}

	if (setsockopt(s->sock, IPPROTO_IP, IP_TOS, &tos, sizeof(tos)) < 0) {
		perror("setsockopt");
		return -1;
	}

	return 0;
}


static int open_socket(struct sigma_dut *dut, struct sigma_stream *s)
{
	switch (s->profile) {
	case SIGMA_PROFILE_FILE_TRANSFER:
		return open_socket_file_transfer(dut, s);
	case SIGMA_PROFILE_MULTICAST:
		return open_socket_multicast(dut, s);
	case SIGMA_PROFILE_IPTV:
		if (open_socket_file_transfer(dut, s) < 0)
			return -1;
		return set_socket_prio(s);
	case SIGMA_PROFILE_TRANSACTION:
		return open_socket_file_transfer(dut, s);
	case SIGMA_PROFILE_UAPSD:
		return open_socket_file_transfer(dut, s);
	case SIGMA_PROFILE_START_SYNC:
		sigma_dut_print(dut, DUT_MSG_INFO, "Traffic stream profile %d "
				"not yet supported", s->profile);
		/* TODO */
		break;
	}

	return -1;
}


static void send_file_fast(struct sigma_stream *s, char *pkt)
{
	struct timeval stop, now;
	int res;
	unsigned int counter = 0;

	gettimeofday(&stop, NULL);
	stop.tv_sec += s->duration;

	while (!s->stop) {
		counter++;
		WPA_PUT_BE32(&pkt[8], counter);

		if ((counter & 0xf) == 0) {
			gettimeofday(&now, NULL);
			if (now.tv_sec > stop.tv_sec ||
			    (now.tv_sec == stop.tv_sec &&
			     now.tv_usec >= stop.tv_usec))
				break;
		}

		s->tx_act_frames++;
		res = send(s->sock, pkt, s->payload_size, MSG_DONTWAIT);
		if (res >= 0) {
			s->tx_frames++;
			s->tx_payload_bytes += res;
		} else {
			switch (errno) {
			case EAGAIN:
			case ENOBUFS:
				usleep(1000);
				break;
			case ECONNRESET:
			case EPIPE:
				s->stop = 1;
				break;
			default:
				perror("send");
				break;
			}
		}
	}
}


static void IRAM_ATTR  send_file(struct sigma_stream *s)
{	
	char *pkt;
	struct timeval stop, now, start;
	int res;
	unsigned int counter = 0, total_sleep_usec = 0, total_pkts;
	int sleep_usec = 0;
	int payloadsizesent = 0;
	if (s->duration <= 0 || s->frame_rate < 0 || s->payload_size < 20)
		return;
	//s->payload_size += 400;
	pkt = malloc(s->payload_size);
	if (pkt == NULL)
		return;
	memset(pkt, 1, s->payload_size);
	if (s->payload_size > 20)
		strlcpy(pkt, "1345678", s->payload_size);

	if (s->frame_rate == 0 && s->no_timestamps) {
		send_file_fast(s, pkt);
		free(pkt);
		return;
	}

	gettimeofday(&stop, NULL);
	stop.tv_sec += s->duration;
	int payloadsizemax = (456000)*(s->duration);
    if (simultaneous_rx_streams_count > 1) {
         printf("\nReduced the rate by 3500000\n");
         payloadsizemax -= 3500000;
    }

	total_pkts = s->duration * s ->frame_rate;

	gettimeofday(&start, NULL);

	while (!s->stop) {
		counter++;
		if (s->payload_size > 20)
			WPA_PUT_BE32(&pkt[8], counter);

		if (sleep_usec) {
			usleep(sleep_usec);
			total_sleep_usec += sleep_usec;
		}

		gettimeofday(&now, NULL);
#if 0
        if (now.tv_sec >= stop.tv_sec || (stop.tv_sec - now.tv_sec) <= 1) {
#else
        if (now.tv_sec > stop.tv_sec ||
			(now.tv_sec == stop.tv_sec && now.tv_usec >= stop.tv_usec)) {
#endif
			break;
	    }

		if (s->frame_rate && (unsigned int) s->tx_frames >= total_pkts)
			break;

		if (s->frame_rate == 0 || s->tx_frames == 0)
			sleep_usec = 0;
		else if (sleep_usec || s->frame_rate < 10 ||
			 counter % (s->frame_rate / 10) == 0) {
			/* Recalculate sleep_usec for every 100 ms approximately
			 */
			struct timeval tmp;
			int diff, duration;

			timersub(&now, &start, &tmp);

			diff = tmp.tv_sec * 1000000 + tmp.tv_usec;
			duration = (1000000 / s->frame_rate) * s->tx_frames;

			if (duration > diff)
				sleep_usec = (total_sleep_usec +
					      (duration - diff)) / s->tx_frames;
			else
				sleep_usec = 0;
		}
		if (s->payload_size > 20) {
			WPA_PUT_BE32(&pkt[12], now.tv_sec);
			WPA_PUT_BE32(&pkt[16], now.tv_usec);
		}
		s->tx_act_frames++;
		res = send(s->sock, pkt, s->payload_size, MSG_DONTWAIT);
		if (res >= 0) {
			s->tx_frames++;
			s->tx_payload_bytes += res;
                        payloadsizesent += res;
			if(((simultaneous_tx_streams_count > 1) || (simultaneous_rx_streams_count > 1)) && (payloadsizesent > payloadsizemax)){
				payloadsizesent = 0;
				sleep(stop.tv_sec - now.tv_sec + 2);

			}
		} else {
			switch (errno) {
			case EAGAIN:
			case ENOBUFS:
				//usleep(1000);
				vTaskDelay(1);
				break;
			case ECONNRESET:
			case EPIPE:
				s->stop = 1;
				break;
			default:
#if 0 //close to avoid printing "send: relative error message of errno" during the test
				perror("send");
#endif
				break;
			}
		}
	}
//	printf(" ---- exit: send_file free size %d\n", heap_caps_get_free_size(MALLOC_CAP_8BIT));
	// sigma_dut_print(s->dut, DUT_MSG_DEBUG,
	// 		"send_file: counter %u s->tx_frames %d total_sleep_usec %u",
	// 		counter, s->tx_frames, total_sleep_usec);

	free(pkt);
}


static void IRAM_ATTR send_transaction(struct sigma_stream *s)
{
	char *pkt, *rpkt;
	struct timeval stop, now;
	int res;
	unsigned int counter = 0, rcounter;
	int wait_time;
	fd_set rfds;
	struct timeval tv;

	if (s->duration <= 0 || s->frame_rate <= 0 || s->payload_size < 20)
		return;

	//s->payload_size += 400;
	pkt = malloc(s->payload_size);
	if (pkt == NULL)
		return;
	rpkt = malloc(s->payload_size);
	if (rpkt == NULL) {
		free(pkt);
		return;
	}
	memset(pkt, 1, s->payload_size);
	strlcpy(pkt, "1345678", s->payload_size);

	gettimeofday(&stop, NULL);
	stop.tv_sec += s->duration;

	wait_time = 1000000 / s->frame_rate;

	while (!s->stop) {
		counter++;
		if (s->max_cnt && (int) counter > s->max_cnt)
			break;
		WPA_PUT_BE32(&pkt[8], counter);

		gettimeofday(&now, NULL);
		if (now.tv_sec > stop.tv_sec ||
		    (now.tv_sec == stop.tv_sec && now.tv_usec >= stop.tv_usec))
			break;
		WPA_PUT_BE32(&pkt[12], now.tv_sec);
		WPA_PUT_BE32(&pkt[16], now.tv_usec);

		res = send(s->sock, pkt, s->payload_size, MSG_DONTWAIT);
		if (res >= 0) {
			s->tx_frames++;
			s->tx_payload_bytes += res;
		} else {
			switch (errno) {
			case EAGAIN:
			case ENOBUFS:
				//usleep(1000);
				vTaskDelay(1);
				break;
			case ECONNRESET:
			case EPIPE:
				s->stop = 1;
				break;
			default:
				perror("send");
				break;
			}
		}

		/* Wait for response */
		tv.tv_sec = 0;
		tv.tv_usec = wait_time;
		FD_ZERO(&rfds);
		FD_SET(s->sock, &rfds);
		res = select(s->sock + 1, &rfds, NULL, NULL, &tv);
		if (res < 0) {
			if (errno == EINTR)
				continue;
			// perror("select");
			break;
		}

		if (res == 0) {
			/* timeout */
			continue;
		}

		if (FD_ISSET(s->sock, &rfds)) {
			/* response received */
			res = recv(s->sock, rpkt, s->payload_size, 0);
			if (res < 0) {
				//perror("recv");
				break;
			}
			rcounter = WPA_GET_BE32(&rpkt[8]);
			if (rcounter != counter)
				s->out_of_seq_frames++;
			s->rx_frames++;
			s->rx_payload_bytes += res;
		}
	}

	sigma_dut_dev_print(" ---- exit: send_transaction\n");
	free(pkt);
	free(rpkt);
}


static void  IRAM_ATTR send_thread(void *ctx)
{
	struct sigma_stream *s = ctx;

	sleep(s->start_delay);

	switch (s->profile) {
	case SIGMA_PROFILE_FILE_TRANSFER:
		send_file(s);
		break;
	case SIGMA_PROFILE_MULTICAST:
		send_file(s);
		break;
	case SIGMA_PROFILE_IPTV:
		send_file(s);
		break;
	case SIGMA_PROFILE_TRANSACTION:
		send_transaction(s);
		break;
	case SIGMA_PROFILE_START_SYNC:
		break;

	case SIGMA_PROFILE_UAPSD:
#if CONFIG_WMM_POWER_SAVE
		send_uapsd_console(s);
#endif /*CONFIG_WMM_POWER_SAVE*/
		break;
	}
	vTaskDelete(NULL);
}


struct traffic_agent_send_data {
	struct sigma_dut *dut;
	struct sigma_conn *conn;
	int streams[MAX_SIGMA_STREAMS];
	int count;
};


static struct sigma_stream * get_stream(struct sigma_dut *dut, int id)
{
	int i;

	for (i = 0; i < dut->num_streams; i++) {
		if ((unsigned int) id == dut->streams[i].stream_id)
			return &dut->streams[i];
	}

	return NULL;
}


static void * send_report_thread(void *ctx)
{
	struct traffic_agent_send_data *data = ctx;
	struct sigma_dut *dut = data->dut;
	int i, ret;
	char *pos;
	char buf[100 + MAX_SIGMA_STREAMS * 60] = {0};

	for (i = 0; i < data->count; i++) {
//		sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: waiting "
//				"for stream %d send to complete",
//				data->streams[i]);
		stop_stream(get_stream(dut, data->streams[i]));
	}
//	printf("\nJoined with the send thread!\n");

	buf[0] = '\0';
	pos = buf;

	pos += snprintf(pos, buf + sizeof(buf) - pos, "streamID,");
	for (i = 0; i < data->count; i++) {
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
			       i > 0 ? " " : "", data->streams[i]);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}
	

	if (dut->program == PROGRAM_60GHZ) {
		sigma_dut_print(dut, DUT_MSG_INFO, "reporting tx_act_frames");
		pos += snprintf(pos, buf + sizeof(buf) - pos, ",txActFrames,");
		for (i = 0; i < data->count; i++) {
			struct sigma_stream *s;

			s = get_stream(dut, data->streams[i]);
			if (!s)
				continue;
			ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
				       i > 0 ? " " : "", s->tx_act_frames);
			if (ret < 0 || ret >= buf + sizeof(buf) - pos)
				break;
			pos += ret;
		}
	}


	pos += snprintf(pos, buf + sizeof(buf) - pos, ",txFrames,");
	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
			       i > 0 ? " " : "", s->tx_frames);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}


	pos += snprintf(pos, buf + sizeof(buf) - pos, ",rxFrames,");
	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
			       i > 0 ? " " : "", s->rx_frames);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}

	pos += snprintf(pos, buf + sizeof(buf) - pos, ",txB,");
	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s)
			continue;

		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%lld",
			       i > 0 ? " " : "", s->tx_payload_bytes);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}

	pos += snprintf(pos, buf + sizeof(buf) - pos, ",rxB,");
	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s)
			continue;

		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%lld",
			       i > 0 ? " " : "", s->rx_payload_bytes);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}

	pos += snprintf(pos, buf + sizeof(buf) - pos, ",OOSF,");
	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d\n",
			       i > 0 ? " " : "", s->out_of_seq_frames);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}

	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);
		if (!s)
			continue;
		s->ta_send_in_progress = 0;
		if (s->trans_proto == IPPROTO_TCP) {
			/*
			 * Close the socket to make sure client side close the
			 * network before the server. Otherwise, the server
			 * might get "Address already in use" when trying to
			 * reuse the port.
			 */
			close(s->sock);
			s->sock = -1;
			//sigma_dut_print(dut, DUT_MSG_DEBUG,"Closed the sender socket");
		}
	}

	buf[sizeof(buf) - 1] = '\0';
	send_resp(dut, SIGMA_COMPLETE, buf);
    for(int f = 0; f < 4; f++){
		send_resp(dut, SIGMA_COMPLETE, buf);
		usleep(100000);
	}
	free(data);
    //printf("\nQweQwe : exiting pthread\n");
    pthread_exit(NULL);
	return NULL;
}
enum sigma_cmd_result cmd_traffic_agent_send(struct sigma_dut *dut,
						    struct sigma_cmd *cmd)
{
	const char *val;
	int i, j, res;
	char buf[100];
	struct traffic_agent_send_data *data;
	val = get_params(cmd, "streamID");
	//printf("get_params: %s, val:%s\n", cmd, val);
	if (val == NULL)
		return INVALID_SEND_STATUS;
	data = calloc(1, sizeof(*data));
	if (data == NULL)
		return ERROR_SEND_STATUS;
	data->dut = dut;
	data->count = get_stream_id(val, data->streams);
    //printf("\nQweQwe :Streams count is %d \n", data->count);
	//sigma_dut_dev_print("get_stream_id: %s, data->streams, %d, data->count:%d\n", val, data->streams[0], data->count);
	if (data->count < 0) {
		free(data);
		return ERROR_SEND_STATUS;
	}
	for (i = 0; i < data->count; i++) {
    
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s) {
			snprintf(buf, sizeof(buf), "errorCode,StreamID %d "
				 "not configured", data->streams[i]);
			send_resp(dut, SIGMA_INVALID, buf);
			free(data);
            //printf("\nQweQwe : Error1\n");
			return STATUS_SENT;
		}
		for (j = 0; j < i; j++)
            //printf("\nQweQwe : In 33\n");
		if (!s->sender) {
			snprintf(buf, sizeof(buf), "errorCode,Not configured "
				 "as sender for streamID %d", data->streams[i]);
			send_resp(dut, SIGMA_INVALID, buf);
			free(data);
            //printf("\nQweQwe : Error2\n");
			return STATUS_SENT;
		}
		if (s->ta_send_in_progress) {
			send_resp(dut, SIGMA_ERROR,
				  "errorCode,Multiple concurrent send cmds on same streamID not supported");
			free(data);
            //printf("\nQweQwe : Error3\n");
			return STATUS_SENT;
		}
	        simultaneous_tx_streams_count++;
	}
	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s)
			continue;
		if (open_socket(dut, s) < 0) {
			free(data);
			return ERROR_SEND_STATUS;
		}
	}

	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (!s)
			continue;

		/*
		 * Provide dut context to the thread to support debugging and
		 * returning of error messages.
		 */
		s->dut = dut;

		char task_id[15];
		sprintf(task_id, "%s%d","send_thread", i);
		int ret;
		int tx_priority = 1;
		switch(s->tc){
			case SIGMA_TC_VIDEO:
				tx_priority = 10;
				break;
			case SIGMA_TC_VOICE:
				tx_priority = 20;
				break;
			case SIGMA_TC_BEST_EFFORT:
				tx_priority = 3;
				break;
			case SIGMA_TC_BACKGROUND:
				tx_priority = 2;
				break;
			default :
		                sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: Could not determine the traffic class priority. Received traffic class : %d ",s->tc);
				tx_priority = 1;
				break;

		}
		ret = xTaskCreatePinnedToCore(send_thread, task_id, 2048, s, tx_priority, NULL, portNUM_PROCESSORS - 1);
		//sigma_dut_dev_print("start:%s\n",task_id);
		if (ret != pdPASS) {
			sigma_dut_dev_print("create task %s failed\n", task_id);
			return ERROR_SEND_STATUS;
		}
		s->started = 1;
	}
    printf("\nFree memory is : %d \n", heap_caps_get_free_size(MALLOC_CAP_8BIT));
	//	sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: start a thread to track sending streams");
	struct sigma_stream *s = get_stream(dut, data->streams[0]);
	usleep((s->duration + 2)* 1000000);
#if CONFIG_IDF_TARGET_ESP32C6 && CONFIG_ESP_WIFI_ENABLE_WIFI_TX_STATS
	wifi_cmd_get_tx_statistics(0, NULL);
#endif
	pthread_attr_t  attrs;
	pthread_attr_init(&attrs);
    attrs.detachstate = PTHREAD_CREATE_DETACHED;
	pthread_attr_setstacksize(&attrs, 4096);
	res = pthread_create(&dut->thr, &attrs, send_report_thread, data);
	if (res) {
		perror( "pthread_create send:");
		//sigma_dut_print(dut, DUT_MSG_INFO, "pthread_create failed: %d",
		//		res);
		free(data);
		return ERROR_SEND_STATUS;
	}

	//sigma_dut_dev_print("data:%p\n", data);
	//sigma_dut_dev_print("data->count:%d,data->streams:%p\n", data->count, data->streams);
	for (i = 0; i < data->count; i++) {
		struct sigma_stream *s = get_stream(dut, data->streams[i]);

		if (s)
			s->ta_send_in_progress = 1;
	}

		//free(data);


	/* Command will be completed in send_report_thread() */

	return STATUS_SENT;
}


static void receive_file(struct sigma_stream *s)
{
	struct timeval tv, now;
	fd_set rfds;
	int res;
	char *pkt;
	int pktlen;
	unsigned int last_rx = 0, counter;

	pktlen = 1400;
	pkt = malloc(pktlen);
	if (pkt == NULL)
		return;

	while (!s->stop) {
		res = recv(s->sock, pkt, pktlen, 0);
		if (res >= 0) {
			s->rx_frames++;
			s->rx_payload_bytes += res;

			counter = WPA_GET_BE32(&pkt[8]);
			if (counter < last_rx)
				s->out_of_seq_frames++;
			last_rx = counter;
		} else {
			//perror("recv");
			//sigma_dut_dev_print("\nThere is an error in the receiving socket!\n");
			break;
		}

		if (res >= 20 && s->stats &&
			s->num_stats < MAX_SIGMA_STATS) {
			struct sigma_frame_stats *stats;
			stats = &s->stats[s->num_stats];
			s->num_stats++;
			gettimeofday(&now, NULL);
			stats->seqnum = counter;
			stats->local_sec = now.tv_sec;
			stats->local_usec = now.tv_usec;
			stats->remote_sec = WPA_GET_BE32(&pkt[12]);
			stats->remote_usec = WPA_GET_BE32(&pkt[16]);
		}
	}

	free(pkt);
}


static void receive_transaction(struct sigma_stream *s)
{
	struct timeval tv;
	fd_set rfds;
	int res;
	char *pkt;
	int pktlen;
	unsigned int last_rx = 0, counter;
	struct sockaddr_in addr;
	socklen_t addrlen;

	if (s->payload_size)
		pktlen = s->payload_size;
	else
		pktlen = 1024 + 1;
		//pktlen = 65536 + 1;
	
	pkt = malloc(pktlen);
	if (pkt == NULL)
		return;

	while (!s->stop) {
		FD_ZERO(&rfds);
		FD_SET(s->sock, &rfds);
		tv.tv_sec = 0;
		tv.tv_usec = 300000;
		res = select(s->sock + 1, &rfds, NULL, NULL, &tv);
		if (res < 0) {
			perror("select");
			usleep(10000);
		} else if (FD_ISSET(s->sock, &rfds)) {
			addrlen = sizeof(addr);
			res = recvfrom(s->sock, pkt, pktlen, 0,
				       (struct sockaddr *) &addr, &addrlen);
			if (res < 0) {
				//perror("recv");
				//sigma_dut_dev_print("\nThere is an error in the receiving socket!\n");
				break;
			}

			s->rx_frames++;
			s->rx_payload_bytes += res;

			counter = WPA_GET_BE32(&pkt[8]);
			if (counter < last_rx)
				s->out_of_seq_frames++;
			last_rx = counter;

			/* send response */
			res = sendto(s->sock, pkt, pktlen, 0,
				     (struct sockaddr *) &addr, addrlen);
			if (res < 0) {
				perror("sendto");
			} else {
				s->tx_frames++;
				s->tx_payload_bytes += res;
			}
		}
	}

	free(pkt);
}


#if USE_PTHREAD
static void * receive_thread(void *ctx)
#else
static void receive_thread(void *ctx)
#endif
{	
	struct sigma_stream *s = ctx;

	if (s->trans_proto == IPPROTO_TCP) {
		/* Wait for socket to be accepted */
		struct sockaddr_in connected_addr;
		int connected_sock; /* returned from accept on sock */
		socklen_t connected_addr_len = sizeof(connected_addr);

		sigma_dut_print(s->dut, DUT_MSG_DEBUG,
				"Traffic agent: Waiting on accept");
		connected_sock = accept(s->sock,
					(struct sockaddr *) &connected_addr,
					&connected_addr_len);
		if (connected_sock < 0) {
			sigma_dut_print(s->dut, DUT_MSG_ERROR,
					"Traffic agent: Failed to accept: %s",
					strerror(errno));
#if USE_PTHREAD
			return NULL;
#endif
		}

		//sigma_dut_print(s->dut, DUT_MSG_DEBUG,"Traffic agent: Accepted client closing parent socket and talk over connected sock.");
		close(s->sock);
		s->sock = connected_sock;
	}

	switch (s->profile) {
	case SIGMA_PROFILE_FILE_TRANSFER:
		receive_file(s);
		break;
	case SIGMA_PROFILE_MULTICAST:
		receive_file(s);
		break;
	case SIGMA_PROFILE_IPTV:
		receive_file(s);
		break;
	case SIGMA_PROFILE_TRANSACTION:
		receive_transaction(s);
		break;
	case SIGMA_PROFILE_START_SYNC:
		break;
	case SIGMA_PROFILE_UAPSD:
#if CONFIG_WMM_POWER_SAVE
		receive_uapsd(s);
#endif /*CONFIG_WMM_POWER_SAVE*/
		break;
	}

#if USE_PTHREAD
	return NULL;
#else

#if 0
	vTaskDelay(portMAX_DELAY);
#else
	vTaskDelete(NULL);
#endif
#endif /* USE_PTHREAD */
}


enum sigma_cmd_result
cmd_traffic_agent_receive_start(struct sigma_dut *dut,
				struct sigma_cmd *cmd)
{
	const char *val;
	int streams[MAX_SIGMA_STREAMS];
	int i, j, count;
	static char buf[100];

	val = get_params(cmd, "streamID");
	if (val == NULL)
		return INVALID_SEND_STATUS;
	count = get_stream_id(val, streams);
	if (count < 0)
		return ERROR_SEND_STATUS;
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s) {
			snprintf(buf, sizeof(buf), "errorCode,StreamID %d "
				 "not configured", streams[i]);
			send_resp(dut, SIGMA_INVALID, buf);
			return STATUS_SENT;
		}
		for (j = 0; j < i; j++)
			if (streams[i] == streams[j])
				return ERROR_SEND_STATUS;
		if (s->sender) {
			snprintf(buf, sizeof(buf), "errorCode,Not configured "
				 "as receiver for streamID %d", streams[i]);
			send_resp(dut, SIGMA_INVALID, buf);
			return STATUS_SENT;
		}
	}

	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s)
			continue;
		//sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: open receive socket for stream %d", streams[i]);
		if (open_socket(dut, s) < 0)
			return ERROR_SEND_STATUS;
	}

	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);
		int res;

		if (!s)
			continue;
		/*
		 * Provide dut context to the thread to support debugging and
		 * returning of error messages. Similarly, provide interface
		 * information to the thread. If the Interface parameter is not
		 * passed, get it from get_station_ifname() since the interface
		 * name is needed for power save mode configuration for Uapsd
		 * cases.
		 */
		s->dut = dut;
		val = get_params(cmd, "Interface");
		strlcpy(s->ifname, (val ? val : "wlan0"),
			sizeof(s->ifname));

		//sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: start " "receive for stream %d", streams[i]);
#if USE_PTHREAD
        printf("\nFree before rx memory is : %d \n", heap_caps_get_free_size(MALLOC_CAP_8BIT));
		res = pthread_create(&s->thr, NULL, receive_thread, s);
        printf("\nFree after rx memory is : %d \n", heap_caps_get_free_size(MALLOC_CAP_8BIT));
		if (res) {
#else
        char task_id[12] = { 0, };
        sprintf(task_id, "%s%d","recv_", s->dst_port);
        int rx_priority = 20;
        if (simultaneous_tx_streams_count > 1) {
            rx_priority = 10; 
        }
        /*As in almost every case, where there are multiple rx streams (mostly in 5.2.27) it is the second stream that is used for measurement.. so just giving it a higher priority than the rest*/
        if (simultaneous_rx_streams_count >= 1) {
            rx_priority++;
        }
        printf("\nFree before rx memory is : %d \n", heap_caps_get_free_size(MALLOC_CAP_8BIT));
        res = xTaskCreatePinnedToCore(receive_thread, task_id, 2048, s, rx_priority, NULL, portNUM_PROCESSORS - 1);
        printf("\nFree after rx memory is : %d \n", heap_caps_get_free_size(MALLOC_CAP_8BIT));
        if (res != pdPASS) {
#endif
		    perror( "pthread_create recv:");
			return ERROR_SEND_STATUS;
		}
#if 0
        if (s->dst_port == 4602) {
            print_real_time_stats_start();
        }
#endif

	    simultaneous_rx_streams_count++;
		s->started = 1;
	}
	return SUCCESS_SEND_STATUS;
}


static void write_frame_stats(struct sigma_dut *dut, struct sigma_stream *s,
			      int id)
{
	static char fname[128];
	FILE *f;
	unsigned int i;

	snprintf(fname, sizeof(fname), SIGMA_TMPDIR "/e2e%u-%d.txt",
		 (unsigned int) time(NULL), id);
	f = fopen(fname, "w");
	if (f == NULL) {
		sigma_dut_print(dut, DUT_MSG_INFO, "Could not write %s",
				fname);
		return;
	}
	fprintf(f, "seqnum:local_sec:local_usec:remote_sec:remote_usec\n");

	//sigma_dut_print(dut, DUT_MSG_DEBUG, "Writing frame stats to %s",fname);

	for (i = 0; i < s->num_stats; i++) {
		struct sigma_frame_stats *stats = &s->stats[i];
		fprintf(f, "%u:%u:%u:%u:%u\n", stats->seqnum,
			stats->local_sec, stats->local_usec,
			stats->remote_sec, stats->remote_usec);
	}

	fclose(f);
}


enum sigma_cmd_result
cmd_traffic_agent_receive_stop(struct sigma_dut *dut,
			       struct sigma_cmd *cmd)
{
	const char *val;
	int streams[MAX_SIGMA_STREAMS];
	int i, j, ret, count;
	char buf[256] = {0};
	char *pos;
	
	val = get_params(cmd, "streamID");
	if (val == NULL)
		return INVALID_SEND_STATUS;
	count = get_stream_id(val, streams);
	if (count < 0)
		return ERROR_SEND_STATUS;
		
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s) {
			snprintf(buf, sizeof(buf), "errorCode,StreamID %d "
				 "not configured", streams[i]);
			send_resp(dut, SIGMA_INVALID, buf);
			return STATUS_SENT;
		}
		for (j = 0; j < i; j++)
			if (streams[i] == streams[j])
				return ERROR_SEND_STATUS;
		if (!s->started) {
			snprintf(buf, sizeof(buf), "errorCode,Receive not "
				 "started for streamID %d", streams[i]);
			send_resp(dut, SIGMA_INVALID, buf);
			return STATUS_SENT;
		}
	}

	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);
		if (s){
			//sigma_dut_print(dut, DUT_MSG_DEBUG, "Traffic agent: stop receive for stream %d", streams[i]);
			s->stop = 1;
			stop_stream(s);
		}
	}
	
	
	buf[0] = '\0';
	pos = buf;
    
	pos += snprintf(pos, buf + sizeof(buf) - pos, "streamID,");
	for (i = 0; i < count; i++) {
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
			       i > 0 ? " " : "", streams[i]);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}
	
	if (dut->program == PROGRAM_60GHZ) {
		pos += snprintf(pos, buf + sizeof(buf) - pos, ",txActFrames,");
		for (i = 0; i < count; i++) {
			struct sigma_stream *s = get_stream(dut, streams[i]);

			if (!s)
				continue;
			ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
				       i > 0 ? " " : "", s->tx_act_frames);
			if (ret < 0 || ret >= buf + sizeof(buf) - pos)
				break;
			pos += ret;
		}
	}

	pos += snprintf(pos, buf + sizeof(buf) - pos, ",txFrames,");
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
			       i > 0 ? " " : "", s->tx_frames);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}

	pos += snprintf(pos, buf + sizeof(buf) - pos, ",rxFrames,");
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
			       i > 0 ? " " : "", s->rx_frames);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}
	pos += snprintf(pos, buf + sizeof(buf) - pos, ",txB,");
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%lld",
			       i > 0 ? " " : "", s->tx_payload_bytes);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}
    
	pos += snprintf(pos, buf + sizeof(buf) - pos, ",rxB,");
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%lld",
			       i > 0 ? " " : "", s->rx_payload_bytes);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}
  
	pos += snprintf(pos, buf + sizeof(buf) - pos, ",OOSF,");
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s)
			continue;
		ret = snprintf(pos, buf + sizeof(buf) - pos, "%s%d",
			       i > 0 ? " " : "", s->out_of_seq_frames);
		if (ret < 0 || ret >= buf + sizeof(buf) - pos)
			break;
		pos += ret;
	}
	
	buf[sizeof(buf) - 1] = '\0';
	for(int f = 0; f < 5; f++){
		send_resp(dut, SIGMA_COMPLETE, buf);
		usleep(100000);
	}
	for (i = 0; i < count; i++) {
		struct sigma_stream *s = get_stream(dut, streams[i]);

		if (!s)
			continue;
		if (s->profile == SIGMA_PROFILE_IPTV && s->num_stats > 0 &&
		    dut->write_stats)
			write_frame_stats(dut, s, streams[i]);

		free(s->stats);
		s->stats = NULL;
		s->num_stats = 0;
	}
	return STATUS_SENT;
}

enum sigma_cmd_result cmd_traffic_agent_version(struct sigma_dut *dut,
						       struct sigma_cmd *cmd)
{
	send_resp(dut, SIGMA_COMPLETE, "version,1.0");
	return STATUS_SENT;
}
