import argparse
import sys
from pathlib import Path
from idf_build_apps import build_apps, find_apps, setup_logging


PROJECT_ROOT = Path(__file__).parents[2].absolute()
IGNORE_WARNINGS = [
    r'expects argument of type',
    r'pkg_resources is deprecated',
    r'unused variable',
    r'discards \'const\'',
    r'format \'%',
    r'is deprecated: Use',
    r'unknown kconfig symbol'
]


def main():
    parser = argparse.ArgumentParser(
        description='Build all the apps for different targets',
    )
    parser.add_argument('paths', nargs='*', help='Paths to the apps to build.')
    parser.add_argument(
        '-t', '--target',
        default='all',
        help='Build apps for given target. could pass "all" to get apps for all targets',
    )

    args = parser.parse_args()
    setup_logging(verbose=1)  # INFO

    paths = args.paths if args.paths else [str(PROJECT_ROOT / 'espsigma' / 'esp_sigma_dut')]
    apps = find_apps(
        paths,
        recursive=True,
        target=args.target,
        build_dir='build_@t_@w',
        config_rules_str=['sdkconfig.ci=default', 'sdkconfig.ci.*=', '=default'],
        check_warnings=True,
        preserve=True,
        manifest_files=[],
    )
    apps = [app for app in apps if app.target not in ['esp32h2', 'esp32p4']]

    res = build_apps(
        apps,
        keep_going=False,
        collect_size_info='size_info.txt',
        ignore_warning_strs=IGNORE_WARNINGS,
        copy_sdkconfig=True,
    )
    sys.exit(res)


if __name__ == '__main__':
    main()
