#!/usr/bin/env bash
set -e

PYTHON_VERSION="3.7.7"

error(){
  echo $1
  exit 1
}

check_path(){
  if [ ! -d ./esp_sigma_ca ]; then
    error "Please run this script at sigma home path."
  fi
}

setup_pyenv(){
  if [ -d ~/.pyenv ]; then
    source ~/.pyenv/activate || error "~/.pyenv already exists. can not install new pyenv!"
    return
  fi
  mkdir ~/.pyenv

  git clone https://github.com/pyenv/pyenv.git ~/.pyenv/pyenv || \
    git clone https://gitee.com/mirrors/pyenv.git ~/.pyenv/pyenv
  cp ./tools/setup/pyenv_activate ~/.pyenv/activate

  source ~/.pyenv/activate
}


install_python(){
  sudo apt-get install -y make build-essential libssl-dev zlib1g-dev \
    libbz2-dev libreadline-dev libsqlite3-dev wget curl llvm \
    libncursesw5-dev xz-utils tk-dev libxml2-dev libxmlsec1-dev libffi-dev liblzma-dev

  PYENV_ROOT=$(pyenv root)
  mkdir -p ${PYENV_ROOT}/cache/
  wget https://www.python.org/ftp/python/${PYTHON_VERSION}/Python-${PYTHON_VERSION}.tar.xz -O ${PYENV_ROOT}/cache/Python-${PYTHON_VERSION}.tar.xz || \
    wget https://npm.taobao.org/mirrors/python/${PYTHON_VERSION}/Python-${PYTHON_VERSION}.tar.xz -O ${PYENV_ROOT}/cache/Python-${PYTHON_VERSION}.tar.xz

  pyenv install ${PYTHON_VERSION}
  pyenv local  ${PYTHON_VERSION}
  pip install --upgrade pip
}


check_path
# sudo apt-get update
sudo apt-get install -y git wget

echo "Seting pyenv ..."
pyenv --version || setup_pyenv

echo "Seting python ${PYTHON_VERSION} ..."
pyenv local ${PYTHON_VERSION} || install_python

echo "Installing requirements ..."
pip install -r requirements.txt

echo "======================== Done ============================="
echo "python ${PYTHON_VERSION} has been installed by pyenv."
echo "You may need to 'source ~/.pyenv/activate' before using python ${PYTHON_VERSION}."
echo "Or:"
echo "  echo source ~/.pyenv/activate >> ~/.bashrc"
echo "then:"
echo "  source ~/.bashrc"
